@@ -14,15 +14,18 @@ typedef struct sasl_message_ sasl_message_t;
 typedef struct sasl_mechanism_ sasl_mechanism_t;
 
 struct sasl_session_ {
-  char uid[IDLEN];
+  char *uid;
   char *buf, *p;
   int len, flags;
 
+  server_t *server;
+
   struct sasl_mechanism_ *mechptr;
   void *mechdata;
 
   char *username;
   char *certfp;
+  char *authzid;
 };
 
 struct sasl_message_ {
@@ -30,15 +33,28 @@ struct sasl_message_ {
   char mode;
   char *buf;
   char *ext;
+
+  server_t *server;
 };
 
 struct sasl_mechanism_ {
-  char name[21];
-  int (*mech_start) (struct sasl_session_ *sptr, char **buffer, int *buflen);
-  int (*mech_step) (struct sasl_session_ *sptr, char *message, int length, char **buffer, int *buflen);
+  char name[60];
+  int (*mech_start) (struct sasl_session_ *sptr, char **buffer, size_t *buflen);
+  int (*mech_step) (struct sasl_session_ *sptr, char *message, size_t length, char **buffer, size_t *buflen);
   void (*mech_finish) (struct sasl_session_ *sptr);
 };
 
+typedef struct {
+  myuser_t *source_mu;
+  myuser_t *target_mu;
+  bool allowed;
+} hook_sasl_may_impersonate_t;
+
+typedef struct {
+	void (*mech_register) (struct sasl_mechanism_ *mech);
+	void (*mech_unregister) (struct sasl_mechanism_ *mech);
+} sasl_mech_register_func_t;
+
 #define ASASL_FAIL 0 /* client supplied invalid credentials / screwed up their formatting */
 #define ASASL_MORE 1 /* everything looks good so far, but we're not done yet */
 #define ASASL_DONE 2 /* client successfully authenticated */
