@@ -22,7 +22,7 @@ DECLARE_MODULE_V1
 #define DB_ATHEME	2
 
 /* loads atheme.db */
-static void flatfile_db_load(void)
+static void flatfile_db_load(const char *filename)
 {
 	myuser_t *mu, *founder = NULL;
 	myuser_name_t *mun;
@@ -38,18 +38,20 @@ static void flatfile_db_load(void)
 	size_t bufsize = BUFSIZE, n;
 	int c;
 	unsigned int their_ca_all = 0;
+	char path[BUFSIZE];
 
-	f = fopen(DATADIR "/atheme.db", "r");
+	snprintf(path, BUFSIZE, "%s/%s", datadir, filename != NULL ? filename : "atheme.db");
+	f = fopen(path, "r");
 	if (f == NULL)
 	{
 		if (errno == ENOENT)
 		{
-			slog(LG_ERROR, "db_load(): %s does not exist.  as no data will be converted, you should use the 'opensex' backend now.", DATADIR "/atheme.db");
+			slog(LG_ERROR, "db_load(): %s does not exist.  as no data will be converted, you should use the 'opensex' backend now.", path);
 			exit(EXIT_FAILURE);
 		}
 		else
 		{
-			slog(LG_ERROR, "db_load(): can't open %s for reading: %s", DATADIR "/atheme.db", strerror(errno));
+			slog(LG_ERROR, "db_load(): can't open %s for reading: %s", path, strerror(errno));
 			slog(LG_ERROR, "db_load(): exiting to avoid data loss");
 			exit(EXIT_FAILURE);
 		}
@@ -76,7 +78,7 @@ static void flatfile_db_load(void)
 
 		if (c == EOF && ferror(f))
 		{
-			slog(LG_ERROR, "db_load(): error while reading %s: %s", DATADIR "/atheme.db", strerror(errno));
+			slog(LG_ERROR, "db_load(): error while reading %s: %s", path, strerror(errno));
 			slog(LG_ERROR, "db_load(): exiting to avoid data loss");
 			exit(EXIT_FAILURE);
 		}
@@ -215,8 +217,8 @@ static void flatfile_db_load(void)
 
 			mz = smalloc(sizeof(mymemo_t));
 
-			strlcpy(mz->sender, sender, NICKLEN);
-			strlcpy(mz->text, text, MEMOLEN);
+			mowgli_strlcpy(mz->sender, sender, NICKLEN);
+			mowgli_strlcpy(mz->text, text, MEMOLEN);
 			mz->sent = mtime;
 			mz->status = status;
 
@@ -234,15 +236,15 @@ static void flatfile_db_load(void)
 			target = strtok(NULL, "\n");
 
 			mu = myuser_find(user);
-			
+
 			if (!mu)
 			{
 				slog(LG_DEBUG, "db_load(): invalid ignore (MI %s %s)", user,target);
 				continue;
 			}
-			
+
 			strbuf = sstrdup(target);
-			
+
 			mowgli_node_add(strbuf, mowgli_node_create(), &mu->memo_ignores);
 		}
 		else if (!strcmp("AC", item))
@@ -309,43 +311,6 @@ static void flatfile_db_load(void)
 
 			mcfp = mycertfp_add(mu, certfp);
 		}
-		else if (!strcmp("SU", item))
-		{
-			/* subscriptions */
-			char *user, *sub_user, *tags, *tag;
-			myuser_t *subscriptor;
-			metadata_subscription_t *md;
-
-			user = strtok(NULL, " ");
-			sub_user = strtok(NULL, " ");
-			tags = strtok(NULL, "\n");
-			if (!user || !sub_user || !tags)
-			{
-				slog(LG_INFO, "db_load(): invalid subscription (line %d)", linecnt);
-				continue;
-			}
-
-			strip(tags);
-
-			mu = myuser_find(user);
-			subscriptor = myuser_find(sub_user);
-			if (!mu || !subscriptor)
-			{
-				slog(LG_INFO, "db_load(): invalid subscription <%s,%s> (line %d)", user, sub_user, linecnt);
-				continue;
-			}
-
-			md = smalloc(sizeof(metadata_subscription_t));
-			md->mu = subscriptor;
-
-			tag = strtok(tags, ",");
-			do
-			{
-				mowgli_node_add(sstrdup(tag), mowgli_node_create(), &md->taglist);
-			} while ((tag = strtok(NULL, ",")) != NULL);
-
-			mowgli_node_add(md, mowgli_node_create(), &mu->subscriptions);
-		}
 		else if (!strcmp("NAM", item))
 		{
 			/* formerly registered name (created by a marked account being dropped) */
@@ -537,7 +502,7 @@ static void flatfile_db_load(void)
 					if (fl == OLD_CA_AOP && versn < 4)
 						fl = CA_AOP_DEF;
 
-					/* 
+					/*
 					 * If the database revision is version 6 or newer, CA entries are
 					 * timestamped... otherwise we use 0 as the last modified TS
 					 *    --nenolod/jilles
@@ -579,9 +544,9 @@ static void flatfile_db_load(void)
 						fl |= CA_USEOWNER;
 
 					if ((!mu) && (validhostmask(causer)))
-						ca = chanacs_add_host(mc, causer, fl, ts);
+						ca = chanacs_add_host(mc, causer, fl, ts, NULL);
 					else
-						ca = chanacs_add(mc, entity(mu), fl, ts);
+						ca = chanacs_add(mc, entity(mu), fl, ts, NULL);
 				}
 				else if (versn == DB_SHRIKE)	/* DB_SHRIKE */
 				{
@@ -609,9 +574,9 @@ static void flatfile_db_load(void)
 					}
 
 					if ((!mu) && (validhostmask(causer)))
-						ca = chanacs_add_host(mc, causer, fl2, ts);
+						ca = chanacs_add_host(mc, causer, fl2, ts, NULL);
 					else
-						ca = chanacs_add(mc, entity(mu), fl2, ts);
+						ca = chanacs_add(mc, entity(mu), fl2, ts, NULL);
 				}
 			}
 		}
