@@ -95,7 +95,7 @@ void _modinit(module_t *m)
 	command_add(&cs_role_del, cs_role_cmds);
 }
 
-void _moddeinit()
+void _moddeinit(module_unload_intent_t intent)
 {
 	service_named_unbind_command("chanserv", &cs_access);
 	service_named_unbind_command("chanserv", &cs_role);
@@ -171,7 +171,7 @@ static void cs_cmd_access(sourceinfo_t *si, int parc, char *parv[])
 	if (parc > 2)
 		snprintf(buf, BUFSIZE, "%s %s", chan, parv[2]);
 	else
-		strlcpy(buf, chan, BUFSIZE);
+		mowgli_strlcpy(buf, chan, BUFSIZE);
 
 	command_exec_split(si->service, si, c->name, buf, cs_access_cmds);
 }
@@ -211,7 +211,7 @@ static void cs_cmd_role(sourceinfo_t *si, int parc, char *parv[])
 	if (parc > 2)
 		snprintf(buf, BUFSIZE, "%s %s", chan, parv[2]);
 	else
-		strlcpy(buf, chan, BUFSIZE);
+		mowgli_strlcpy(buf, chan, BUFSIZE);
 
 	command_exec_split(si->service, si, c->name, buf, cs_role_cmds);
 }
@@ -250,6 +250,24 @@ static int compare_template_nodes(mowgli_node_t *a, mowgli_node_t *b, void *opaq
 	return count_bits(ta->level) - count_bits(tb->level);
 }
 
+static template_t *find_template(mowgli_list_t *l, const char *key)
+{
+	mowgli_node_t *iter;
+
+	return_val_if_fail(l != NULL, NULL);
+	return_val_if_fail(key != NULL, NULL);
+
+	MOWGLI_ITER_FOREACH(iter, l->head)
+	{
+		template_t *t = iter->data;
+
+		if (!strcasecmp(t->name, key))
+			return t;
+	}
+
+	return NULL;
+}
+
 static int append_global_template(const char *key, void *data, void *privdata)
 {
 	template_t *t;
@@ -265,8 +283,11 @@ static int append_global_template(const char *key, void *data, void *privdata)
 			return 0;
 	}
 
+	if ((t = find_template(ti->l, key)) != NULL)
+		return 0;
+
 	t = smalloc(sizeof(template_t));
-	strlcpy(t->name, key, sizeof(t->name));
+	mowgli_strlcpy(t->name, key, sizeof(t->name));
 	t->level = get_template_flags(ti->mc, key);
 	mowgli_node_add(t, &t->node, ti->l);
 
@@ -302,19 +323,19 @@ static mowgli_list_t *build_template_list(mychan_t *mc)
 			r = strchr(q, ' ');
 			if (r != NULL && r < q)
 				break;
-			strlcpy(ss, q, sizeof ss);
+			mowgli_strlcpy(ss, q, sizeof ss);
 			if (r != NULL && r - q < (int)(sizeof ss - 1))
 			{
 				ss[r - q] = '\0';
 			}
 
-			strlcpy(flagname, p, sizeof flagname);
+			mowgli_strlcpy(flagname, p, sizeof flagname);
 			s = strchr(flagname, '=');
 			if (s != NULL)
 				*s = '\0';
 
 			t = smalloc(sizeof(template_t));
-			strlcpy(t->name, flagname, sizeof(t->name));
+			mowgli_strlcpy(t->name, flagname, sizeof(t->name));
 			t->level = flags_to_bitmask(ss, 0);
 			mowgli_node_add(t, &t->node, l);
 
@@ -388,9 +409,9 @@ static const char *get_template_name(mychan_t *mc, unsigned int level)
 	}
 
 	if (exact_t != NULL)
-		strlcpy(flagname, exact_t->name, sizeof flagname);
+		mowgli_strlcpy(flagname, exact_t->name, sizeof flagname);
 	else
-		strlcpy(flagname, "<Custom>", sizeof flagname);
+		mowgli_strlcpy(flagname, "<Custom>", sizeof flagname);
 
 	free_template_list(l);
 
@@ -412,6 +433,7 @@ static void update_role_entry(sourceinfo_t *si, mychan_t *mc, const char *role,
 	mowgli_node_t *n, *tn;
 	chanacs_t *ca;
 	int changes = 0;
+	hook_channel_acl_req_t req;
 
 	flagstr = bitmask_to_flags2(flags, 0);
 	oldflags = get_template_flags(mc, role);
@@ -421,7 +443,7 @@ static void update_role_entry(sourceinfo_t *si, mychan_t *mc, const char *role,
 	if (md != NULL)
 	{
 		p = md->value;
-		strlcpy(newstr, p, sizeof newstr);
+		mowgli_strlcpy(newstr, p, sizeof newstr);
 		while (p != NULL)
 		{
 			while (*p == ' ')
@@ -432,7 +454,7 @@ static void update_role_entry(sourceinfo_t *si, mychan_t *mc, const char *role,
 			r = strchr(q, ' ');
 			if (r != NULL && r < q)
 				break;
-			strlcpy(ss, q, sizeof ss);
+			mowgli_strlcpy(ss, q, sizeof ss);
 			if (r != NULL && r - q < (int)(sizeof ss - 1))
 			{
 				ss[r - q] = '\0';
@@ -444,12 +466,12 @@ static void update_role_entry(sourceinfo_t *si, mychan_t *mc, const char *role,
 				if (flags == 0)
 				{
 					if (p == md->value)
-						strlcpy(newstr, r != NULL ? r + 1 : "", sizeof newstr);
+						mowgli_strlcpy(newstr, r != NULL ? r + 1 : "", sizeof newstr);
 					else
 					{
 						/* otherwise, zap the space before it */
 						p--;
-						strlcpy(newstr + (p - md->value), r != NULL ? r : "", sizeof newstr - (p - md->value));
+						mowgli_strlcpy(newstr + (p - md->value), r != NULL ? r : "", sizeof newstr - (p - md->value));
 					}
 				}
 				else
@@ -493,13 +515,19 @@ static void update_role_entry(sourceinfo_t *si, mychan_t *mc, const char *role,
 			if (ca->level != oldflags)
 				continue;
 
-			/* don't change entries involving foundership status. */
-			if (oldflags & CA_FOUNDER)
+			/* don't change foundership status. */
+			if ((oldflags ^ flags) & CA_FOUNDER)
 				continue;
 
+			req.ca = ca;
+			req.oldlevel = ca->level;
+
 			changes++;
 			chanacs_modify_simple(ca, flags, ~flags);
 
+			req.newlevel = ca->level;
+
+			hook_call_channel_acl_change(&req);
 			chanacs_close(ca);
 		}
 	}
@@ -510,23 +538,14 @@ static void update_role_entry(sourceinfo_t *si, mychan_t *mc, const char *role,
 		command_success_nodata(si, _("%d access entries updated accordingly."), changes);
 }
 
-static unsigned int xflag_apply_batch(unsigned int in, int parc, char *parv[], unsigned int restrictflags)
+static unsigned int xflag_apply_batch(unsigned int in, int parc, char *parv[])
 {
 	unsigned int out;
 	int i;
 
 	out = in;
 	for (i = 0; i < parc; i++)
-	{
-#ifdef NOTYET
-		unsigned int flag;
-		flag = xflag_apply(0, parv[i]);
-		if (flag & restrictflags)
-			continue;
-#endif
-
 		out = xflag_apply(out, parv[i]);
-	}
 
 	return out;
 }
@@ -559,7 +578,7 @@ static void cs_cmd_access_list(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	if (!chanacs_source_has_flag(mc, si, CA_ACLVIEW))
+	if (!(mc->flags & MC_PUBACL) && !chanacs_source_has_flag(mc, si, CA_ACLVIEW))
 	{
 		if (has_priv(si, PRIV_CHAN_AUSPEX))
 			operoverride = true;
@@ -636,7 +655,7 @@ static void cs_cmd_access_info(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	if (!chanacs_source_has_flag(mc, si, CA_ACLVIEW))
+	if (!(mc->flags & MC_PUBACL) && !chanacs_source_has_flag(mc, si, CA_ACLVIEW))
 	{
 		if (has_priv(si, PRIV_CHAN_AUSPEX))
 			operoverride = true;
@@ -657,7 +676,7 @@ static void cs_cmd_access_info(sourceinfo_t *si, int parc, char *parv[])
 			return;
 		}
 		target = mt->name;
-		ca = chanacs_find(mc, mt, 0);
+		ca = chanacs_find_literal(mc, mt, 0);
 	}
 
 	if (ca == NULL)
@@ -675,7 +694,7 @@ static void cs_cmd_access_info(sourceinfo_t *si, int parc, char *parv[])
 	role = get_template_name(mc, ca->level);
 
 	tm = *localtime(&ca->tmodified);
-	strftime(strfbuf, sizeof(strfbuf) - 1,"%b %d %H:%M:%S %Y", &tm);
+	strftime(strfbuf, sizeof strfbuf, TIME_FORMAT, &tm);
 
 	command_success_nodata(si, _("Access for \2%s\2 in \2%s\2:"), target, channel);
 
@@ -713,6 +732,7 @@ static void cs_cmd_access_del(sourceinfo_t *si, int parc, char *parv[])
 	chanacs_t *ca;
 	myentity_t *mt;
 	mychan_t *mc;
+	hook_channel_acl_req_t req;
 	unsigned int restrictflags;
 	const char *channel = parv[0];
 	const char *target = parv[1];
@@ -744,7 +764,7 @@ static void cs_cmd_access_del(sourceinfo_t *si, int parc, char *parv[])
 	}
 
 	if (validhostmask(target))
-		ca = chanacs_open(mc, NULL, target, true);
+		ca = chanacs_open(mc, NULL, target, true, entity(si->smu));
 	else
 	{
 		if (mt == NULL)
@@ -753,7 +773,7 @@ static void cs_cmd_access_del(sourceinfo_t *si, int parc, char *parv[])
 			return;
 		}
 		target = mt->name;
-		ca = chanacs_open(mc, mt, NULL, true);
+		ca = chanacs_open(mc, mt, NULL, true, entity(si->smu));
 	}
 
 	if (ca->level == 0)
@@ -772,6 +792,10 @@ static void cs_cmd_access_del(sourceinfo_t *si, int parc, char *parv[])
 
 	role = get_template_name(mc, ca->level);
 
+	req.ca = ca;
+	req.oldlevel = ca->level;
+	req.newlevel = 0;
+
 	restrictflags = chanacs_source_flags(mc, si);
 	if (restrictflags & CA_FOUNDER)
 		restrictflags = ca_all;
@@ -787,14 +811,13 @@ static void cs_cmd_access_del(sourceinfo_t *si, int parc, char *parv[])
 			restrictflags = allow_flags(mc, restrictflags);
 	}
 
-	unsigned int i = 0;
-
-	if (!chanacs_modify(ca, &i, &ca->level, restrictflags))
+	if (!chanacs_modify(ca, &req.newlevel, &req.oldlevel, restrictflags))
 	{
 		command_fail(si, fault_noprivs, _("You may not remove \2%s\2 from the \2%s\2 role."), target, role);
 		return;
 	}
 
+	hook_call_channel_acl_change(&req);
 	chanacs_close(ca);
 
 	command_success_nodata(si, _("\2%s\2 was removed from the \2%s\2 role in \2%s\2."), target, role, channel);
@@ -813,8 +836,9 @@ static void cs_cmd_access_del(sourceinfo_t *si, int parc, char *parv[])
 static void cs_cmd_access_add(sourceinfo_t *si, int parc, char *parv[])
 {
 	chanacs_t *ca;
+	myentity_t *mt = NULL;
 	mychan_t *mc;
-	myentity_t *mt;
+	hook_channel_acl_req_t req;
 	unsigned int oldflags, restrictflags;
 	unsigned int newflags, addflags, removeflags;
 	const char *channel = parv[0];
@@ -845,7 +869,7 @@ static void cs_cmd_access_add(sourceinfo_t *si, int parc, char *parv[])
 	}
 
 	if (validhostmask(target))
-		ca = chanacs_open(mc, NULL, target, true);
+		ca = chanacs_open(mc, NULL, target, true, entity(si->smu));
 	else
 	{
 		if (!(mt = myentity_find_ext(target)))
@@ -854,7 +878,7 @@ static void cs_cmd_access_add(sourceinfo_t *si, int parc, char *parv[])
 			return;
 		}
 		target = mt->name;
-		ca = chanacs_open(mc, mt, NULL, true);
+		ca = chanacs_open(mc, mt, NULL, true, entity(si->smu));
 	}
 
 	if (ca->level == 0 && chanacs_is_table_full(ca))
@@ -871,6 +895,9 @@ static void cs_cmd_access_add(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
+	req.ca = ca;
+	req.oldlevel = ca->level;
+
 	newflags = get_template_flags(mc, role);
 	if (newflags == 0)
 	{
@@ -890,8 +917,38 @@ static void cs_cmd_access_add(sourceinfo_t *si, int parc, char *parv[])
 			return;
 		}
 
-		if (entity(si->smu) != mt)
+		if (mt != entity(si->smu))
 			restrictflags = allow_flags(mc, restrictflags);
+		else
+			restrictflags |= allow_flags(mc, restrictflags);
+	}
+
+	if (!(ca->level & CA_FOUNDER) && newflags & CA_FOUNDER)
+	{
+		if (mychan_num_founders(mc) >= chansvs.maxfounders)
+		{
+			command_fail(si, fault_noprivs, _("Only %d founders allowed per channel."), chansvs.maxfounders);
+			chanacs_close(ca);
+			return;
+		}
+		if (mt == NULL)
+		{
+			command_fail(si, fault_badparams, _("You may not set founder status on a hostmask."));
+			chanacs_close(ca);
+			return;
+		}
+		if (!myentity_can_register_channel(mt))
+		{
+			command_fail(si, fault_toomany, _("\2%s\2 has too many channels registered."), mt->name);
+			chanacs_close(ca);
+			return;
+		}
+		if (!myentity_allow_foundership(mt))
+		{
+			command_fail(si, fault_toomany, _("\2%s\2 cannot take foundership of a channel."), mt->name);
+			chanacs_close(ca);
+			return;
+		}
 	}
 
 	oldflags = ca->level;
@@ -905,6 +962,9 @@ static void cs_cmd_access_add(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
+	req.newlevel = newflags;
+
+	hook_call_channel_acl_change(&req);
 	chanacs_close(ca);
 
 	command_success_nodata(si, _("\2%s\2 was added with the \2%s\2 role in \2%s\2."), target, role, channel);
@@ -923,8 +983,9 @@ static void cs_cmd_access_add(sourceinfo_t *si, int parc, char *parv[])
 static void cs_cmd_access_set(sourceinfo_t *si, int parc, char *parv[])
 {
 	chanacs_t *ca;
-	myentity_t *mt;
+	myentity_t *mt = NULL;
 	mychan_t *mc;
+	hook_channel_acl_req_t req;
 	unsigned int oldflags, restrictflags;
 	unsigned int newflags, addflags, removeflags;
 	const char *channel = parv[0];
@@ -955,7 +1016,7 @@ static void cs_cmd_access_set(sourceinfo_t *si, int parc, char *parv[])
 	}
 
 	if (validhostmask(target))
-		ca = chanacs_open(mc, NULL, target, true);
+		ca = chanacs_open(mc, NULL, target, true, entity(si->smu));
 	else
 	{
 		if (!(mt = myentity_find_ext(target)))
@@ -964,7 +1025,7 @@ static void cs_cmd_access_set(sourceinfo_t *si, int parc, char *parv[])
 			return;
 		}
 		target = mt->name;
-		ca = chanacs_open(mc, mt, NULL, true);
+		ca = chanacs_open(mc, mt, NULL, true, entity(si->smu));
 	}
 
 	if (ca->level == 0 && chanacs_is_table_full(ca))
@@ -989,13 +1050,8 @@ static void cs_cmd_access_set(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	newflags = get_template_flags(mc, role);
-	if (newflags == 0)
-	{
-		chanacs_close(ca);
-		command_fail(si, fault_toomany, _("Role \2%s\2 does not exist."), role);
-		return;
-	}
+	req.ca = ca;
+	req.oldlevel = ca->level;
 
 	restrictflags = chanacs_source_flags(mc, si);
 	if (restrictflags & CA_FOUNDER)
@@ -1008,8 +1064,32 @@ static void cs_cmd_access_set(sourceinfo_t *si, int parc, char *parv[])
 			return;
 		}
 
-		if (entity(si->smu) != mt)
+		if (mt != entity(si->smu))
 			restrictflags = allow_flags(mc, restrictflags);
+		else
+			restrictflags |= allow_flags(mc, restrictflags);
+	}
+
+	if (!(ca->level & CA_FOUNDER) && newflags & CA_FOUNDER)
+	{
+		if (mychan_num_founders(mc) >= chansvs.maxfounders)
+		{
+			command_fail(si, fault_noprivs, _("Only %d founders allowed per channel."), chansvs.maxfounders);
+			chanacs_close(ca);
+			return;
+		}
+		if (mt == NULL)
+		{
+			command_fail(si, fault_badparams, _("You may not set founder status on a hostmask."));
+			chanacs_close(ca);
+			return;
+		}
+		if (!myentity_can_register_channel(mt))
+		{
+			command_fail(si, fault_toomany, _("\2%s\2 has too many channels registered."), mt->name);
+			chanacs_close(ca);
+			return;
+		}
 	}
 
 	oldflags = ca->level;
@@ -1023,6 +1103,9 @@ static void cs_cmd_access_set(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
+	req.newlevel = newflags;
+
+	hook_call_channel_acl_change(&req);
 	chanacs_close(ca);
 
 	command_success_nodata(si, _("\2%s\2 now has the \2%s\2 role in \2%s\2."), target, role, channel);
@@ -1112,6 +1195,10 @@ static void cs_cmd_role_add(sourceinfo_t *si, int parc, char *parv[])
 	}
 
 	restrictflags = chanacs_source_flags(mc, si);
+	if (restrictflags & CA_FOUNDER)
+		restrictflags = ca_all;
+	else
+		restrictflags = allow_flags(mc, restrictflags);
 	oldflags = get_template_flags(mc, role);
 
 	if (oldflags != 0)
@@ -1120,7 +1207,14 @@ static void cs_cmd_role_add(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	newflags = xflag_apply_batch(oldflags, parc - 2, parv + 2, restrictflags);
+	newflags = xflag_apply_batch(oldflags, parc - 2, parv + 2);
+	if (newflags & ~restrictflags)
+	{
+		unsigned int delta = newflags & ~restrictflags;
+
+		command_fail(si, fault_badparams, _("You do not have appropriate permissions to add flags: \2%s\2"), xflag_tostr(delta));
+		return;
+	}
 
 	if (newflags & CA_FOUNDER)
 		newflags |= CA_FLAGS;
@@ -1192,6 +1286,10 @@ static void cs_cmd_role_set(sourceinfo_t *si, int parc, char *parv[])
 	}
 
 	restrictflags = chanacs_source_flags(mc, si);
+	if (restrictflags & CA_FOUNDER)
+		restrictflags = ca_all;
+	else
+		restrictflags = allow_flags(mc, restrictflags);
 	oldflags = get_template_flags(mc, role);
 
 	if (oldflags == 0)
@@ -1200,7 +1298,20 @@ static void cs_cmd_role_set(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	newflags = xflag_apply_batch(oldflags, parc - 2, parv + 2, restrictflags);
+	newflags = xflag_apply_batch(oldflags, parc - 2, parv + 2);
+	if ((oldflags | newflags) & ~restrictflags)
+	{
+		unsigned int delta = (oldflags | newflags) & ~restrictflags;
+
+		command_fail(si, fault_badparams, _("You do not have appropriate permissions to set flags: \2%s\2"), xflag_tostr(delta));
+		return;
+	}
+
+	if ((oldflags ^ newflags) & CA_FOUNDER)
+	{
+		command_fail(si, fault_unimplemented, _("Adding or removing founder status from a role is not implemented."));
+		return;
+	}
 
 	if (newflags & CA_FOUNDER)
 		newflags |= CA_FLAGS;
@@ -1246,9 +1357,9 @@ static void cs_cmd_role_set(sourceinfo_t *si, int parc, char *parv[])
 static void cs_cmd_role_del(sourceinfo_t *si, int parc, char *parv[])
 {
 	mychan_t *mc;
-	unsigned int level;
 	const char *channel = parv[0];
 	const char *role = parv[1];
+	unsigned int oldflags, restrictflags;
 
 	mc = mychan_find(channel);
 	if (!mc)
@@ -1270,13 +1381,26 @@ static void cs_cmd_role_del(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	level = get_template_flags(mc, role);
-	if (level == 0)
+	restrictflags = chanacs_source_flags(mc, si);
+	if (restrictflags & CA_FOUNDER)
+		restrictflags = ca_all;
+	else
+		restrictflags = allow_flags(mc, restrictflags);
+	oldflags = get_template_flags(mc, role);
+	if (oldflags == 0)
 	{
 		command_fail(si, fault_toomany, _("Role \2%s\2 does not exist."), role);
 		return;
 	}
 
+	if (oldflags & ~restrictflags)
+	{
+		unsigned int delta = oldflags & ~restrictflags;
+
+		command_fail(si, fault_badparams, _("You do not have appropriate permissions to set flags: \2%s\2"), xflag_tostr(delta));
+		return;
+	}
+
 	command_success_nodata(si, _("Role \2%s\2 has been deleted."), role);
 	update_role_entry(si, mc, role, 0);
 }
