@@ -7,11 +7,13 @@
  */
 
 #include "atheme.h"
-#include "conf.h" /* XXX conf_ci_table */
+#include "chanserv.h"
+#include "template.h"
+#include <limits.h>
 
 DECLARE_MODULE_V1
 (
-	"chanserv/main", false, _modinit, _moddeinit,
+	"chanserv/main", true, _modinit, _moddeinit,
 	PACKAGE_STRING,
 	"Atheme Development Group <http://www.atheme.org>"
 );
@@ -19,13 +21,17 @@ DECLARE_MODULE_V1
 static void cs_join(hook_channel_joinpart_t *hdata);
 static void cs_part(hook_channel_joinpart_t *hdata);
 static void cs_register(hook_channel_req_t *mc);
+static void cs_succession(hook_channel_succession_req_t *data);
 static void cs_newchan(channel_t *c);
 static void cs_keeptopic_topicset(channel_t *c);
 static void cs_topiccheck(hook_channel_topic_check_t *data);
 static void cs_tschange(channel_t *c);
 static void cs_leave_empty(void *unused);
+static void cs_bounce_mode_change(hook_channel_mode_change_t *data);
 static void on_shutdown(void *unused);
 
+static mowgli_eventloop_timer_t *cs_leave_empty_timer = NULL;
+
 static void join_registered(bool all)
 {
 	mychan_t *mc;
@@ -94,14 +100,14 @@ static void chanserv(sourceinfo_t *si, int parc, char *parv[])
 	}
 
 	/* make a copy of the original for debugging */
-	strlcpy(orig, parv[parc - 1], BUFSIZE);
+	mowgli_strlcpy(orig, parv[parc - 1], BUFSIZE);
 
 	/* lets go through this to get the command */
 	cmd = strtok(parv[parc - 1], " ");
 
 	if (!cmd)
 		return;
-	if (*cmd == '\001')
+	if (*orig == '\001')
 	{
 		handle_ctcp_common(si, cmd, strtok(NULL, ""));
 		return;
@@ -115,22 +121,23 @@ static void chanserv(sourceinfo_t *si, int parc, char *parv[])
 		metadata_t *md = metadata_find(mc, "private:prefix");
 		const char *prefix = (md ? md->value : chansvs.trigger);
 
-		if (strlen(cmd) >= 2 && strchr(prefix, cmd[0]) && isalpha(*++cmd))
+		if (strlen(cmd) >= 2 && strchr(prefix, cmd[0]) && isalpha((unsigned char)*++cmd))
 		{
+			const char *realcmd = service_resolve_alias(si->service, NULL, cmd);
 
 			/* XXX not really nice to look up the command twice
 			 * -- jilles */
-			if (command_find(si->service->commands, service_resolve_alias(si->service, NULL, cmd)) == NULL)
+			if (command_find(si->service->commands, realcmd) == NULL)
 				return;
 			if (floodcheck(si->su, si->service->me))
 				return;
 			/* construct <channel> <args> */
-			strlcpy(newargs, parv[parc - 2], sizeof newargs);
+			mowgli_strlcpy(newargs, parv[parc - 2], sizeof newargs);
 			args = strtok(NULL, "");
 			if (args != NULL)
 			{
-				strlcat(newargs, " ", sizeof newargs);
-				strlcat(newargs, args, sizeof newargs);
+				mowgli_strlcat(newargs, " ", sizeof newargs);
+				mowgli_strlcat(newargs, args, sizeof newargs);
 			}
 			/* let the command know it's called as fantasy cmd */
 			si->c = mc->chan;
@@ -138,23 +145,26 @@ static void chanserv(sourceinfo_t *si, int parc, char *parv[])
 			 * (a little ugly but this way we can !set verbose)
 			 */
 			mc->flags |= MC_FORCEVERBOSE;
-			command_exec_split(si->service, si, cmd, newargs, si->service->commands);
+			command_exec_split(si->service, si, realcmd, newargs, si->service->commands);
 			mc->flags &= ~MC_FORCEVERBOSE;
 		}
-		else if (!ircncasecmp(cmd, chansvs.nick, strlen(chansvs.nick)) && !isalnum(cmd[strlen(chansvs.nick)]) && (cmd = strtok(NULL, "")) != NULL)
+		else if (!ircncasecmp(cmd, chansvs.nick, strlen(chansvs.nick)) && !isalnum((unsigned char)cmd[strlen(chansvs.nick)]) && (cmd = strtok(NULL, "")) != NULL)
 		{
+			const char *realcmd;
 			char *pptr;
 
-			strlcpy(newargs, parv[parc - 2], sizeof newargs);
+			mowgli_strlcpy(newargs, parv[parc - 2], sizeof newargs);
 			while (*cmd == ' ')
 				cmd++;
 			if ((pptr = strchr(cmd, ' ')) != NULL)
 			{
-				strlcat(newargs, pptr, sizeof newargs);
+				mowgli_strlcat(newargs, pptr, sizeof newargs);
 				*pptr = '\0';
 			}
 
-			if (command_find(si->service->commands, service_resolve_alias(si->service, NULL, cmd)) == NULL)
+			realcmd = service_resolve_alias(si->service, NULL, cmd);
+
+			if (command_find(si->service->commands, realcmd) == NULL)
 				return;
 			if (floodcheck(si->su, si->service->me))
 				return;
@@ -165,7 +175,7 @@ static void chanserv(sourceinfo_t *si, int parc, char *parv[])
 			 * (a little ugly but this way we can !set verbose)
 			 */
 			mc->flags |= MC_FORCEVERBOSE;
-			command_exec_split(si->service, si, cmd, newargs, si->service->commands);
+			command_exec_split(si->service, si, realcmd, newargs, si->service->commands);
 			mc->flags &= ~MC_FORCEVERBOSE;
 		}
 	}
@@ -184,33 +194,127 @@ static void chanserv_config_ready(void *unused)
 		join_registered(false); /* !config_options.leave_chans */
 }
 
+static int c_ci_vop(mowgli_config_file_entry_t *ce)
+{
+	if (ce->vardata == NULL)
+	{
+		conf_report_warning(ce, "no parameter for configuration option");
+		return 0;
+	}
+
+	set_global_template_flags("VOP", flags_to_bitmask(ce->vardata, 0));
+
+	return 0;
+}
+
+static int c_ci_hop(mowgli_config_file_entry_t *ce)
+{
+	if (ce->vardata == NULL)
+	{
+		conf_report_warning(ce, "no parameter for configuration option");
+		return 0;
+	}
+
+	set_global_template_flags("HOP", flags_to_bitmask(ce->vardata, 0));
+
+	return 0;
+}
+
+static int c_ci_aop(mowgli_config_file_entry_t *ce)
+{
+	if (ce->vardata == NULL)
+	{
+		conf_report_warning(ce, "no parameter for configuration option");
+		return 0;
+	}
+
+	set_global_template_flags("AOP", flags_to_bitmask(ce->vardata, 0));
+
+	return 0;
+}
+
+static int c_ci_sop(mowgli_config_file_entry_t *ce)
+{
+	if (ce->vardata == NULL)
+	{
+		conf_report_warning(ce, "no parameter for configuration option");
+		return 0;
+	}
+
+	set_global_template_flags("SOP", flags_to_bitmask(ce->vardata, 0));
+
+	return 0;
+}
+
+static int c_ci_templates(mowgli_config_file_entry_t *ce)
+{
+	mowgli_config_file_entry_t *flce;
+
+	MOWGLI_ITER_FOREACH(flce, ce->entries)
+	{
+		if (flce->vardata == NULL)
+		{
+			conf_report_warning(ce, "no parameter for configuration option");
+			return 0;
+		}
+
+		set_global_template_flags(flce->varname, flags_to_bitmask(flce->vardata, 0));
+	}
+
+	return 0;
+}
+
 void _modinit(module_t *m)
 {
 	hook_add_event("config_ready");
 	hook_add_config_ready(chanserv_config_ready);
 
-	chansvs.me = service_add("chanserv", chanserv, &conf_ci_table);
+	chansvs.me = service_add("chanserv", chanserv);
 
 	hook_add_event("channel_join");
 	hook_add_event("channel_part");
 	hook_add_event("channel_register");
+	hook_add_event("channel_succession");
 	hook_add_event("channel_add");
 	hook_add_event("channel_topic");
 	hook_add_event("channel_can_change_topic");
 	hook_add_event("channel_tschange");
+	hook_add_event("channel_mode_change");
+	hook_add_event("user_identify");
 	hook_add_event("shutdown");
 	hook_add_channel_join(cs_join);
 	hook_add_channel_part(cs_part);
 	hook_add_channel_register(cs_register);
+	hook_add_channel_succession(cs_succession);
 	hook_add_channel_add(cs_newchan);
 	hook_add_channel_topic(cs_keeptopic_topicset);
 	hook_add_channel_can_change_topic(cs_topiccheck);
 	hook_add_channel_tschange(cs_tschange);
+	hook_add_channel_mode_change(cs_bounce_mode_change);
 	hook_add_shutdown(on_shutdown);
-	event_add("cs_leave_empty", cs_leave_empty, NULL, 300);
+
+	cs_leave_empty_timer = mowgli_timer_add(base_eventloop, "cs_leave_empty", cs_leave_empty, NULL, 300);
+
+	/* chanserv{} block */
+	add_bool_conf_item("FANTASY", &chansvs.me->conf_table, 0, &chansvs.fantasy, false);
+	add_conf_item("VOP", &chansvs.me->conf_table, c_ci_vop);
+	add_conf_item("HOP", &chansvs.me->conf_table, c_ci_hop);
+	add_conf_item("AOP", &chansvs.me->conf_table, c_ci_aop);
+	add_conf_item("SOP", &chansvs.me->conf_table, c_ci_sop);
+	add_conf_item("TEMPLATES", &chansvs.me->conf_table, c_ci_templates);
+	add_bool_conf_item("CHANGETS", &chansvs.me->conf_table, 0, &chansvs.changets, false);
+	add_bool_conf_item("HIDE_XOP", &chansvs.me->conf_table, 0, &chansvs.hide_xop, false);
+	add_dupstr_conf_item("TRIGGER", &chansvs.me->conf_table, 0, &chansvs.trigger, "!");
+	add_duration_conf_item("EXPIRE", &chansvs.me->conf_table, 0, &chansvs.expiry, "d", 0);
+	add_uint_conf_item("MAXCHANS", &chansvs.me->conf_table, 0, &chansvs.maxchans, 1, INT_MAX, 5);
+	add_uint_conf_item("MAXCHANACS", &chansvs.me->conf_table, 0, &chansvs.maxchanacs, 0, INT_MAX, 0);
+	add_uint_conf_item("MAXFOUNDERS", &chansvs.me->conf_table, 0, &chansvs.maxfounders, 1, (512 - 60) / (9 + 2), 4); /* fit on a line */
+	add_dupstr_conf_item("FOUNDER_FLAGS", &chansvs.me->conf_table, 0, &chansvs.founder_flags, NULL);
+	add_dupstr_conf_item("DEFTEMPLATES", &chansvs.me->conf_table, 0, &chansvs.deftemplates, NULL);
+	add_duration_conf_item("AKICK_TIME", &chansvs.me->conf_table, 0, &chansvs.akick_time, "m", 0);
 }
 
-void _moddeinit(void)
+void _moddeinit(module_unload_intent_t intent)
 {
 	if (chansvs.me)
 	{
@@ -226,12 +330,15 @@ void _moddeinit(void)
 	hook_del_channel_join(cs_join);
 	hook_del_channel_part(cs_part);
 	hook_del_channel_register(cs_register);
+	hook_del_channel_succession(cs_succession);
 	hook_del_channel_add(cs_newchan);
 	hook_del_channel_topic(cs_keeptopic_topicset);
 	hook_del_channel_can_change_topic(cs_topiccheck);
 	hook_del_channel_tschange(cs_tschange);
+	hook_del_channel_mode_change(cs_bounce_mode_change);
 	hook_del_shutdown(on_shutdown);
-	event_delete(cs_leave_empty, NULL);
+
+	mowgli_timer_destroy(base_eventloop, cs_leave_empty_timer);
 }
 
 static void cs_join(hook_channel_joinpart_t *hdata)
@@ -276,10 +383,10 @@ static void cs_join(hook_channel_joinpart_t *hdata)
 	if ((mc->flags & MC_RESTRICTED) && !(flags & CA_ALLPRIVS) && !has_priv_user(u, PRIV_JOIN_STAFFONLY))
 	{
 		/* Stay on channel if this would empty it -- jilles */
-		if (chan->nummembers <= (mc->flags & MC_GUARD ? 2 : 1))
+		if (chan->nummembers - chan->numsvcmembers == 1)
 		{
 			mc->flags |= MC_INHABIT;
-			if (!(mc->flags & MC_GUARD))
+			if (chan->numsvcmembers == 0)
 				join(chan->name, chansvs.nick);
 		}
 		if (mc->mlock_on & CMODE_INVITE || chan->modes & CMODE_INVITE)
@@ -299,13 +406,13 @@ static void cs_join(hook_channel_joinpart_t *hdata)
 		return;
 	}
 
-	if (flags & CA_AKICK && !(flags & CA_REMOVE))
+	if (flags & CA_AKICK && !(flags & CA_EXEMPT))
 	{
 		/* Stay on channel if this would empty it -- jilles */
-		if (chan->nummembers <= (mc->flags & MC_GUARD ? 2 : 1))
+		if (chan->nummembers - chan->numsvcmembers == 1)
 		{
 			mc->flags |= MC_INHABIT;
-			if (!(mc->flags & MC_GUARD))
+			if (chan->numsvcmembers == 0)
 				join(chan->name, chansvs.nick);
 		}
 		/* use a user-given ban mask if possible -- jilles */
@@ -314,12 +421,9 @@ static void cs_join(hook_channel_joinpart_t *hdata)
 		{
 			if (chanban_find(chan, ca2->host, 'b') == NULL)
 			{
-				char str[512];
-
 				chanban_add(chan, ca2->host, 'b');
-				snprintf(str, sizeof str, "+b %s", ca2->host);
-				/* ban immediately */
-				mode_sts(chansvs.nick, chan, str);
+				modestack_mode_param(chansvs.nick, chan, MTYPE_ADD, 'b', ca2->host);
+				modestack_flush_channel(chan);
 			}
 		}
 		else
@@ -367,13 +471,13 @@ static void cs_join(hook_channel_joinpart_t *hdata)
 	 */
 	if (mc->mlock_on & CMODE_INVITE && !(flags & CA_INVITE) &&
 			(!me.bursting || mc->flags & MC_RECREATED) &&
-			(!(u->server->flags & SF_EOB) || (chan->nummembers <= 2 && (chan->nummembers <= 1 || chanuser_find(chan, chansvs.me->me)))) &&
+			(!(u->server->flags & SF_EOB) || (chan->nummembers - chan->numsvcmembers == 1)) &&
 			(!ircd->invex_mchar || !next_matching_ban(chan, u, ircd->invex_mchar, chan->bans.head)))
 	{
-		if (chan->nummembers <= (mc->flags & MC_GUARD ? 2 : 1))
+		if (chan->nummembers - chan->numsvcmembers == 1)
 		{
 			mc->flags |= MC_INHABIT;
-			if (!(mc->flags & MC_GUARD))
+			if (chan->numsvcmembers == 0)
 				join(chan->name, chansvs.nick);
 		}
 		if (!(chan->modes & CMODE_INVITE))
@@ -390,10 +494,10 @@ static void cs_join(hook_channel_joinpart_t *hdata)
 	 * triggering autocycle-for-ops scripts and immediately
 	 * destroying channels with kick on split riding.
 	 */
-	if (mc->flags & MC_INHABIT && chan->nummembers >= 3)
+	if (mc->flags & MC_INHABIT && chan->nummembers - chan->numsvcmembers >= 2)
 	{
 		mc->flags &= ~MC_INHABIT;
-		if (!(mc->flags & MC_GUARD) && (!config_options.chan || irccasecmp(chan->name, config_options.chan)) && !(chan->flags & CHAN_LOG) && chanuser_find(chan, chansvs.me->me))
+		if (!(mc->flags & MC_GUARD) && !(chan->flags & CHAN_LOG) && chanuser_find(chan, chansvs.me->me))
 			part(chan->name, chansvs.nick);
 	}
 
@@ -471,9 +575,13 @@ static void cs_join(hook_channel_joinpart_t *hdata)
 		}
 	}
 
-	if (u->server->flags & SF_EOB && (md = metadata_find(mc, "private:entrymsg"))) {
-		if (!u->myuser || !(u->myuser->flags & MU_NOGREET))
-			notice(chansvs.nick, cu->user->nick, "[%s] %s", mc->name, md->value);
+	if (u->server->flags & SF_EOB && (md = metadata_find(mc, "private:entrymsg")))
+	{
+		if (metadata_find(mc, "private:botserv:bot-assigned") == NULL)
+		{
+			if (!u->myuser || !(u->myuser->flags & MU_NOGREET))
+				notice(chansvs.nick, cu->user->nick, "[%s] %s", mc->name, md->value);
+		}
 	}
 
 	if (u->server->flags & SF_EOB && (md = metadata_find(mc, "url")))
@@ -500,27 +608,44 @@ static void cs_part(hook_channel_joinpart_t *hdata)
 	if (CURRTIME - mc->used >= 3600)
 		if (chanacs_user_flags(mc, cu->user) & CA_USEDUPDATE)
 			mc->used = CURRTIME;
+
 	/*
 	 * When channel_part is fired, we haven't yet removed the
 	 * user from the room. So, the channel will have two members
 	 * if ChanServ is joining channels: the triggering user and
 	 * itself.
 	 *
-	 * Do not part if we're enforcing an akick/close in an otherwise
-	 * empty channel (MC_INHABIT). -- jilles
+	 * This if block was utter nonsense.  Refactor it into multiple
+	 * branches for better clarity and debugging ability. --nenolod
 	 */
-	if ((mc->flags & MC_GUARD)
-		&& config_options.leave_chans
-		&& !(mc->flags & MC_INHABIT)
-		&& (cu->chan->nummembers <= 2)
-		&& !is_internal_client(cu->user))
-		part(cu->chan->name, chansvs.nick);
+
+	/* we're not parting if we've been told to never part. */
+	if (!config_options.leave_chans)
+		return;
+
+	/* we're not parting if the channel has more than one person on it */
+	if (cu->chan->nummembers - cu->chan->numsvcmembers > 1)
+		return;
+
+	/* internal clients parting a channel shouldn't cause chanserv to leave. */
+	if (is_internal_client(cu->user))
+		return;
+
+	/* if we're enforcing an akick, we're MC_INHABIT.  do not part. */
+	if (mc->flags & MC_INHABIT)
+	{
+		slog(LG_DEBUG, "cs_part(): not leaving channel %s due to MC_INHABIT flag", mc->name);
+		return;
+	}
+
+	/* seems we've met all conditions to be parted from a channel. */
+	part(cu->chan->name, chansvs.nick);
 }
 
 static user_t *get_changets_user(mychan_t *mc)
 {
 	metadata_t *md;
-	
+
 	return_val_if_fail(mc != NULL, chansvs.me->me);
 
 	md = metadata_find(mc, "private:botserv:bot-assigned");
@@ -549,10 +674,16 @@ static void cs_register(hook_channel_req_t *hdata)
 			return;
 
 		mlock_sts(mc->chan);
+		topiclock_sts(mc->chan);
 		check_modes(mc, true);
 	}
 }
 
+static void cs_succession(hook_channel_succession_req_t *data)
+{
+	chanacs_change_simple(data->mc, entity(data->mu), NULL, custom_founder_check(), 0, NULL);
+}
+
 /* Called on every set of a topic, after updating our internal structures */
 static void cs_keeptopic_topicset(channel_t *c)
 {
@@ -646,13 +777,13 @@ static void cs_newchan(channel_t *c)
 	if (!(mc = mychan_find(c->name)))
 		return;
 
+	/* set channel_t.mychan */
+	c->mychan = mc;
+
 	/* schedule a mode lock check when we know the current modes
 	 * -- jilles */
 	mc->flags |= MC_MLOCK_CHECK;
 
-	/* ...and send the MLOCK to the ircd to do its bit. */
-	mlock_sts(c);
-
 	md = metadata_find(mc, "private:channelts");
 	if (md != NULL)
 		channelts = atol(md->value);
@@ -688,6 +819,8 @@ static void cs_newchan(channel_t *c)
 		metadata_add(mc, "private:channelts", str);
 	}
 	else if (!(MC_TOPICLOCK & mc->flags) && MOWGLI_LIST_LENGTH(&c->members) == 0)
+	{
+		mlock_sts(c);
 		/* Same channel, let's assume ircd has kept topic.
 		 * However, if topiclock is enabled, we must change it back
 		 * regardless.
@@ -695,6 +828,10 @@ static void cs_newchan(channel_t *c)
 		 * being created by a service and we must restore.
 		 * -- jilles */
 		return;
+	}
+
+	mlock_sts(c);
+	topiclock_sts(c);
 
 	if (!(MC_KEEPTOPIC & mc->flags))
 		return;
@@ -736,6 +873,7 @@ static void cs_tschange(channel_t *c)
 
 	/* reset the mlock if needed */
 	mlock_sts(c);
+	topiclock_sts(c);
 }
 
 static void on_shutdown(void *unused)
@@ -755,14 +893,13 @@ static void cs_leave_empty(void *unused)
 		if (!(mc->flags & MC_INHABIT))
 			continue;
 		/* If there is only one user, stay indefinitely. */
-		if (mc->chan != NULL && mc->chan->nummembers == 2)
+		if (mc->chan != NULL && mc->chan->nummembers - mc->chan->numsvcmembers == 1)
 			continue;
 		mc->flags &= ~MC_INHABIT;
 		if (mc->chan != NULL &&
-				(!config_options.chan || irccasecmp(mc->name, config_options.chan)) &&
 				!(mc->chan->flags & CHAN_LOG) &&
 				(!(mc->flags & MC_GUARD) ||
-				 (config_options.leave_chans && mc->chan->nummembers == 1) ||
+				 (config_options.leave_chans && mc->chan->nummembers == mc->chan->numsvcmembers) ||
 				 metadata_find(mc, "private:close:closer")) &&
 				chanuser_find(mc->chan, chansvs.me->me))
 		{
@@ -772,6 +909,42 @@ static void cs_leave_empty(void *unused)
 	}
 }
 
+static void cs_bounce_mode_change(hook_channel_mode_change_t *data)
+{
+	mychan_t *mc;
+	chanuser_t *cu;
+	channel_t *chan;
+
+	/* if we have SECURE mode enabled, then we want to bounce any changes */
+	cu = data->cu;
+	chan = cu->chan;
+	mc = chan->mychan;
+
+	if (mc == NULL || (mc != NULL && !(mc->flags & MC_SECURE)))
+		return;
+
+	if (ircd->uses_owner && data->mchar == ircd->owner_mchar[1] && !(chanacs_user_flags(mc, cu->user) & (CA_USEOWNER)))
+	{
+		modestack_mode_param(chansvs.nick, chan, MTYPE_DEL, ircd->owner_mchar[1], CLIENT_NAME(cu->user));
+		cu->modes &= ~data->mvalue;
+	}
+	else if (ircd->uses_protect && data->mchar == ircd->protect_mchar[1] && !(chanacs_user_flags(mc, cu->user) & (CA_USEPROTECT)))
+	{
+		modestack_mode_param(chansvs.nick, chan, MTYPE_DEL, ircd->protect_mchar[1], CLIENT_NAME(cu->user));
+		cu->modes &= ~data->mvalue;
+	}
+	else if (data->mchar == 'o' && !(chanacs_user_flags(mc, cu->user) & (CA_OP | CA_AUTOOP)))
+	{
+		modestack_mode_param(chansvs.nick, chan, MTYPE_DEL, 'o', CLIENT_NAME(cu->user));
+		cu->modes &= ~data->mvalue;
+	}
+	else if (ircd->uses_halfops && data->mchar == ircd->halfops_mchar[1] && !(chanacs_user_flags(mc, cu->user) & (CA_HALFOP | CA_AUTOHALFOP)))
+	{
+		modestack_mode_param(chansvs.nick, chan, MTYPE_DEL, ircd->halfops_mchar[1], CLIENT_NAME(cu->user));
+		cu->modes &= ~data->mvalue;
+	}
+}
+
 /* vim:cinoptions=>s,e0,n0,f0,{0,}0,^0,=s,ps,t0,c3,+s,(2s,us,)20,*30,gs,hs
  * vim:ts=8
  * vim:sw=8
