@@ -1,4 +1,5 @@
 #include "atheme.h"
+#include "serno.h"
 
 DECLARE_MODULE_V1
 (
@@ -17,7 +18,7 @@ void _modinit(module_t *m)
         service_named_bind_command("chanserv", &cs_version);
 }
 
-void _moddeinit()
+void _moddeinit(module_unload_intent_t intent)
 {
 	service_named_unbind_command("chanserv", &cs_version);
 }
@@ -25,7 +26,7 @@ void _moddeinit()
 static void cs_cmd_version(sourceinfo_t *si, int parc, char *parv[])
 {
         char buf[BUFSIZE];
-        snprintf(buf, BUFSIZE, "Atheme %s [%s] #%s", version, revision, generation);
+        snprintf(buf, BUFSIZE, "%s [%s]", PACKAGE_STRING, SERNO);
         command_success_string(si, buf, "%s", buf);
         return;
 }
