@@ -1,8 +1,8 @@
-#include "atheme.h"
+#include "atheme-compat.h"
 
 DECLARE_MODULE_V1
 (
-	"operserv/tabletest", false, _modinit, _moddeinit,
+	"contrib/os_tabletest", false, _modinit, _moddeinit,
 	PACKAGE_STRING,
 	"Atheme Development Group <http://www.atheme.org>"
 );
@@ -16,7 +16,7 @@ void _modinit(module_t *m)
         service_named_bind_command("operserv", &os_tabletest);
 }
 
-void _moddeinit()
+void _moddeinit(module_unload_intent_t intent)
 {
 	service_named_unbind_command("operserv", &os_tabletest);
 }
