@@ -5,12 +5,12 @@
  * Looks for users and performs actions on them.
  */
 
-#include "atheme.h"
+#include "atheme-compat.h"
 #include <limits.h>
 
 DECLARE_MODULE_V1
 (
-	"operserv/trace", false, _modinit, _moddeinit,
+	"contrib/os_trace", false, _modinit, _moddeinit,
 	"Copyright (c) 2010 William Pitcock <nenolod@atheme.org>",
 	"Atheme Development Group <http://www.atheme.org>"
 );
@@ -100,7 +100,7 @@ char *reason_extract(char **args)
 			(*args)++;
 			break;
 		}
-		
+
 		(*args)++;
 	}
 
@@ -132,7 +132,7 @@ static void *trace_regexp_prepare(char **args)
 	domain = scalloc(sizeof(trace_query_regexp_domain_t), 1);
 	domain->pattern = regex_extract(*args, &(*args), &domain->flags);
 	domain->regex = regex_create(domain->pattern, domain->flags);
-	
+
 	return domain;
 }
 
@@ -212,6 +212,57 @@ static void trace_server_cleanup(void *q)
 
 trace_query_constructor_t trace_server = { trace_server_prepare, trace_server_exec, trace_server_cleanup };
 
+typedef struct {
+	trace_query_domain_t domain;
+	char *pattern;
+} trace_query_glob_domain_t;
+
+static void *trace_glob_prepare(char **args)
+{
+	char *pattern;
+	trace_query_glob_domain_t *domain;
+
+	return_val_if_fail(args != NULL, NULL);
+	return_val_if_fail(*args != NULL, NULL);
+
+	/* split out the next space */
+	pattern = strtok(*args, " ");
+
+	domain = scalloc(sizeof(trace_query_glob_domain_t), 1);
+	domain->pattern = sstrdup(pattern);
+
+	*args = strtok(NULL, "");
+
+	return domain;
+}
+
+static bool trace_glob_exec(user_t *u, void *q)
+{
+	char usermask[512];
+	trace_query_glob_domain_t *domain = (trace_query_glob_domain_t *) q;
+
+	return_val_if_fail(domain != NULL, false);
+	return_val_if_fail(u != NULL, false);
+
+	if (domain->pattern == NULL)
+		return false;
+
+	snprintf(usermask, 512, "%s!%s@%s", u->nick, u->user, u->host);
+
+	return !match(domain->pattern, usermask);
+}
+
+static void trace_glob_cleanup(void *q)
+{
+	trace_query_glob_domain_t *domain = (trace_query_glob_domain_t *) q;
+
+	return_if_fail(domain != NULL);
+
+	free(domain);
+}
+
+trace_query_constructor_t trace_glob = { trace_glob_prepare, trace_glob_exec, trace_glob_cleanup };
+
 typedef struct {
 	trace_query_domain_t domain;
 	channel_t *channel;
@@ -244,6 +295,9 @@ static bool trace_channel_exec(user_t *u, void *q)
 	return_val_if_fail(domain != NULL, false);
 	return_val_if_fail(u != NULL, false);
 
+	if (domain->channel == NULL)
+		return false;
+
 	return (chanuser_find(domain->channel, u) != NULL);
 }
 
@@ -525,7 +579,7 @@ static void trace_kill_exec(user_t *u, trace_action_t *act)
 		return;
 	if (u->myuser && is_soper(u->myuser))
 		return;
-	if ((svs = service_find("operserv")) != NULL)
+	if ((svs = service_find("operserv")) == NULL)
 		return;
 
 	act->matched = true;
@@ -568,7 +622,7 @@ static trace_action_t *trace_akill_prepare(sourceinfo_t *si, char **args)
 	/* Extract a token, but only if there's one to remove.
 	 * Otherwise, this would clip a word off the reason. */
 	token = 0;
-	s = *args; 
+	s = *args;
 	if (!strncasecmp(s, "!T", 2) || !strncasecmp(s, "!P", 2))
 	{
 		if (s[2] == ' ')
@@ -592,7 +646,7 @@ static trace_action_t *trace_akill_prepare(sourceinfo_t *si, char **args)
 			return NULL;
 
 		duration = (atol(s) * 60);
-		while (isdigit(*s))
+		while (isdigit((unsigned char)*s))
 		s++;
 		if (*s == 'h' || *s == 'H')
 			duration *= 60;
@@ -604,17 +658,17 @@ static trace_action_t *trace_akill_prepare(sourceinfo_t *si, char **args)
 			;
 		else
 			duration = 0;
-		
+
 		if (duration == 0)
 			return NULL;
 	}
 	else if (token == 'p')
 		duration = 0;
-	
+
 	reason = reason_extract(args);
 	if (reason == NULL)
 		return NULL;
-	
+
 	a = scalloc(sizeof(trace_action_akill_t), 1);
 	trace_action_init(&a->base, si);
 	a->duration = duration;
@@ -626,7 +680,7 @@ static trace_action_t *trace_akill_prepare(sourceinfo_t *si, char **args)
 
 static void trace_akill_exec(user_t *u, trace_action_t *act)
 {
-	char *kuser, *khost;
+	const char *kuser, *khost;
 	trace_action_akill_t *a = (trace_action_akill_t *) act;
 
 	return_if_fail(u != NULL);
@@ -727,6 +781,7 @@ void _modinit(module_t *m)
 	trace_cmdtree = mowgli_patricia_create(strcasecanon);
 	mowgli_patricia_add(trace_cmdtree, "REGEXP", &trace_regexp);
 	mowgli_patricia_add(trace_cmdtree, "SERVER", &trace_server);
+	mowgli_patricia_add(trace_cmdtree, "GLOB", &trace_glob);
 	mowgli_patricia_add(trace_cmdtree, "CHANNEL", &trace_channel);
 	mowgli_patricia_add(trace_cmdtree, "NICKAGE", &trace_nickage);
 	mowgli_patricia_add(trace_cmdtree, "NUMCHAN", &trace_numchan);
@@ -739,7 +794,7 @@ void _modinit(module_t *m)
 	mowgli_patricia_add(trace_acttree, "COUNT", &trace_count);
 }
 
-void _moddeinit(void)
+void _moddeinit(module_unload_intent_t intent)
 {
 	mowgli_patricia_destroy(trace_cmdtree, NULL, NULL);
 
@@ -788,7 +843,7 @@ static void os_cmd_trace(sourceinfo_t *si, int parc, char *parv[])
 	MOWGLI_ITER_FOREACH_SAFE(n, tn, crit.head)
 	{
 		trace_query_domain_t *q = (trace_query_domain_t *) n->data;
-		q->cons->cleanup(q);		
+		q->cons->cleanup(q);
 	}
 	actcons->cleanup(act, succeeded);
 
