@@ -16,14 +16,13 @@ DECLARE_MODULE_V1
 );
 
 service_t *gs;
-mowgli_list_t gs_conftable;
 
 void _modinit(module_t *m)
 {
-	gs = service_add("gameserv", NULL, &gs_conftable);
+	gs = service_add("gameserv", NULL);
 }
 
-void _moddeinit(void)
+void _moddeinit(module_unload_intent_t intent)
 {
         if (gs != NULL)
                 service_delete(gs);
