@@ -50,6 +50,44 @@ command_t ns_regain = { "REGAIN", N_("Regain usage of a nickname."), AC_NONE, 2,
 
 mowgli_patricia_t **ns_set_cmdtree;
 
+static mowgli_eventloop_timer_t *enforce_timeout_check_timer = NULL;
+static mowgli_eventloop_timer_t *enforce_remove_enforcers_timer = NULL;
+
+/* logs a released nickname out */
+static bool log_enforce_victim_out(user_t *u, myuser_t *mu)
+{
+	mynick_t *mn;
+	mowgli_node_t *n, *tn;
+
+	return_val_if_fail(u != NULL, false);
+
+	if (u->myuser == NULL || u->myuser != mu)
+		return false;
+
+	u->myuser->lastlogin = CURRTIME;
+
+	if ((mn = mynick_find(u->nick)) != NULL)
+		mn->lastseen = CURRTIME;
+
+	if (!ircd_on_logout(u, entity(u->myuser)->name))
+	{
+		MOWGLI_ITER_FOREACH_SAFE(n, tn, u->myuser->logins.head)
+		{
+			if (n->data == u)
+			{
+				mowgli_node_delete(n, &u->myuser->logins);
+				mowgli_node_free(n);
+				break;
+			}
+		}
+
+		u->myuser = NULL;
+		return false;
+	}
+
+	return true;
+}
+
 /* sends an FNC for the given user */
 static void guest_nickname(user_t *u)
 {
@@ -58,7 +96,7 @@ static void guest_nickname(user_t *u)
 
 	/* Generate a new guest nickname and check if it already exists
 	 * This will try to generate a new nickname 30 different times
-	 * if nicks are in use. If it runs into 30 nicks in use, maybe 
+	 * if nicks are in use. If it runs into 30 nicks in use, maybe
 	 * you shouldn't use this module. */
 	for (tries = 0; tries < 30; tries++)
 	{
@@ -97,12 +135,6 @@ static void ns_cmd_set_enforce(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	if (!si->smu)
-	{
-		command_fail(si, fault_noprivs, _("You are not logged in."));
-		return;
-	}
-
 	if (strcasecmp(setting, "ON") == 0)
 	{
 		if (metadata_find(si->smu, "private:doenforce"))
@@ -139,8 +171,8 @@ static void ns_cmd_set_enforce(sourceinfo_t *si, int parc, char *parv[])
 static void ns_cmd_release(sourceinfo_t *si, int parc, char *parv[])
 {
 	mynick_t *mn;
-	char *target = parv[0];
-	char *password = parv[1];
+	const char *target = parv[0];
+	const char *password = parv[1];
 	user_t *u;
 	mowgli_node_t *n, *tn;
 	enforce_timeout_t *timeout;
@@ -164,19 +196,31 @@ static void ns_cmd_release(sourceinfo_t *si, int parc, char *parv[])
 
 	u = user_find_named(target);
 	mn = mynick_find(target);
-	
+
 	if (!mn)
 	{
 		command_fail(si, fault_nosuch_target, _("\2%s\2 is not a registered nickname."), target);
 		return;
 	}
-	
+
 	/* The != NULL check is required to make releasing an enforcer via xmlrpc work */
 	if (u != NULL && u == si->su)
 	{
 		command_fail(si, fault_noprivs, _("You cannot RELEASE yourself."));
 		return;
 	}
+	if (password && metadata_find(mn->owner, "private:freeze:freezer"))
+	{
+		command_fail(si, fault_authfail, "You cannot release \2%s\2 because the account has been frozen.", target);
+		logcommand(si, CMDLOG_DO, "failed RELEASE \2%s\2 (frozen)", target);
+		return;
+	}
+	if (password && (mn->owner->flags & MU_NOPASSWORD))
+	{
+		command_fail(si, fault_authfail, _("Password authentication is disabled for this account."));
+		logcommand(si, CMDLOG_DO, "failed RELEASE \2%s\2 (password authentication disabled)", target);
+		return;
+	}
 	if ((si->smu == mn->owner) || verify_password(mn->owner, password))
 	{
 		/* if this (nick, host) is waiting to be enforced, remove it */
@@ -206,13 +250,17 @@ static void ns_cmd_release(sourceinfo_t *si, int parc, char *parv[])
 		else
 		{
 			notice(nicksvs.nick, target, "%s has released your nickname.", get_source_mask(si));
-			guest_nickname(u);
-			if (ircd->flags & IRCD_HOLDNICK)
-				holdnick_sts(nicksvs.me->me, 60 + arc4random() % 60, u->nick, mn->owner);
-			else
-				u->flags |= UF_DOENFORCE;
-			command_success_nodata(si, _("%s has been released."), target);
-			logcommand(si, CMDLOG_DO, "RELEASE: \2%s!%s@%s\2", u->nick, u->user, u->vhost);
+
+			if (!log_enforce_victim_out(u, mn->owner))
+			{
+				guest_nickname(u);
+				if (ircd->flags & IRCD_HOLDNICK)
+					holdnick_sts(nicksvs.me->me, 60 + arc4random() % 60, u->nick, mn->owner);
+				else
+					u->flags |= UF_DOENFORCE;
+				command_success_nodata(si, _("\2%s\2 has been released."), target);
+				logcommand(si, CMDLOG_DO, "RELEASE: \2%s!%s@%s\2", u->nick, u->user, u->vhost);
+			}
 		}
 		return;
 	}
@@ -233,11 +281,12 @@ static void ns_cmd_release(sourceinfo_t *si, int parc, char *parv[])
 static void ns_cmd_regain(sourceinfo_t *si, int parc, char *parv[])
 {
 	mynick_t *mn;
-	char *target = parv[0];
-	char *password = parv[1];
+	const char *target = parv[0];
+	const char *password = parv[1];
 	user_t *u;
 	mowgli_node_t *n, *tn;
 	enforce_timeout_t *timeout;
+	char lau[BUFSIZE];
 
 	/* Absolutely do not do anything like this if nicks
 	 * are not considered owned */
@@ -258,20 +307,38 @@ static void ns_cmd_regain(sourceinfo_t *si, int parc, char *parv[])
 
 	u = user_find_named(target);
 	mn = mynick_find(target);
-	
+
 	if (!mn)
 	{
 		command_fail(si, fault_nosuch_target, _("\2%s\2 is not a registered nickname."), target);
 		return;
 	}
-	
+
 	if (u == si->su)
 	{
 		command_fail(si, fault_noprivs, _("You cannot REGAIN yourself."));
 		return;
 	}
+	if (password && metadata_find(mn->owner, "private:freeze:freezer"))
+	{
+		command_fail(si, fault_authfail, "You cannot regain \2%s\2 because the account has been frozen.", target);
+		logcommand(si, CMDLOG_DO, "failed REGAIN \2%s\2 (frozen)", target);
+		return;
+	}
+	if (password && (mn->owner->flags & MU_NOPASSWORD))
+	{
+		command_fail(si, fault_authfail, _("Password authentication is disabled for this account."));
+		logcommand(si, CMDLOG_DO, "failed REGAIN \2%s\2 (password authentication disabled)", target);
+		return;
+	}
 	if ((si->smu == mn->owner) || verify_password(mn->owner, password))
 	{
+		if (si->su != NULL && (user_is_channel_banned(si->su, 'b') || user_is_channel_banned(si->su, 'q')))
+		{
+			command_fail(si, fault_noprivs, _("You can not regain your nickname while banned or quieted on a channel."));
+			return;
+		}
+
 		/* if this (nick, host) is waiting to be enforced, remove it */
 		if (si->su != NULL)
 		{
@@ -300,11 +367,63 @@ static void ns_cmd_regain(sourceinfo_t *si, int parc, char *parv[])
 		else
 		{
 			notice(nicksvs.nick, target, "\2%s\2 has regained your nickname.", get_source_mask(si));
-			guest_nickname(u);
+			if (!log_enforce_victim_out(u, mn->owner))
+			{
+				guest_nickname(u);
+				if (ircd->flags & IRCD_HOLDNICK)
+					holdnick_sts(nicksvs.me->me, 60 + arc4random() % 60, u->nick, mn->owner);
+				else
+					u->flags |= UF_DOENFORCE;
+				command_success_nodata(si, _("\2%s\2 has been regained."), target);
+				logcommand(si, CMDLOG_DO, "REGAIN: \2%s!%s@%s\2", u->nick, u->user, u->vhost);
+			}
 			fnc_sts(nicksvs.me->me, si->su, target, FNC_FORCE);
-			command_success_nodata(si, _("\2%s\2 has been regained."), target);
-			logcommand(si, CMDLOG_DO, "REGAIN: \2%s!%s@%s\2", u->nick, u->user, u->vhost);
 		}
+
+		if (MOWGLI_LIST_LENGTH(&mn->owner->logins) >= me.maxlogins)
+		{
+			command_fail(si, fault_toomany, _("You were not logged in."));
+			command_fail(si, fault_toomany, _("There are already \2%zu\2 sessions logged in to \2%s\2 (maximum allowed: %u)."), MOWGLI_LIST_LENGTH(&mn->owner->logins), entity(mn->owner)->name, me.maxlogins);
+
+			lau[0] = '\0';
+			MOWGLI_ITER_FOREACH(n, mn->owner->logins.head)
+			{
+				if (lau[0] != '\0')
+					mowgli_strlcat(lau, ", ", sizeof lau);
+				mowgli_strlcat(lau, ((user_t *)n->data)->nick, sizeof lau);
+			}
+			command_fail(si, fault_toomany, _("Logged in nicks are: %s"), lau);
+
+			return;
+		}
+
+		/* identify them to the target nick's account if they aren't yet */
+		if (si->smu != mn->owner)
+		{
+			/* if they are identified to another account, nuke their session first */
+			if (si->smu != NULL)
+			{
+				command_success_nodata(si, _("You have been logged out of \2%s\2."), entity(si->smu)->name);
+
+				if (ircd_on_logout(si->su, entity(si->smu)->name))
+					/* logout killed the user... */
+					return;
+				si->smu->lastlogin = CURRTIME;
+				MOWGLI_ITER_FOREACH_SAFE(n, tn, si->smu->logins.head)
+				{
+					if (n->data == si->su)
+					{
+						mowgli_node_delete(n, &si->smu->logins);
+						mowgli_node_free(n);
+						break;
+					}
+				}
+				si->su->myuser = NULL;
+			}
+
+			myuser_login(si->service, si->su, mn->owner, true);
+		}
+
 		return;
 	}
 	if (!password)
@@ -352,7 +471,7 @@ void enforce_timeout_check(void *arg)
 		if (timeout->timelimit > CURRTIME)
 		{
 			enforce_next = timeout->timelimit;
-			event_add_once("enforce_timeout_check", enforce_timeout_check, NULL, enforce_next - CURRTIME);
+			enforce_timeout_check_timer = mowgli_timer_add_once(base_eventloop, "enforce_timeout_check", enforce_timeout_check, NULL, enforce_next - CURRTIME);
 			break; /* assume sorted list */
 		}
 		u = user_find_named(timeout->nick);
@@ -383,19 +502,22 @@ void enforce_timeout_check(void *arg)
 
 static void show_enforce(hook_user_req_t *hdata)
 {
-
 	if (metadata_find(hdata->mu, "private:doenforce"))
 		command_success_nodata(hdata->si, "%s has enabled nick protection", entity(hdata->mu)->name);
 }
 
 static void check_registration(hook_user_register_check_t *hdata)
 {
-	int prefixlen = strlen(nicksvs.enforce_prefix);
+	int prefixlen;
+
+	return_if_fail(nicksvs.enforce_prefix != NULL);
+
+	prefixlen = strlen(nicksvs.enforce_prefix);
 
 	if (hdata->approved)
 		return;
-	
-	if (!strncasecmp(hdata->account, nicksvs.enforce_prefix, prefixlen) && isdigit(hdata->account[prefixlen]))
+
+	if (!strncasecmp(hdata->account, nicksvs.enforce_prefix, prefixlen) && isdigit((unsigned char)hdata->account[prefixlen]))
 	{
 		command_fail(hdata->si, fault_badparams, "The nick \2%s\2 is reserved and cannot be registered.", hdata->account);
 		hdata->approved = 1;
@@ -439,8 +561,8 @@ static void check_enforce(hook_nick_enforce_t *hdata)
 	if (timeout == NULL)
 	{
 		timeout = mowgli_heap_alloc(enforce_timeout_heap);
-		strlcpy(timeout->nick, hdata->mn->nick, sizeof timeout->nick);
-		strlcpy(timeout->host, hdata->u->host, sizeof timeout->host);
+		mowgli_strlcpy(timeout->nick, hdata->mn->nick, sizeof timeout->nick);
+		mowgli_strlcpy(timeout->host, hdata->u->host, sizeof timeout->host);
 
 		if (!metadata_find(hdata->mn->owner, "private:enforcetime"))
 			timeout->timelimit = CURRTIME + nicksvs.enforce_delay;
@@ -468,9 +590,9 @@ static void check_enforce(hook_nick_enforce_t *hdata)
 		if (enforce_next == 0 || enforce_next > timeout->timelimit)
 		{
 			if (enforce_next != 0)
-				event_delete(enforce_timeout_check, NULL);
+				mowgli_timer_destroy(base_eventloop, enforce_timeout_check_timer);
 			enforce_next = timeout->timelimit;
-			event_add_once("enforce_timeout_check", enforce_timeout_check, NULL, enforce_next - CURRTIME);
+			enforce_timeout_check_timer = mowgli_timer_add_once(base_eventloop, "enforce_timeout_check", enforce_timeout_check, NULL, enforce_next - CURRTIME);
 		}
 	}
 
@@ -491,8 +613,7 @@ static int idcheck_foreach_cb(myentity_t *mt, void *privdata)
 
 void _modinit(module_t *m)
 {
-	MODULE_TRY_REQUEST_DEPENDENCY(m, "nickserv/set_core");
-	MODULE_USE_SYMBOL(ns_set_cmdtree, "nickserv/set_core", "ns_set_cmdtree");
+	MODULE_TRY_REQUEST_SYMBOL(m, ns_set_cmdtree, "nickserv/set_core", "ns_set_cmdtree");
 
 	/* Leave this for compatibility with old versions of this code
 	 * -- jilles
@@ -507,7 +628,7 @@ void _modinit(module_t *m)
 		m->mflags = MODTYPE_FAIL;
 		return;
 	}
-	
+
 	enforce_timeout_heap = mowgli_heap_create(sizeof(enforce_timeout_t), 128, BH_NOW);
 	if (enforce_timeout_heap == NULL)
 	{
@@ -515,7 +636,8 @@ void _modinit(module_t *m)
 		return;
 	}
 
-	event_add("enforce_remove_enforcers", enforce_remove_enforcers, NULL, 300);
+	enforce_remove_enforcers_timer = mowgli_timer_add(base_eventloop, "enforce_remove_enforcers", enforce_remove_enforcers, NULL, 300);
+
 	service_named_bind_command("nickserv", &ns_release);
 	service_named_bind_command("nickserv", &ns_regain);
 	command_add(&ns_set_enforce, *ns_set_cmdtree);
@@ -527,12 +649,15 @@ void _modinit(module_t *m)
 	hook_add_nick_enforce(check_enforce);
 }
 
-void _moddeinit()
+void _moddeinit(module_unload_intent_t intent)
 {
 	enforce_remove_enforcers(NULL);
-	event_delete(enforce_remove_enforcers, NULL);
+
+	mowgli_timer_destroy(base_eventloop, enforce_remove_enforcers_timer);
+
 	if (enforce_next)
-		event_delete(enforce_timeout_check, NULL);
+		mowgli_timer_destroy(base_eventloop, enforce_timeout_check_timer);
+
 	service_named_unbind_command("nickserv", &ns_release);
 	service_named_unbind_command("nickserv", &ns_regain);
 	command_delete(&ns_set_enforce, *ns_set_cmdtree);
