@@ -38,7 +38,7 @@ void _modinit(module_t *m)
 	MODULE_TRY_REQUEST_DEPENDENCY(m, "nickserv/set_quietchg");
 }
 
-void _moddeinit()
+void _moddeinit(module_unload_intent_t intent)
 {
 }
 
