@@ -15,9 +15,8 @@ DECLARE_MODULE_V1
 	"Atheme Development Group <http://www.atheme.org>"
 );
 
-#define DEFAULT_WARN_CLONES	3 /* IPs with more than this are warned about */
-#define EXEMPT_GRACE		10 /* exempt IPs exceeding their allowance by this are banned */
-#define CLONESDB_VERSION	2
+#define CLONESDB_VERSION	3
+#define CLONES_GRACE_TIMEPERIOD	180
 
 static void clones_newuser(hook_user_nick_t *data);
 static void clones_userquit(user_t *u);
@@ -35,10 +34,9 @@ static void os_cmd_clones_duration(sourceinfo_t *si, int parc, char *parv[]);
 static void write_exemptdb(database_handle_t *db);
 
 
-static void clones_kill_users(void *dummy);
-
 static void db_h_ck(database_handle_t *db, const char *type);
 static void db_h_cd(database_handle_t *db, const char *type);
+static void db_h_gr(database_handle_t *db, const char *type);
 static void db_h_ex(database_handle_t *db, const char *type);
 static void db_h_clonesdbv(database_handle_t *db, const char *type);
 
@@ -48,14 +46,13 @@ service_t *serviceinfo;
 
 static mowgli_list_t clone_exempts;
 bool kline_enabled;
+unsigned int grace_count;
 mowgli_patricia_t *hostlist;
 mowgli_heap_t *hostentry_heap;
 static long kline_duration;
-static int clones_allowed, clones_warn, clones_kill;
+static int clones_allowed, clones_warn;
 static unsigned int clones_dbversion = 1;
 
-static mowgli_list_t clones_kill_queue;
-
 typedef struct cexcept_ cexcept_t;
 struct cexcept_
 {
@@ -63,7 +60,6 @@ struct cexcept_
 	int allowed;
 	int warn;
 	char *reason;
-	int kill;
 	long expires;
 };
 
@@ -72,8 +68,8 @@ struct hostentry_
 {
 	char ip[HOSTIPLEN];
 	mowgli_list_t clients;
-	time_t lastaction;
-	unsigned int lastaction_clones;
+	time_t firstkill;
+	unsigned int gracekills;
 };
 
 static inline bool cexempt_expired(cexcept_t *c)
@@ -98,24 +94,6 @@ static void clones_configready(void *unused)
 {
 	clones_allowed = config_options.default_clone_allowed;
 	clones_warn = config_options.default_clone_warn;
-	clones_kill = config_options.default_clone_limit;
-}
-
-static void clones_kill_users(void *dummy)
-{
-	mowgli_node_t *n, *tn;
-	user_t *u;
-
-	MOWGLI_ITER_FOREACH_SAFE(n, tn, clones_kill_queue.head)
-	{
-		u = n->data;
-		kill_user(serviceinfo->me, u, "Too many connections from this host.");
-
-		/* This node delete() and free() will be handled inside the user_quit()
-		 * mowgli_node_delete(n, &clones_kill_queue);
-		 * mowgli_node_free(n);
-		 */
-	}
 }
 
 void _modinit(module_t *m)
@@ -125,7 +103,7 @@ void _modinit(module_t *m)
 
 	if (!module_find_published("backend/opensex"))
 	{
-		slog(LG_INFO, "Module %s requires use of the OpenSEX database backend, refusing to load.", m->header->name);
+		slog(LG_INFO, "Module %s requires use of the OpenSEX database backend, refusing to load.", m->name);
 		m->mflags = MODTYPE_FAIL;
 		return;
 	}
@@ -154,6 +132,7 @@ void _modinit(module_t *m)
 	db_register_type_handler("CLONES-DBV", db_h_clonesdbv);
 	db_register_type_handler("CLONES-CK", db_h_ck);
 	db_register_type_handler("CLONES-CD", db_h_cd);
+	db_register_type_handler("CLONES-GR", db_h_gr);
 	db_register_type_handler("CLONES-EX", db_h_ex);
 
 	hostlist = mowgli_patricia_create(noopcanon);
@@ -185,7 +164,7 @@ static void free_hostentry(const char *key, void *data, void *privdata)
 	mowgli_heap_free(hostentry_heap, he);
 }
 
-void _moddeinit(void)
+void _moddeinit(module_unload_intent_t intent)
 {
 	mowgli_node_t *n, *tn;
 
@@ -204,19 +183,6 @@ void _moddeinit(void)
 		mowgli_node_free(n);
 	}
 
-	if (MOWGLI_LIST_LENGTH(&clones_kill_queue) > 0)
-	{
-		/* Cannot safely delete users from here, so just forget
-		 * about them.
-		 */
-		event_delete(clones_kill_users, NULL);
-		MOWGLI_ITER_FOREACH_SAFE(n, tn, clones_kill_queue.head)
-		{
-			mowgli_node_delete(n, &clones_kill_queue);
-			mowgli_node_free(n);
-		}
-	}
-
 	service_named_unbind_command("operserv", &os_clones);
 
 	command_delete(&os_clones_kline, os_clones_cmds);
@@ -255,6 +221,9 @@ static void write_exemptdb(database_handle_t *db)
 	db_start_row(db, "CLONES-CD");
 	db_write_uint(db, kline_duration);
 	db_commit_row(db);
+	db_start_row(db, "CLONES-GR");
+	db_write_uint(db, grace_count);
+	db_commit_row(db);
 
 	MOWGLI_ITER_FOREACH_SAFE(n, tn, clone_exempts.head)
 	{
@@ -273,7 +242,6 @@ static void write_exemptdb(database_handle_t *db)
 			db_write_word(db, c->ip);
 			db_write_uint(db, c->allowed);
 			db_write_uint(db, c->warn);
-			db_write_uint(db, c->kill);
 			db_write_time(db, c->expires);
 			db_write_str(db, c->reason);
 			db_commit_row(db);
@@ -295,22 +263,30 @@ static void db_h_cd(database_handle_t *db, const char *type)
 	kline_duration = db_sread_uint(db);
 }
 
+static void db_h_gr(database_handle_t *db, const char *type)
+{
+	grace_count = db_sread_uint(db);
+}
+
 static void db_h_ex(database_handle_t *db, const char *type)
 {
-	unsigned int allowed, warn, clonekill;
+	unsigned int allowed, warn;
 
 	const char *ip = db_sread_word(db);
 	allowed = db_sread_uint(db);
 
-	if (clones_dbversion == 2)
+	if (clones_dbversion == 3)
 	{
 		warn = db_sread_uint(db);
-		clonekill = db_sread_uint(db);
-	} 
+	}
+	else if (clones_dbversion == 2)
+	{
+		warn = db_sread_uint(db);
+		db_sread_uint(db); /* trash the old KILL value */
+	}
 	else
 	{
 		warn = allowed;
-		clonekill = allowed+1;
 	}
 
 	time_t expires = db_sread_time(db);
@@ -320,37 +296,11 @@ static void db_h_ex(database_handle_t *db, const char *type)
 	c->ip = sstrdup(ip);
 	c->allowed = allowed;
 	c->warn = warn;
-	c->kill = clonekill;
 	c->expires = expires;
 	c->reason = sstrdup(reason);
 	mowgli_node_add(c, mowgli_node_create(), &clone_exempts);
 }
 
-static unsigned int is_exempt(const char *ip)
-{
-	mowgli_node_t *n;
-
-	/* first check for an exact match */
-	MOWGLI_ITER_FOREACH(n, clone_exempts.head)
-	{
-		cexcept_t *c = n->data;
-
-		if (!strcmp(ip, c->ip))
-			return c->allowed;
-	}
-
-	/* then look for cidr */
-	MOWGLI_ITER_FOREACH(n, clone_exempts.head)
-	{
-		cexcept_t *c = n->data;
-
-		if (!match_ips(c->ip, ip))
-			return c->allowed;
-	}
-
-	return 0;
-}
-
 static cexcept_t * find_exempt(const char *ip)
 {
 	mowgli_node_t *n;
@@ -380,15 +330,15 @@ static void os_cmd_clones(sourceinfo_t *si, int parc, char *parv[])
 {
 	command_t *c;
 	char *cmd = parv[0];
-	
+
 	/* Bad/missing arg */
 	if (!cmd)
 	{
 		command_fail(si, fault_needmoreparams, STR_INSUFFICIENT_PARAMS, "CLONES");
-		command_fail(si, fault_needmoreparams, _("Syntax: CLONES KLINE|LIST|ADDEXEMPT|DELEXEMPT|LISTEXEMPT|SETEXEMPT [parameters]"));
+		command_fail(si, fault_needmoreparams, _("Syntax: CLONES KLINE|LIST|ADDEXEMPT|DELEXEMPT|LISTEXEMPT|SETEXEMPT|DURATION [parameters]"));
 		return;
 	}
-	
+
 	c = command_find(os_clones_cmds, cmd);
 	if (c == NULL)
 	{
@@ -408,12 +358,13 @@ static void os_cmd_clones_kline(sourceinfo_t *si, int parc, char *parv[])
 
 	if (!strcasecmp(arg, "ON"))
 	{
-		if (kline_enabled)
+		if (kline_enabled && grace_count == 0)
 		{
 			command_fail(si, fault_nochange, _("CLONES klines are already enabled."));
 			return;
 		}
 		kline_enabled = true;
+		grace_count = 0;
 		command_success_nodata(si, _("Enabled CLONES klines."));
 		wallops("\2%s\2 enabled CLONES klines", get_oper_name(si));
 		logcommand(si, CMDLOG_ADMIN, "CLONES:KLINE:ON");
@@ -430,10 +381,28 @@ static void os_cmd_clones_kline(sourceinfo_t *si, int parc, char *parv[])
 		wallops("\2%s\2 disabled CLONES klines", get_oper_name(si));
 		logcommand(si, CMDLOG_ADMIN, "CLONES:KLINE:OFF");
 	}
+	else if (isdigit((unsigned char)arg[0]))
+	{
+		unsigned int newgrace = atol(arg);
+		if (kline_enabled && grace_count == newgrace)
+		{
+			command_fail(si, fault_nochange, _("CLONES kline grace is already enabled and set to %d kills."), grace_count);
+		}
+		kline_enabled = true;
+		grace_count = newgrace;
+		command_success_nodata(si, _("Enabled CLONES klines with a grace of %d kills"), grace_count);
+		wallops("\2%s\2 enabled CLONES klines with a grace of %d kills", get_oper_name(si), grace_count);
+		logcommand(si, CMDLOG_ADMIN, "CLONES:KLINE:ON grace %d", grace_count);
+	}
 	else
 	{
 		if (kline_enabled)
-			command_success_string(si, "ON", _("CLONES klines are currently enabled."));
+		{
+			if (grace_count)
+				command_success_string(si, "ON", _("CLONES klines are currently enabled with a grace of %d kills."), grace_count);
+			else
+				command_success_string(si, "ON", _("CLONES klines are currently enabled."));
+		}
 		else
 			command_success_string(si, "OFF", _("CLONES klines are currently disabled."));
 	}
@@ -444,7 +413,6 @@ static void os_cmd_clones_list(sourceinfo_t *si, int parc, char *parv[])
 	hostentry_t *he;
 	int k = 0;
 	mowgli_patricia_iteration_state_t state;
-	int allowed = 0;
 
 	MOWGLI_PATRICIA_FOREACH(he, &state, hostlist)
 	{
@@ -452,8 +420,9 @@ static void os_cmd_clones_list(sourceinfo_t *si, int parc, char *parv[])
 
 		if (k > 3)
 		{
-			if ((allowed = is_exempt(he->ip)))
-				command_success_nodata(si, _("%d from %s (\2EXEMPT\2; allowed %d)"), k, he->ip, allowed);
+			cexcept_t *c = find_exempt(he->ip);
+			if (c)
+				command_success_nodata(si, _("%d from %s (\2EXEMPT\2; allowed %d)"), k, he->ip, c->allowed);
 			else
 				command_success_nodata(si, _("%d from %s"), k, he->ip);
 		}
@@ -481,13 +450,15 @@ static void os_cmd_clones_addexempt(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	}
 
-	clones = atoi(clonesstr);
-	if (clones <= DEFAULT_WARN_CLONES)
+	if (!valid_ip_or_mask(ip))
 	{
-		command_fail(si, fault_badparams, _("Allowed clones count must be more than %d"), DEFAULT_WARN_CLONES);
+		command_fail(si, fault_badparams, _("Invalid IP/mask given."));
+		command_fail(si, fault_badparams, _("Syntax: CLONES ADDEXEMPT <ip> <clones> [!P|!T <minutes>] <reason>"));
 		return;
 	}
 
+	clones = atoi(clonesstr);
+
 	if (expiry && !strcasecmp(expiry, "!P"))
 	{
 		if (!reason)
@@ -498,7 +469,7 @@ static void os_cmd_clones_addexempt(sourceinfo_t *si, int parc, char *parv[])
 		}
 
 		duration = 0;
-		strlcpy(rreason, reason, BUFSIZE);
+		mowgli_strlcpy(rreason, reason, BUFSIZE);
 	}
 	else if (expiry && !strcasecmp(expiry, "!T"))
 	{
@@ -514,33 +485,24 @@ static void os_cmd_clones_addexempt(sourceinfo_t *si, int parc, char *parv[])
 			*reason++ = '\0';
 		expiry += 3;
 
-		if (expiry)
-		{
-			duration = (atol(expiry) * 60);
-			while (isdigit(*expiry))
-				++expiry;
-			if (*expiry == 'h' || *expiry == 'H')
-				duration *= 60;
-			else if (*expiry == 'd' || *expiry == 'D')
-				duration *= 1440;
-			else if (*expiry == 'w' || *expiry == 'W')
-				duration *= 10080;
-			else if (*expiry == '\0')
-				;
-			else
-				duration = 0;
-
-			if (duration == 0)
-			{
-				command_fail(si, fault_badparams, _("Invalid duration given."));
-				command_fail(si, fault_badparams, _("Syntax: CLONES ADDEXEMPT <ip> <clones> [!P|!T <minutes>] <reason>"));
-				return;
-			}
-		}
+		duration = (atol(expiry) * 60);
+		while (isdigit((unsigned char)*expiry))
+			++expiry;
+		if (*expiry == 'h' || *expiry == 'H')
+			duration *= 60;
+		else if (*expiry == 'd' || *expiry == 'D')
+			duration *= 1440;
+		else if (*expiry == 'w' || *expiry == 'W')
+			duration *= 10080;
+		else if (*expiry == '\0')
+			;
 		else
+			duration = 0;
+
+		if (duration == 0)
 		{
-			command_fail(si, fault_needmoreparams, STR_INSUFFICIENT_PARAMS, "CLONES ADDEXEMPT");
-			command_fail(si, fault_needmoreparams, _("Syntax: AKILL ADD <nick|hostmask> [!P|!T <minutes>] <reason>"));
+			command_fail(si, fault_badparams, _("Invalid duration given."));
+			command_fail(si, fault_badparams, _("Syntax: CLONES ADDEXEMPT <ip> <clones> [!P|!T <minutes>] <reason>"));
 			return;
 		}
 
@@ -551,17 +513,17 @@ static void os_cmd_clones_addexempt(sourceinfo_t *si, int parc, char *parv[])
 			return;
 		}
 
-		strlcpy(rreason, reason, BUFSIZE);
+		mowgli_strlcpy(rreason, reason, BUFSIZE);
 	}
 	else if (expiry)
 	{
 		duration = config_options.clone_time;
 
-		strlcpy(rreason, expiry, BUFSIZE);
+		mowgli_strlcpy(rreason, expiry, BUFSIZE);
 		if (reason)
 		{
-			strlcat(rreason, " ", BUFSIZE);
-			strlcat(rreason, reason, BUFSIZE);
+			mowgli_strlcat(rreason, " ", BUFSIZE);
+			mowgli_strlcat(rreason, reason, BUFSIZE);
 		}
 	}
 	else
@@ -607,7 +569,6 @@ static void os_cmd_clones_addexempt(sourceinfo_t *si, int parc, char *parv[])
 
 	c->allowed = clones;
 	c->warn = clones;
-	c->kill = clones+1;
 	c->expires = duration ? (CURRTIME + duration) : 0;
 
 	logcommand(si, CMDLOG_ADMIN, "CLONES:ADDEXEMPT: \2%s\2 \2%d\2 (reason: \2%s\2) (duration: \2%s\2)", ip, clones, c->reason, timediff(duration));
@@ -619,7 +580,11 @@ static void os_cmd_clones_delexempt(sourceinfo_t *si, int parc, char *parv[])
 	char *arg = parv[0];
 
 	if (!arg)
+	{
+		command_fail(si, fault_needmoreparams, STR_INSUFFICIENT_PARAMS, "CLONES DELEXEMPT");
+		command_fail(si, fault_needmoreparams, _("Syntax: CLONES DELEXEMPT <ip>"));
 		return;
+	}
 
 	MOWGLI_ITER_FOREACH_SAFE(n, tn, clone_exempts.head)
 	{
@@ -657,90 +622,45 @@ static void os_cmd_clones_setexempt(sourceinfo_t *si, int parc, char *parv[])
 	char *subcmd = parv[1];
 	char *clonesstr = parv[2];
 	char *reason = parv[3];
-	int clones;
 	char rreason[BUFSIZE];
 
 	long duration;
 
+
 	if (!ip || !subcmd || !clonesstr)
 	{
 		command_fail(si, fault_needmoreparams, STR_INSUFFICIENT_PARAMS, "CLONES SETEXEMPT");
-		command_fail(si, fault_needmoreparams, _("Syntax: CLONES SETEXEMPT [DEFAULT | <ip>] <ALLOWED | WARN | KILL> <limit>"));
+		command_fail(si, fault_needmoreparams, _("Syntax: CLONES SETEXEMPT [DEFAULT | <ip>] <ALLOWED | WARN> <limit>"));
 		command_fail(si, fault_needmoreparams, _("Syntax: CLONES SETEXEMPT <ip> <REASON | DURATION> <value>"));
 		return;
 	}
 
-	if (!strcasecmp(ip,"DEFAULT"))
-	{
-		clones = atoi(clonesstr);
-		if (!strcasecmp(subcmd,"ALLOWED"))
-		{
-			if (clones < DEFAULT_WARN_CLONES)
-			{
-				command_fail(si, fault_badparams, _("Default allowed clone limit must be more than %d."),DEFAULT_WARN_CLONES);
-				return;
-			}
-			else if (clones > clones_warn)
-			{
-				command_fail(si, fault_badparams, _("Default allowed clone limit cannot be higher than the default warn limit (currently %d)."),clones_warn);
-				return;
-			}
-			else if (clones >= clones_kill)
-			{
-				command_fail(si, fault_badparams, _("Default allowed clone limit must be lower than the default kill limit (currently %d)."),clones_kill);
-				return;
-			}
+	int clones = atoi(clonesstr);
 
+	if (!strcasecmp(ip, "DEFAULT"))
+	{
+		if (!strcasecmp(subcmd, "ALLOWED"))
+		{
 			clones_allowed = clones;
-			command_success_nodata(si,_("Default allowed clone limit set to \2%d\2."),clones_allowed);
+			command_success_nodata(si, _("Default allowed clone limit set to \2%d\2."), clones_allowed);
 		}
 		else if (!strcasecmp(subcmd,"WARN"))
 		{
 			if (clones == 0)
 			{
-				command_success_nodata(si,_("Default clone warning has been disabled."));
+				command_success_nodata(si, _("Default clone warning has been disabled."));
 				clones_warn = 0;
 				return;
 			}
-			else if (clones < clones_allowed)
-			{
-				command_fail(si, fault_badparams, _("Default warned clone limit cannot be lower than the default allowed limit (currently %d)."), clones_allowed);
-				return;
-			}
-			else if (clones >= clones_kill && clones_kill > 0)
-			{
-				command_fail(si, fault_badparams, _("Default warned clone limit must be less than the default kill limit (currently %d)."), clones_kill);
-				return;
-			}
 
 			clones_warn = clones;
-			command_success_nodata(si,_("Default warned clone limit set to \2%d\2"),clones_warn);
-		}
-		else if (!strcasecmp(subcmd,"KILL"))
-		{
-			if (clones == 0)
-			{
-				command_success_nodata(si,_("Default clone kill limit has been disabled."));
-			}
-			else if (clones <= clones_warn && clones_warn > 0)
-			{
-				command_fail(si, fault_badparams, _("Default kill clone limit must be more than the default warn limit (currently %d)."), clones_warn);
-				return;
-			}
-			else if (clones <= clones_allowed)
-			{
-				command_fail(si, fault_badparams, _("Default kill clone limit must be more than the default allowed limit (currently %d) since warnings are disabled."), clones_allowed);
-				return;
-			}
-
-			clones_kill = clones;
-			command_success_nodata(si,_("Default kill clone limit set to \2%d\2"),clones_kill);
+			command_success_nodata(si, _("Default warned clone limit set to \2%d\2"), clones_warn);
 		}
 		else
 		{
 			/* Invalid parameters */
 			command_fail(si, fault_badparams, _("Invalid syntax given."));
-			command_fail(si, fault_badparams, _("Syntax: CLONES SETEXEMPT DEFAULT <ALLOWED | WARN | KILL> <limit>"));
+			command_fail(si, fault_badparams, _("Syntax: CLONES SETEXEMPT DEFAULT <ALLOWED | WARN> <limit>"));
 			return;
 		}
 
@@ -761,81 +681,46 @@ static void os_cmd_clones_setexempt(sourceinfo_t *si, int parc, char *parv[])
 			}
 			else if (!strcmp(c->ip, ip))
 			{
-				if (!strcasecmp(subcmd,"ALLOWED"))
+				if (!strcasecmp(subcmd, "ALLOWED"))
 				{
-					clones = atoi(clonesstr);
-					if (clones < DEFAULT_WARN_CLONES)
+					if (clones < c->warn)
 					{
-						command_fail(si, fault_badparams, _("Allowed clones must be more than %d."),DEFAULT_WARN_CLONES);
-						return;
-					}
-					else if (clones > c->warn)
-					{
-						command_fail(si, fault_badparams, _("The ALLOWED clone limit must be lower than the WARN limit (currently %d)"),c->warn);
-						return;
-					}
-					else if (clones >= c->kill)
-					{
-						command_fail(si, fault_badparams, _("The ALLOWED clone limit must be lower than the KILL limit (currently %d)"),c->kill);
+						command_fail(si, fault_badparams, _("Allowed clones limit must be greater than or equal to the warned limit of %d"), c->warn);
 						return;
 					}
 
 					c->allowed = clones;
-					command_success_nodata(si,_("Allowed clones limit for host \2%s\2 set to \2%d\2"),ip,c->allowed);
+					command_success_nodata(si, _("Allowed clones limit for host \2%s\2 set to \2%d\2"), ip, c->allowed);
 				}
-				else if (!strcasecmp(subcmd,"WARN"))
+				else if (!strcasecmp(subcmd, "WARN"))
 				{
-					clones = atoi(clonesstr);
 					if (clones == 0)
 					{
-						command_success_nodata(si,_("Clone warning messages will be disabled for host \2%s\2"),ip);
+						command_success_nodata(si, _("Clone warning messages will be disabled for host \2%s\2"), ip);
 						c->warn = 0;
 						return;
 					}
-					else if (clones < c->allowed)
+					else if (clones > c->allowed)
 					{
-						command_fail(si, fault_badparams, _("Warned clones limit cannot be less than the allowed limit of %d"), c->allowed);
-						return;
-					}
-					else if (clones >= c->kill && c->kill > 0)
-					{
-						command_fail(si, fault_badparams, _("Warned clones limit must be less than the kill limit of %d."), c->kill);
+						command_fail(si, fault_badparams, _("Warned clones limit must be lower than or equal to the allowed limit of %d"), c->allowed);
 						return;
 					}
 
 					c->warn = clones;
-					command_success_nodata(si,_("Warned clones limit for host \2%s\2 set to \2%d\2"),ip,c->warn);
+					command_success_nodata(si, _("Warned clones limit for host \2%s\2 set to \2%d\2"), ip, c->warn);
 				}
-				else if (!strcasecmp(subcmd,"KILL"))
-				{
-					clones = atoi(clonesstr);
-					if (clones == 0)
-					{
-						command_success_nodata(si,_("Clone kills will be disabled for host \2%s\2"),ip);
-						c->kill = 0;
-						return;
-					}
-					else if (clones <= clones_warn)
-					{
-						command_fail(si, fault_badparams, _("Killed clones limit must be more than the warn limit of %d"), c->warn);
-						return;
-					}
-
-					c->kill = clones;
-					command_success_nodata(si,_("Kill clone limit for host \2%s\2 set to \2%d\2"),ip,c->kill);
-				}
-				else if (!strcasecmp(subcmd,"DURATION"))
+				else if (!strcasecmp(subcmd, "DURATION"))
 				{
 					char *expiry = clonesstr;
-					if (!strcmp(expiry,"0"))
+					if (!strcmp(expiry, "0"))
 					{
 						c->expires = 0;
-						command_success_nodata(si,_("Clone exemption duration for host \2%s\2 set to \2permanent\2"),ip);
+						command_success_nodata(si, _("Clone exemption duration for host \2%s\2 set to \2permanent\2"), ip);
 					}
 					else
 					{
 						duration = (atol(expiry) * 60);
-						while (isdigit(*expiry))
+						while (isdigit((unsigned char)*expiry))
 							++expiry;
 						if (*expiry == 'h' || *expiry == 'H')
 							duration *= 60;
@@ -855,34 +740,35 @@ static void os_cmd_clones_setexempt(sourceinfo_t *si, int parc, char *parv[])
 							return;
 						}
 						c->expires = CURRTIME + duration;
-						command_success_nodata(si,_("Clone exemption duration for host \2%s\2 set to \2%s\2 (%ld seconds)"),ip,clonesstr,duration);
+						command_success_nodata(si, _("Clone exemption duration for host \2%s\2 set to \2%s\2 (%ld seconds)"), ip, clonesstr, duration);
 					}
 
 				}
-				else if (!strcasecmp(subcmd,"REASON"))
+				else if (!strcasecmp(subcmd, "REASON"))
 				{
-					strlcpy(rreason, clonesstr, BUFSIZE);
+					mowgli_strlcpy(rreason, clonesstr, BUFSIZE);
 					if (reason)
 					{
-						strlcat(rreason, " ", BUFSIZE);
-						strlcat(rreason, reason, BUFSIZE);
+						mowgli_strlcat(rreason, " ", BUFSIZE);
+						mowgli_strlcat(rreason, reason, BUFSIZE);
 					}
-					/* strlcat(rreason,clonesstr,BUFSIZE); */
+					/* mowgli_strlcat(rreason,clonesstr,BUFSIZE); */
 					free(c->reason);
 					c->reason = sstrdup(rreason);
-					command_success_nodata(si,_("Clone exemption reason for host \2%s\2 changed to \2%s\2"),ip,c->reason);
+					command_success_nodata(si, _("Clone exemption reason for host \2%s\2 changed to \2%s\2"), ip, c->reason);
 				}
 				else
 				{
 					/* Invalid parameters */
 					command_fail(si, fault_badparams, _("Invalid syntax given."));
-					command_fail(si, fault_badparams, _("Syntax: CLONES SETEXEMPT <IP> <ALLOWED | WARN | KILL | DURATION | REASON> <value>"));
+					command_fail(si, fault_badparams, _("Syntax: CLONES SETEXEMPT <IP> <ALLOWED | WARN | DURATION | REASON> <value>"));
 					return;
 				}
+
+				logcommand(si, CMDLOG_ADMIN, "CLONES:SETEXEMPT: \2%s\2 \2%d\2 (reason: \2%s\2) (duration: \2%s\2)", ip, clones, c->reason, timediff((c->expires - CURRTIME)));
+
 				return;
 			}
-
-			logcommand(si, CMDLOG_ADMIN, "CLONES:SETEXEMPT: \2%s\2 \2%d\2 (reason: \2%s\2) (duration: \2%s\2)", ip, clones, c->reason, timediff(duration));
 		}
 
 		command_fail(si, fault_nosuch_target, _("\2%s\2 not found in clone exempt list."), ip);
@@ -901,7 +787,7 @@ static void os_cmd_clones_duration(sourceinfo_t *si, int parc, char *parv[])
 	}
 
 	duration = (atol(s) * 60);
-	while (isdigit(*s))
+	while (isdigit((unsigned char)*s))
 		s++;
 	if (*s == 'h' || *s == 'H')
 		duration *= 60;
@@ -927,7 +813,7 @@ static void os_cmd_clones_duration(sourceinfo_t *si, int parc, char *parv[])
 static void os_cmd_clones_listexempt(sourceinfo_t *si, int parc, char *parv[])
 {
 
-	command_success_nodata(si, "DEFAULT - allowed limit %d, warn on %d, kill after %d",clones_allowed,clones_warn,clones_kill);
+	command_success_nodata(si, _("DEFAULT - allowed limit %d, warn on %d"), clones_allowed, clones_warn);
 	mowgli_node_t *n, *tn;
 
 	MOWGLI_ITER_FOREACH_SAFE(n, tn, clone_exempts.head)
@@ -943,9 +829,9 @@ static void os_cmd_clones_listexempt(sourceinfo_t *si, int parc, char *parv[])
 			mowgli_node_free(n);
 		}
 		else if (c->expires)
-			command_success_nodata(si, "%s - allowed limit %d, warn on %d, kill after %d - expires in %s - \2%s\2", c->ip, c->allowed, c->warn, c->kill, timediff(c->expires > CURRTIME ? c->expires - CURRTIME : 0), c->reason);
+			command_success_nodata(si, _("%s - allowed limit %d, warn on %d - expires in %s - \2%s\2"), c->ip, c->allowed, c->warn, timediff(c->expires > CURRTIME ? c->expires - CURRTIME : 0), c->reason);
 		else
-			command_success_nodata(si, "%s - allowed limit %d, warn on %d, kill after %d - \2permanent\2 - \2%s\2", c->ip, c->allowed, c->warn, c->kill, c->reason);
+			command_success_nodata(si, _("%s - allowed limit %d, warn on %d - \2permanent\2 - \2%s\2"), c->ip, c->allowed, c->warn, c->reason);
 	}
 	command_success_nodata(si, _("End of CLONES LISTEXEMPT"));
 	logcommand(si, CMDLOG_ADMIN, "CLONES:LISTEXEMPT");
@@ -956,21 +842,22 @@ static void clones_newuser(hook_user_nick_t *data)
 	user_t *u = data->u;
 	unsigned int i;
 	hostentry_t *he;
-	unsigned int allowed, warn, clonekill;
+	unsigned int allowed, warn;
+	mowgli_node_t *n;
 
 	/* If the user has been killed, don't do anything. */
 	if (!u)
 		return;
 
-	/* User has no IP, ignore him */
-	if (is_internal_client(u) || *u->ip == '\0')
+	/* User has no IP, ignore them */
+	if (is_internal_client(u) || u->ip == NULL)
 		return;
 
 	he = mowgli_patricia_retrieve(hostlist, u->ip);
 	if (he == NULL)
 	{
 		he = mowgli_heap_alloc(hostentry_heap);
-		strlcpy(he->ip, u->ip, sizeof he->ip);
+		mowgli_strlcpy(he->ip, u->ip, sizeof he->ip);
 		mowgli_patricia_add(hostlist, he->ip, he);
 	}
 	mowgli_node_add(u, mowgli_node_create(), &he->clients);
@@ -981,46 +868,78 @@ static void clones_newuser(hook_user_nick_t *data)
 	{
 		allowed = clones_allowed;
 		warn = clones_warn;
-		clonekill = clones_kill;
 	}
 	else
 	{
-		allowed = clones_allowed > c->allowed ? clones_allowed : c->allowed;
-		warn = clones_warn > c->warn && c->warn > 0 ? clones_warn : c->warn;
-		clonekill = clones_kill > c->kill && c->kill > 0? clones_kill : c->kill;
+		allowed = c->allowed;
+		warn = c->warn;
 	}
 
-	if (i >= allowed)
+	if (config_options.clone_increase)
 	{
-		if (i >= allowed && (i < clonekill || clonekill == 0))
+		unsigned int real_allowed = allowed;
+		unsigned int real_warn = warn;
+
+		MOWGLI_ITER_FOREACH(n, he->clients.head)
 		{
-			slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (\2%d\2 allowed)", i, u->ip, u->nick, u->user, u->host, allowed);
-			if (i >= warn && warn > 0)
-				msg(serviceinfo->nick, u->nick, "\2WARNING\2: You may not have any more than \2%d\2 clients on the network at once. Any further connections over this limit risks being removed.", allowed);
+			user_t *tu = n->data;
+
+			if (tu->myuser == NULL)
+				continue;
+			if (allowed != 0)
+				allowed++;
+			if (warn != 0)
+				warn++;
 		}
-		else if (i >= clonekill && is_autokline_exempt(u) && clonekill > 0)
+
+		/* A hard limit of 2x the "real" limit sounds good IMO --jdhore */
+		if (allowed > (real_allowed * 2))
+			allowed = real_allowed * 2;
+		if (warn > (real_warn * 2))
+			warn = real_warn * 2;
+	}
+
+	if (i > allowed && allowed != 0)
+	{
+		/* User has exceeded the maximum number of allowed clones. */
+		if (is_autokline_exempt(u))
 			slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (user is autokline exempt)", i, u->ip, u->nick, u->user, u->host);
-		else if (i >= clonekill && !kline_enabled && clonekill > 0)
+		else if (!kline_enabled || he->gracekills < grace_count || (grace_count > 0 && he->firstkill < time(NULL) - CLONES_GRACE_TIMEPERIOD))
 		{
-			slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (TKLINE disabled, killing user)", i, u->ip, u->nick, u->user, u->host);
-
-			if (MOWGLI_LIST_LENGTH(&clones_kill_queue) == 0)
+			if (he->firstkill < time(NULL) - CLONES_GRACE_TIMEPERIOD)
 			{
-				event_add_once("clones_kill_users", clones_kill_users, NULL, 0);
+				he->firstkill = time(NULL);
+				he->gracekills = 1;
+			}
+			else
+			{
+				he->gracekills++;
 			}
 
-			if (!mowgli_node_find(u, &clones_kill_queue))
-			{
-				mowgli_node_add(u, mowgli_node_create(), &clones_kill_queue);
+			if (!kline_enabled)
+				slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (TKLINE disabled, killing user)", i, u->ip, u->nick, u->user, u->host);
+			else
+				slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (grace period, killing user, %d grace kills remaining)", i, u->ip, u->nick,
+					u->user, u->host, grace_count - he->gracekills);
+
+			kill_user(serviceinfo->me, u, "Too many connections from this host.");
+			data->u = NULL; /* Required due to kill_user being called during user_add hook. --mr_flea */
+		}
+		else
+		{
+			if (! (u->flags & UF_KLINESENT)) {
+				slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (TKLINE due to excess clones)", i, u->ip, u->nick, u->user, u->host);
+				kline_sts("*", "*", u->ip, kline_duration, "Excessive clones");
+				u->flags |= UF_KLINESENT;
 			}
 		}
-		else if (i >= clonekill && clonekill > 0)
-		{
-			slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (TKLINE due to excess clones)", i, u->ip, u->nick, u->user, u->host);
-			kline_sts("*", "*", u->ip, kline_duration, "Excessive clones");
-		}
 
 	}
+	else if (i >= warn && warn != 0)
+	{
+		slog(LG_INFO, "CLONES: \2%d\2 clones on \2%s\2 (%s!%s@%s) (\2%d\2 allowed)", i, u->ip, u->nick, u->user, u->host, allowed);
+		msg(serviceinfo->nick, u->nick, _("\2WARNING\2: You may not have more than \2%d\2 clients connected to the network at once. Any further connections risks being removed."), allowed);
+	}
 }
 
 static void clones_userquit(user_t *u)
@@ -1028,8 +947,8 @@ static void clones_userquit(user_t *u)
 	mowgli_node_t *n;
 	hostentry_t *he;
 
-	/* User has no IP, ignore him */
-	if (is_internal_client(u) || *u->ip == '\0')
+	/* User has no IP, ignore them */
+	if (is_internal_client(u) || u->ip == NULL)
 		return;
 
 	he = mowgli_patricia_retrieve(hostlist, u->ip);
@@ -1045,17 +964,11 @@ static void clones_userquit(user_t *u)
 		mowgli_node_free(n);
 		if (MOWGLI_LIST_LENGTH(&he->clients) == 0)
 		{
+			/* TODO: free later if he->firstkill > time(NULL) - CLONES_GRACE_TIMEPERIOD. */
 			mowgli_patricia_delete(hostlist, he->ip);
 			mowgli_heap_free(hostentry_heap, he);
 		}
 	}
-
-	n = mowgli_node_find(u, &clones_kill_queue);
-	if (n != NULL)
-	{
-		mowgli_node_delete(n, &clones_kill_queue);
-		mowgli_node_free(n);
-	}
 }
 
 /* vim:cinoptions=>s,e0,n0,f0,{0,}0,^0,=s,ps,t0,c3,+s,(2s,us,)20,*30,gs,hs
