@@ -26,7 +26,7 @@ void _modinit(module_t *m)
 	service_named_bind_command("operserv", &os_modlist);
 }
 
-void _moddeinit()
+void _moddeinit(module_unload_intent_t intent)
 {
 	service_named_unbind_command("operserv", &os_modlist);
 }
@@ -42,7 +42,7 @@ static void os_cmd_modlist(sourceinfo_t *si, int parc, char *parv[])
 		module_t *m = n->data;
 
 		command_success_nodata(si, _("%2d: %-20s [loaded at 0x%lx]"),
-			++i, m->header->name, (unsigned long)m->address);
+			++i, m->name, (unsigned long)m->address);
 	}
 
 	command_success_nodata(si, _("\2%d\2 modules loaded."), i);
