@@ -17,42 +17,43 @@ DECLARE_MODULE_V1("protocol/bahamut", true, _modinit, NULL, PACKAGE_STRING, "Ath
 /* *INDENT-OFF* */
 
 ircd_t Bahamut = {
-        "Bahamut 1.8.x",                /* IRCd name */
-        "$",                            /* TLD Prefix, used by Global. */
-        false,                          /* Whether or not we use IRCNet/TS6 UID */
-        false,                          /* Whether or not we use RCOMMAND */
-        false,                          /* Whether or not we support channel owners. */
-        false,                          /* Whether or not we support channel protection. */
-        false,                          /* Whether or not we support halfops. */
-	false,				/* Whether or not we use P10 */
-	false,				/* Whether or not we use vHosts. */
-	CMODE_OPERONLY,			/* Oper-only cmodes */
-        0,                              /* Integer flag for owner channel flag. */
-        0,                              /* Integer flag for protect channel flag. */
-        0,                              /* Integer flag for halfops. */
-        "+",                            /* Mode we set for owner. */
-        "+",                            /* Mode we set for protect. */
-        "+",                            /* Mode we set for halfops. */
-	PROTOCOL_BAHAMUT,		/* Protocol type */
-	0,                              /* Permanent cmodes */
-	0,                              /* Oper-immune cmode */
-	"beI",                          /* Ban-like cmodes */
-	'e',                            /* Except mchar */
-	'I',                            /* Invex mchar */
-	IRCD_HOLDNICK                   /* Flags */
+	.ircdname = "Bahamut 1.8.x",
+	.tldprefix = "$",
+	.uses_uid = false,
+	.uses_rcommand = false,
+	.uses_owner = false,
+	.uses_protect = false,
+	.uses_halfops = false,
+	.uses_p10 = false,
+	.uses_vhost = false,
+	.oper_only_modes = CMODE_OPERONLY,
+	.owner_mode = 0,
+	.protect_mode = 0,
+	.halfops_mode = 0,
+	.owner_mchar = "+",
+	.protect_mchar = "+",
+	.halfops_mchar = "+",
+	.type = PROTOCOL_BAHAMUT,
+	.perm_mode = 0,
+	.oimmune_mode = 0,
+	.ban_like_modes = "beI",
+	.except_mchar = 'e',
+	.invex_mchar = 'I',
+	.flags = IRCD_HOLDNICK,
 };
 
 struct cmode_ bahamut_mode_list[] = {
-  { 'i', CMODE_INVITE   },
-  { 'm', CMODE_MOD      },
-  { 'n', CMODE_NOEXT    },
-  { 'p', CMODE_PRIV     },
-  { 's', CMODE_SEC      },
-  { 't', CMODE_TOPIC    },
-  { 'c', CMODE_NOCOLOR  },
-  { 'M', CMODE_MODREG   },
-  { 'R', CMODE_REGONLY  },
+  { 'i', CMODE_INVITE	},
+  { 'm', CMODE_MOD	},
+  { 'n', CMODE_NOEXT	},
+  { 'p', CMODE_PRIV	},
+  { 's', CMODE_SEC	},
+  { 't', CMODE_TOPIC	},
+  { 'c', CMODE_NOCOLOR	},
+  { 'M', CMODE_MODREG	},
+  { 'R', CMODE_REGONLY	},
   { 'O', CMODE_OPERONLY },
+  { 'r', CMODE_CHANREG	},
   { '\0', 0 }
 };
 
@@ -100,7 +101,7 @@ static bool check_jointhrottle(const char *value, channel_t *c, mychan_t *mc, us
 				return false;
 			arg2 = p + 1;
 		}
-		else if (!isdigit(*p))
+		else if (!isdigit((unsigned char)*p))
 			return false;
 		p++;
 	}
@@ -129,23 +130,23 @@ static bool check_jointhrottle(const char *value, channel_t *c, mychan_t *mc, us
 	return true;
 }
 
+static bool use_nickipstr = false;
+
 /* login to our uplink */
 static unsigned int bahamut_server_login(void)
 {
 	int ret;
 
-	ret = sts("PASS %s :TS", curr_uplink->pass);
+	ret = sts("PASS %s :TS", curr_uplink->send_pass);
 	if (ret == 1)
 		return 1;
 
 	me.bursting = true;
 
-	sts("CAPAB SSJOIN NOQUIT BURST ZIP NICKIP TSMODE");
+	sts("CAPAB SSJOIN NOQUIT BURST ZIP NICKIP TSMODE NICKIPSTR");
 	sts("SERVER %s 1 :%s", me.name, me.desc);
 	sts("SVINFO 5 3 0 :%lu", (unsigned long)CURRTIME);
 
-	services_init();
-
 	return 0;
 }
 
@@ -154,7 +155,10 @@ static void bahamut_introduce_nick(user_t *u)
 {
 	const char *umode = user_get_umodestr(u);
 
-	sts("NICK %s 1 %lu %s %s %s %s 0 0 :%s", u->nick, (unsigned long)u->ts, umode, u->user, u->host, me.name, u->gecos);
+	if (use_nickipstr)
+		sts("NICK %s 1 %lu %s %s %s %s 0 0.0.0.0 :%s", u->nick, (unsigned long)u->ts, umode, u->user, u->host, me.name, u->gecos);
+	else
+		sts("NICK %s 1 %lu %s %s %s %s 0 0 :%s", u->nick, (unsigned long)u->ts, umode, u->user, u->host, me.name, u->gecos);
 }
 
 /* invite a user to a channel */
@@ -165,9 +169,6 @@ static void bahamut_invite_sts(user_t *sender, user_t *target, channel_t *channe
 
 static void bahamut_quit_sts(user_t *u, const char *reason)
 {
-	if (!me.connected)
-		return;
-
 	sts(":%s QUIT :%s", u->nick, reason);
 }
 
@@ -307,9 +308,6 @@ static void bahamut_kline_sts(const char *server, const char *user, const char *
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s AKILL %s %s %ld %s %lu :%s", me.name, host, user, duration, svs != NULL ? svs->nick : me.name, (unsigned long)CURRTIME, reason);
 }
@@ -317,17 +315,13 @@ static void bahamut_kline_sts(const char *server, const char *user, const char *
 /* server-to-server UNKLINE wrapper */
 static void bahamut_unkline_sts(const char *server, const char *user, const char *host)
 {
-	if (!me.connected)
-		return;
-
 	sts(":%s RAKILL %s %s", me.name, host, user);
 }
 
 /* topic wrapper */
 static void bahamut_topic_sts(channel_t *c, user_t *source, const char *setter, time_t ts, time_t prevts, const char *topic)
 {
-	if (!me.connected || !c)
-		return;
+	return_if_fail(c != NULL);
 
 	sts(":%s TOPIC %s %s %lu :%s", source->nick, c->name, setter, (unsigned long)ts, topic);
 }
@@ -335,8 +329,9 @@ static void bahamut_topic_sts(channel_t *c, user_t *source, const char *setter,
 /* mode wrapper */
 static void bahamut_mode_sts(char *sender, channel_t *target, char *modes)
 {
-	if (!me.connected)
-		return;
+	return_if_fail(sender != NULL);
+	return_if_fail(target != NULL);
+	return_if_fail(modes != NULL);
 
 	sts(":%s MODE %s %lu %s", sender, target->name, (unsigned long)target->ts, modes);
 }
@@ -344,17 +339,13 @@ static void bahamut_mode_sts(char *sender, channel_t *target, char *modes)
 /* ping wrapper */
 static void bahamut_ping_sts(void)
 {
-	if (!me.connected)
-		return;
-
 	sts("PING :%s", me.name);
 }
 
 /* protocol-specific stuff to do on login */
 static void bahamut_on_login(user_t *u, myuser_t *account, const char *wantedhost)
 {
-	if (!me.connected || u == NULL)
-		return;
+	return_if_fail(u != NULL);
 
 	/* Can only do this for nickserv, and can only record identified
 	 * state if logged in to correct nick, sorry -- jilles
@@ -366,8 +357,7 @@ static void bahamut_on_login(user_t *u, myuser_t *account, const char *wantedhos
 /* protocol-specific stuff to do on login */
 static bool bahamut_on_logout(user_t *u, const char *account)
 {
-	if (!me.connected)
-		return false;
+	return_val_if_fail(u != NULL, false);
 
 	if (!nicksvs.no_nick_ownership)
 		sts(":%s SVSMODE %s -r+d %lu", nicksvs.nick, u->nick, (unsigned long)CURRTIME);
@@ -378,9 +368,6 @@ static void bahamut_jupe(const char *server, const char *reason)
 {
 	server_t *s;
 
-	if (!me.connected)
-		return;
-
 	sts(":%s SQUIT %s :%s", me.name, server, reason);
 	s = server_find(server);
 	/* If the server is not directly connected to our uplink, we
@@ -392,7 +379,7 @@ static void bahamut_jupe(const char *server, const char *reason)
 		sts(":%s SERVER %s 2 :%s", me.name, server, reason);
 }
 
-static void bahamut_fnc_sts(user_t *source, user_t *u, char *newnick, int type)
+static void bahamut_fnc_sts(user_t *source, user_t *u, const char *newnick, int type)
 {
 	sts(":%s SVSNICK %s %s %lu", source->nick, u->nick, newnick,
 			(unsigned long)(CURRTIME - 60));
@@ -442,7 +429,7 @@ static void m_pong(sourceinfo_t *si, int parc, char *parv[])
 	if (s->uplink != me.me)
 		handle_eob(s);
 
-	if (irccasecmp(me.actual, parv[0]))
+	if (s != si->s)
 		return;
 
 	me.uplinkpong = CURRTIME;
@@ -499,7 +486,7 @@ static void m_sjoin(sourceinfo_t *si, int parc, char *parv[])
 {
 	/*
 	 *  -> :proteus.malkier.net SJOIN 1073516550 #shrike +tn :@sycobuny @+rakaur
-	 *      also:
+	 *	also:
 	 *  -> :nenolod_ SJOIN 1117334567 #chat
 	 */
 
@@ -631,7 +618,6 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 {
 	server_t *s;
 	user_t *u;
-	struct in_addr ip;
 	char ipstring[64];
 	bool realchange;
 
@@ -647,10 +633,18 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 
 		slog(LG_DEBUG, "m_nick(): new user on `%s': %s", s->name, parv[0]);
 
-		ip.s_addr = ntohl(strtoul(parv[8], NULL, 10));
-		ipstring[0] = '\0';
-		if (!inet_ntop(AF_INET, &ip, ipstring, sizeof ipstring))
+		if (!use_nickipstr)
+		{
+			struct in_addr ip;
+
+			ip.s_addr = ntohl(strtoul(parv[8], NULL, 10));
 			ipstring[0] = '\0';
+			if (!inet_ntop(AF_INET, &ip, ipstring, sizeof ipstring))
+				ipstring[0] = '\0';
+		}
+		else
+			mowgli_strlcpy(ipstring, parv[8], sizeof ipstring);
+
 		u = user_add(parv[0], parv[4], parv[5], NULL, ipstring, NULL, parv[9], s, atoi(parv[2]));
 		if (u == NULL)
 			return;
@@ -851,7 +845,7 @@ static void m_join(sourceinfo_t *si, int parc, char *parv[])
 
 static void m_pass(sourceinfo_t *si, int parc, char *parv[])
 {
-	if (strcmp(curr_uplink->pass, parv[0]))
+	if (strcmp(curr_uplink->receive_pass, parv[0]))
 	{
 		slog(LG_INFO, "m_pass(): password mismatch from uplink; aborting");
 		runflags |= RF_SHUTDOWN;
@@ -868,6 +862,25 @@ static void m_motd(sourceinfo_t *si, int parc, char *parv[])
 	handle_motd(si->su);
 }
 
+static void m_capab(sourceinfo_t *si, int parc, char *parv[])
+{
+	int i;
+
+	use_nickipstr = false;
+
+	for (i = 0; i < parc; i++)
+	{
+		if (!irccasecmp(parv[i], "NICKIPSTR"))
+		{
+			slog(LG_DEBUG, "m_capab(): uplink supports string-based IP addresses, enabling support.");
+			use_nickipstr = true;
+		}
+	}
+
+	/* now burst services with NICKIP set correctly. */
+	services_init();
+}
+
 static void nick_group(hook_user_req_t *hdata)
 {
 	user_t *u;
@@ -888,6 +901,8 @@ static void nick_ungroup(hook_user_req_t *hdata)
 
 void _modinit(module_t * m)
 {
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "transport/rfc1459");
+
 	/* Symbol relocation voodoo. */
 	server_login = &bahamut_server_login;
 	introduce_nick = &bahamut_introduce_nick;
@@ -926,6 +941,7 @@ void _modinit(module_t * m)
 
 	ircd = &Bahamut;
 
+	pcommand_add("CAPAB", m_capab, 0, MSRC_UNREG);
 	pcommand_add("PING", m_ping, 1, MSRC_USER | MSRC_SERVER);
 	pcommand_add("PONG", m_pong, 1, MSRC_SERVER);
 	pcommand_add("PRIVMSG", m_privmsg, 2, MSRC_USER);
