@@ -18,29 +18,29 @@ DECLARE_MODULE_V1("protocol/ircnet", true, _modinit, NULL, PACKAGE_STRING, "Athe
 /* *INDENT-OFF* */
 
 ircd_t IRCNet = {
-        "ircd 2.11.1p1 or later",       /* IRCd name */
-        "$$",                           /* TLD Prefix, used by Global. */
-        true,                           /* Whether or not we use IRCNet/TS6 UID */
-        false,                          /* Whether or not we use RCOMMAND */
-        false,                          /* Whether or not we support channel owners. */
-        false,                          /* Whether or not we support channel protection. */
-        false,                          /* Whether or not we support halfops. */
-	false,				/* Whether or not we use P10 */
-	false,				/* Whether or not we use vHosts. */
-	0,				/* Oper-only cmodes */
-        0,                              /* Integer flag for owner channel flag. */
-        0,                              /* Integer flag for protect channel flag. */
-        0,                              /* Integer flag for halfops. */
-        "+",                            /* Mode we set for owner. */
-        "+",                            /* Mode we set for protect. */
-        "+",                            /* Mode we set for halfops. */
-	PROTOCOL_IRCNET,		/* Protocol type */
-	0,                              /* Permanent cmodes */
-	0,                              /* Oper-immune cmode */
-	"beIR",                         /* Ban-like cmodes */
-	'e',                            /* Except mchar */
-	'I',                            /* Invex mchar */
-	IRCD_CIDR_BANS                  /* Flags */
+	.ircdname = "ircd 2.11.1p1 or later",
+	.tldprefix = "$$",
+	.uses_uid = true,
+	.uses_rcommand = false,
+	.uses_owner = false,
+	.uses_protect = false,
+	.uses_halfops = false,
+	.uses_p10 = false,
+	.uses_vhost = false,
+	.oper_only_modes = 0,
+	.owner_mode = 0,
+	.protect_mode = 0,
+	.halfops_mode = 0,
+	.owner_mchar = "+",
+	.protect_mchar = "+",
+	.halfops_mchar = "+",
+	.type = PROTOCOL_IRCNET,
+	.perm_mode = 0,
+	.oimmune_mode = 0,
+	.ban_like_modes = "beIR",
+	.except_mchar = 'e',
+	.invex_mchar = 'I',
+	.flags = IRCD_CIDR_BANS,
 };
 
 struct cmode_ ircnet_mode_list[] = {
@@ -82,7 +82,7 @@ static unsigned int ircnet_server_login(void)
 {
 	int ret;
 
-	ret = sts("PASS %s 0211010000 IRC|aDEFiIJMuw P", curr_uplink->pass);
+	ret = sts("PASS %s 0211010000 IRC|aDEFiIJMuw P", curr_uplink->send_pass);
 	if (ret == 1)
 		return 1;
 
@@ -125,18 +125,9 @@ static void ircnet_invite_sts(user_t *sender, user_t *target, channel_t *channel
 
 static void ircnet_quit_sts(user_t *u, const char *reason)
 {
-	if (!me.connected)
-		return;
-
 	sts(":%s QUIT :%s", u->nick, reason);
 }
 
-/* WALLOPS wrapper */
-static void ircnet_wallops_sts(const char *text)
-{
-	sts(":%s WALLOPS :%s", me.name, text);
-}
-
 /* join a channel */
 static void ircnet_join_sts(channel_t *c, user_t *u, bool isnew, char *modes)
 {
@@ -250,9 +241,6 @@ static void ircnet_kline_sts(const char *server, const char *user, const char *h
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	/* this won't propagate!
 	 * you'll need some bot/service on each server to do that */
 	if (irccasecmp(server, me.actual) && cnt.server > 2)
@@ -267,9 +255,6 @@ static void ircnet_unkline_sts(const char *server, const char *user, const char
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	if (irccasecmp(server, me.actual) && cnt.server > 2)
 		wallops("Missed an untkline");
 
@@ -282,8 +267,7 @@ static void ircnet_topic_sts(channel_t *c, user_t *source, const char *setter, t
 {
 	int joined = 0;
 
-	if (!me.connected || !c)
-		return;
+	return_if_fail(c != NULL);
 
 	/* Need to join to set topic -- jilles */
 	if (!chanuser_find(c, source))
@@ -302,8 +286,9 @@ static void ircnet_mode_sts(char *sender, channel_t *target, char *modes)
 {
 	user_t *u;
 
-	if (!me.connected)
-		return;
+	return_if_fail(sender != NULL);
+	return_if_fail(target != NULL);
+	return_if_fail(modes != NULL);
 
 	u = user_find(sender);
 
@@ -315,9 +300,6 @@ static void ircnet_mode_sts(char *sender, channel_t *target, char *modes)
 /* ping wrapper */
 static void ircnet_ping_sts(void)
 {
-	if (!me.connected)
-		return;
-
 	sts("PING :%s", me.name);
 }
 
@@ -342,9 +324,6 @@ static void ircnet_jupe(const char *server, const char *reason)
 	int i;
 	server_t *s;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s SQUIT %s :%s", svs != NULL ? CLIENT_NAME(svs->me) : me.actual, server, reason);
 
@@ -358,7 +337,7 @@ static void ircnet_jupe(const char *server, const char *reason)
 
 	/* dirty dirty make up some sid */
 	if (sid[0] == '\0')
-		strlcpy(sid, me.numeric, sizeof sid);
+		mowgli_strlcpy(sid, me.numeric, sizeof sid);
 	do
 	{
 		i = 3;
@@ -534,11 +513,11 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 	/* if it's only 1 then it's a nickname change */
 	else if (parc == 1)
 	{
-                if (!si->su)
-                {       
-                        slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
-                        return;
-                }
+		if (!si->su)
+		{
+			slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
+			return;
+		}
 
 		slog(LG_DEBUG, "m_nick(): nickname change from `%s': %s", si->su->nick, parv[0]);
 
@@ -727,7 +706,7 @@ static void m_join(sourceinfo_t *si, int parc, char *parv[])
 
 static void m_pass(sourceinfo_t *si, int parc, char *parv[])
 {
-	if (strcmp(curr_uplink->pass, parv[0]))
+	if (strcmp(curr_uplink->receive_pass, parv[0]))
 	{
 		slog(LG_INFO, "m_pass(): password mismatch from uplink; aborting");
 		runflags |= RF_SHUTDOWN;
@@ -746,11 +725,13 @@ static void m_motd(sourceinfo_t *si, int parc, char *parv[])
 
 void _modinit(module_t * m)
 {
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "transport/rfc1459");
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "protocol/base36uid");
+
 	/* Symbol relocation voodoo. */
 	server_login = &ircnet_server_login;
 	introduce_nick = &ircnet_introduce_nick;
 	quit_sts = &ircnet_quit_sts;
-	wallops_sts = &ircnet_wallops_sts;
 	join_sts = &ircnet_join_sts;
 	kick = &ircnet_kick;
 	msg = &ircnet_msg;
