@@ -18,29 +18,29 @@ DECLARE_MODULE_V1("protocol/nefarious", true, _modinit, NULL, PACKAGE_STRING, "A
 /* *INDENT-OFF* */
 
 ircd_t Nefarious = {
-        "Nefarious IRCU 0.4.0 or later", /* IRCd name */
-        "$",                            /* TLD Prefix, used by Global. */
-        true,                           /* Whether or not we use IRCNet/TS6 UID */
-        false,                          /* Whether or not we use RCOMMAND */
-        false,                          /* Whether or not we support channel owners. */
-        false,                          /* Whether or not we support channel protection. */
-        true,                           /* Whether or not we support halfops. */
-	true,				/* Whether or not we use P10 */
-	true,				/* Whether or not we use vhosts. */
-	CMODE_PERM|CMODE_OPERONLY|CMODE_ADMONLY, /* Oper-only cmodes */
-        0,                              /* Integer flag for owner channel flag. */
-        0,                              /* Integer flag for protect channel flag. */
-        CSTATUS_HALFOP,                   /* Integer flag for halfops. */
-        "+",                            /* Mode we set for owner. */
-        "+",                            /* Mode we set for protect. */
-        "+",                            /* Mode we set for halfops. */
-	PROTOCOL_NEFARIOUS,		/* Protocol type */
-	CMODE_PERM,                     /* Permanent cmodes */
-	0,                              /* Oper-immune cmode */
-	"be",                           /* Ban-like cmodes */
-	'e',                            /* Except mchar */
-	'e',                            /* Invex mchar (+e also exempts from +i in Nefarious) */
-	IRCD_CIDR_BANS                  /* Flags */
+	.ircdname = "Nefarious IRCU 0.4.0 or later",
+	.tldprefix = "$",
+	.uses_uid = true,
+	.uses_rcommand = false,
+	.uses_owner = false,
+	.uses_protect = false,
+	.uses_halfops = true,
+	.uses_p10 = true,
+	.uses_vhost = true,
+	.oper_only_modes = CMODE_PERM | CMODE_OPERONLY | CMODE_ADMONLY,
+	.owner_mode = 0,
+	.protect_mode = 0,
+	.halfops_mode = CSTATUS_HALFOP,
+	.owner_mchar = "+",
+	.protect_mchar = "+",
+	.halfops_mchar = "+",
+	.type = PROTOCOL_NEFARIOUS,
+	.perm_mode = CMODE_PERM,
+	.oimmune_mode = 0,
+	.ban_like_modes = "be",
+	.except_mchar = 'e',
+	.invex_mchar = 'e',
+	.flags = IRCD_CIDR_BANS,
 };
 
 struct cmode_ nefarious_mode_list[] = {
@@ -73,16 +73,16 @@ struct extmode nefarious_ignore_mode_list[] = {
 };
 
 struct cmode_ nefarious_status_mode_list[] = {
-  { 'o', CSTATUS_OP     },
+  { 'o', CSTATUS_OP	},
   { 'h', CSTATUS_HALFOP },
-  { 'v', CSTATUS_VOICE  },
+  { 'v', CSTATUS_VOICE	},
   { '\0', 0 }
 };
 
 struct cmode_ nefarious_prefix_mode_list[] = {
-  { '@', CSTATUS_OP     },
+  { '@', CSTATUS_OP	},
   { '%', CSTATUS_HALFOP },
-  { '+', CSTATUS_VOICE  },
+  { '+', CSTATUS_VOICE	},
   { '\0', 0 }
 };
 
@@ -133,8 +133,7 @@ static void nefarious_notice_channel_sts(user_t *from, channel_t *target, const
 /* topic wrapper */
 static void nefarious_topic_sts(channel_t *c, user_t *source, const char *setter, time_t ts, time_t prevts, const char *topic)
 {
-	if (!me.connected || !c)
-		return;
+	return_if_fail(c != NULL);
 
 	/* for nefarious, set topicsetter iff we can set the proper topicTS */
 	if (ts > prevts || prevts == 0)
@@ -149,22 +148,11 @@ static void nefarious_topic_sts(channel_t *c, user_t *source, const char *setter
 	}
 }
 
-/* mode wrapper */
-static void nefarious_mode_sts(char *sender, channel_t *target, char *modes)
-{
-	user_t *fptr = user_find_named(sender);
-
-	if (!fptr)
-		return;
-
-	sts("%s M %s %s", fptr->uid, target->name, modes);
-}
-
 /* protocol-specific stuff to do on login */
 static void nefarious_on_login(user_t *u, myuser_t *mu, const char *wantedhost)
 {
-	if (!me.connected || u == NULL)
-		return;
+	return_if_fail(u != NULL);
+	return_if_fail(mu != NULL);
 
 	sts("%s AC %s R %s %lu", me.numeric, u->uid, entity(mu)->name,
 			(unsigned long)mu->registered);
@@ -172,13 +160,12 @@ static void nefarious_on_login(user_t *u, myuser_t *mu, const char *wantedhost)
 }
 
 /* P10 does not support logout, so kill the user
- * we can't keep track of which logins are stale and which aren't -- jilles 
+ * we can't keep track of which logins are stale and which aren't -- jilles
  * Except we can in Nefarious --nenolod
  */
 static bool nefarious_on_logout(user_t *u, const char *account)
 {
-	if (!me.connected || u == NULL)
-		return false;
+	return_val_if_fail(u != NULL, false);
 
 	sts("%s AC %s U", me.numeric, u->uid);
 	if (u->flags & UF_HIDEHOSTREQ && me.hidehostsuffix != NULL &&
@@ -186,12 +173,25 @@ static bool nefarious_on_logout(user_t *u, const char *account)
 	{
 		slog(LG_DEBUG, "nefarious_on_logout(): removing +x vhost for %s: %s -> %s",
 				u->nick, u->vhost, u->host);
-		strlcpy(u->vhost, u->host, sizeof u->vhost);
+
+		strshare_unref(u->vhost);
+		u->vhost = strshare_get(u->host);
 	}
 
 	return false;
 }
 
+static void nefarious_sasl_sts(char *target, char mode, char *data)
+{
+	sts("%s SASL %c%c %s %c %s", me.numeric, target[0], target[1], target, mode, data);
+}
+
+static void nefarious_svslogin_sts(char *target, char *nick, char *user, char *host, myuser_t *account)
+{
+	sts("%s SASL %c%c %s L %s %lu", me.numeric, target[0], target[1], target,
+			entity(account)->name, (unsigned long)account->registered);
+}
+
 static void nefarious_sethost_sts(user_t *source, user_t *target, const char *host)
 {
 	sts("%s FA %s %s", me.numeric, target->uid, host);
@@ -200,10 +200,15 @@ static void nefarious_sethost_sts(user_t *source, user_t *target, const char *ho
 		sts("%s M %s +x", me.numeric, target->uid);
 }
 
+static void nefarious_quarantine_sts(user_t *source, user_t *victim, long duration, const char *reason)
+{
+	sts("%s SU * +*@%s %lu :%s", me.numeric, victim->host, CURRTIME + duration, reason);
+}
+
 static void m_topic(sourceinfo_t *si, int parc, char *parv[])
 {
 	channel_t *c = channel_find(parv[0]);
-	char *source;
+	const char *source;
 	time_t ts = 0;
 
 	if (!c)
@@ -337,8 +342,8 @@ static void m_burst(sourceinfo_t *si, int parc, char *parv[])
 						p++;
 					}
 				}
-				strlcpy(newnick, prefix, sizeof newnick);
-				strlcat(newnick, userv[i], sizeof newnick);
+				mowgli_strlcpy(newnick, prefix, sizeof newnick);
+				mowgli_strlcat(newnick, userv[i], sizeof newnick);
 				chanuser_add(c, newnick);
 			}
 		}
@@ -386,20 +391,33 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 			{
 				p = strchr(parv[5+i], '@');
 				if (p == NULL)
-					strlcpy(u->vhost, parv[5+i], sizeof u->vhost);
+				{
+					strshare_unref(u->vhost);
+					u->vhost = strshare_get(parv[5 + i]);
+				}
 				else
 				{
-					strlcpy(u->vhost, p + 1, sizeof u->vhost);
-					strlcpy(u->user, parv[5+i], sizeof u->user);
-					p = strchr(u->user, '@');
+					char userbuf[USERLEN];
+
+					strshare_unref(u->vhost);
+					u->vhost = strshare_get(p + 1);
+
+					mowgli_strlcpy(userbuf, parv[5+i], sizeof userbuf);
+
+					p = strchr(userbuf, '@');
 					if (p != NULL)
 						*p = '\0';
+
+					strshare_unref(u->user);
+					u->user = strshare_get(userbuf);
 				}
 				i++;
 			}
 			if (strchr(parv[5], 'f'))
 			{
-				strlcpy(u->vhost, parv[5+i], sizeof u->vhost);
+				strshare_unref(u->vhost);
+				u->vhost = strshare_get(parv[5 + i]);
+
 				i++;
 			}
 			if (strchr(parv[5], 'x'))
@@ -415,11 +433,11 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 	/* if it's only 2 then it's a nickname change */
 	else if (parc == 2)
 	{
-                if (!si->su)
-                {       
-                        slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
-                        return;
-                }
+		if (!si->su)
+		{
+			slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
+			return;
+		}
 
 		slog(LG_DEBUG, "m_nick(): nickname change from `%s': %s", si->su->nick, parv[0]);
 
@@ -466,14 +484,24 @@ static void m_mode(sourceinfo_t *si, int parc, char *parv[])
 				/* assume +h */
 				p = strchr(parv[2], '@');
 				if (p == NULL)
-					strlcpy(u->vhost, parv[2], sizeof u->vhost);
+				{
+					strshare_unref(u->vhost);
+					u->vhost = strshare_get(parv[2]);
+				}
 				else
 				{
-					strlcpy(u->vhost, p + 1, sizeof u->vhost);
-					strlcpy(u->user, parv[2], sizeof u->user);
-					p = strchr(u->user, '@');
+					char userbuf[USERLEN];
+
+					strshare_unref(u->vhost);
+					u->vhost = strshare_get(p + 1);
+
+					mowgli_strlcpy(userbuf, parv[2], sizeof userbuf);
+					p = strchr(userbuf, '@');
 					if (p != NULL)
 						*p = '\0';
+
+					strshare_unref(u->user);
+					u->user = strshare_get(userbuf);
 				}
 				slog(LG_DEBUG, "m_mode(): user %s setting vhost %s@%s", u->nick, u->user, u->vhost);
 			}
@@ -482,7 +510,10 @@ static void m_mode(sourceinfo_t *si, int parc, char *parv[])
 				/* must be -h */
 				/* XXX we don't know the original ident */
 				slog(LG_DEBUG, "m_mode(): user %s turning off vhost", u->nick);
-				strlcpy(u->vhost, u->host, sizeof u->vhost);
+
+				strshare_unref(u->vhost);
+				u->vhost = strshare_get(u->host);
+
 				/* revert to +x vhost if applicable */
 				check_hidehost(u);
 			}
@@ -611,9 +642,26 @@ static void m_account(sourceinfo_t *si, int parc, char *parv[])
 	}
 }
 
+static void m_sasl(sourceinfo_t *si, int parc, char *parv[])
+{
+	sasl_message_t smsg;
+
+	if (parc < 4)
+		return;
+
+	smsg.uid = parv[1];
+	smsg.mode = *parv[2];
+	smsg.buf = parv[3];
+	smsg.ext = parc >= 4 ? parv[4] : NULL;
+	smsg.server = si->s ? si->s : NULL;
+
+	hook_call_sasl_input(&smsg);
+}
+
 static void check_hidehost(user_t *u)
 {
 	static bool warned = false;
+	char buf[HOSTLEN + 1];
 
 	/* do they qualify? */
 	if (!(u->flags & UF_HIDEHOSTREQ) || u->myuser == NULL || (u->myuser->flags & MU_WAITAUTH))
@@ -633,8 +681,12 @@ static void check_hidehost(user_t *u)
 		}
 		return;
 	}
-	snprintf(u->vhost, sizeof u->vhost, "%s.%s", entity(u->myuser)->name,
-			me.hidehostsuffix);
+
+	snprintf(buf, sizeof buf, "%s.%s", entity(u->myuser)->name, me.hidehostsuffix);
+
+	strshare_unref(u->vhost);
+	u->vhost = strshare_get(buf);
+
 	slog(LG_DEBUG, "check_hidehost(): %s -> %s", u->nick, u->vhost);
 }
 
@@ -647,10 +699,12 @@ void _modinit(module_t * m)
 	kick = &nefarious_kick;
 	notice_channel_sts = &nefarious_notice_channel_sts;
 	topic_sts = &nefarious_topic_sts;
-	mode_sts = &nefarious_mode_sts;
 	ircd_on_login = &nefarious_on_login;
 	ircd_on_logout = &nefarious_on_logout;
+	sasl_sts = &nefarious_sasl_sts;
 	sethost_sts = &nefarious_sethost_sts;
+	svslogin_sts = &nefarious_svslogin_sts;
+	quarantine_sts = &nefarious_quarantine_sts;
 
 	mode_list = nefarious_mode_list;
 	ignore_mode_list = nefarious_ignore_mode_list;
@@ -661,6 +715,8 @@ void _modinit(module_t * m)
 
 	ircd = &Nefarious;
 
+	pcommand_add("SASL", m_sasl, 4, MSRC_SERVER);
+
 	/* override these */
 	pcommand_delete("B");
 	pcommand_delete("N");
