@@ -21,7 +21,7 @@ static unsigned int p10_server_login(void)
 {
 	int ret;
 
-	ret = sts("PASS :%s", curr_uplink->pass);
+	ret = sts("PASS :%s", curr_uplink->send_pass);
 	if (ret == 1)
 		return 1;
 
@@ -54,9 +54,6 @@ static void p10_invite_sts(user_t *sender, user_t *target, channel_t *channel)
 
 static void p10_quit_sts(user_t *u, const char *reason)
 {
-	if (!me.connected)
-		return;
-
 	sts("%s Q :%s", u->uid, reason);
 }
 
@@ -204,9 +201,6 @@ static void p10_part_sts(channel_t *c, user_t *u)
 /* server-to-server KLINE wrapper */
 static void p10_kline_sts(const char *server, const char *user, const char *host, long duration, const char *reason)
 {
-	if (!me.connected)
-		return;
-
 	/* hold permanent akills for four weeks -- jilles */
 	sts("%s GL * +%s@%s %ld :%s", me.numeric, user, host, duration > 0 ? duration : 2419200, reason);
 }
@@ -214,35 +208,23 @@ static void p10_kline_sts(const char *server, const char *user, const char *host
 /* server-to-server UNKLINE wrapper */
 static void p10_unkline_sts(const char *server, const char *user, const char *host)
 {
-	if (!me.connected)
-		return;
-
 	sts("%s GL * -%s@%s", me.numeric, user, host);
 }
 
 static void p10_xline_sts(const char *server, const char *realname, long duration, const char *reason)
 {
-	if (!me.connected)
-		return;
-
 	/* hold permanent sglines for four weeks -- jilles */
 	sts("%s GL * +$R%s %ld :%s", me.numeric, realname, duration > 0 ? duration : 2419200, reason);
 }
 
 static void p10_unxline_sts(const char *server, const char *realname)
 {
-	if (!me.connected)
-		return;
-
 	sts("%s GL * -$R%s", me.numeric, realname);
 }
 
 static void p10_qline_sts(const char *server, const char *name, long duration, const char *reason)
 {
-	if (!me.connected)
-		return;
-
-	if (*name != '#' && *name != '&')
+	if (!VALID_CHANNEL_PFX(name))
 	{
 		slog(LG_INFO, "SQLINE: Could not set SQLINE on \2%s\2, not supported by ircu.", name);
 		return;
@@ -254,10 +236,7 @@ static void p10_qline_sts(const char *server, const char *name, long duration, c
 
 static void p10_unqline_sts(const char *server, const char *name)
 {
-	if (!me.connected)
-		return;
-
-	if (*name != '#' && *name != '&')
+	if (!VALID_CHANNEL_PFX(name))
 	{
 		slog(LG_INFO, "SQLINE: Could not remove SQLINE on \2%s\2, not supported by ircu.", name);
 		return;
@@ -269,9 +248,6 @@ static void p10_unqline_sts(const char *server, const char *name)
 /* topic wrapper */
 static void p10_topic_sts(channel_t *c, user_t *source, const char *setter, time_t ts, time_t prevts, const char *topic)
 {
-	if (!me.connected || !c)
-		return;
-
 	if (ts > prevts || prevts == 0)
 		sts("%s T %s %lu %lu :%s", source->uid, c->name, (unsigned long)c->ts, (unsigned long)ts, topic);
 	else
@@ -287,10 +263,15 @@ static void p10_topic_sts(channel_t *c, user_t *source, const char *setter, time
 /* mode wrapper */
 static void p10_mode_sts(char *sender, channel_t *target, char *modes)
 {
-	user_t *fptr = user_find_named(sender);
+	user_t *fptr;
 
-	if (!fptr)
-		return;
+	return_if_fail(sender != NULL);
+	return_if_fail(target != NULL);
+	return_if_fail(modes != NULL);
+
+	fptr = user_find_named(sender);
+
+	return_if_fail(fptr != NULL);
 
 	if (chanuser_find(target, fptr))
 		sts("%s M %s %s", fptr->uid, target->name, modes);
@@ -301,17 +282,13 @@ static void p10_mode_sts(char *sender, channel_t *target, char *modes)
 /* ping wrapper */
 static void p10_ping_sts(void)
 {
-	if (!me.connected)
-		return;
-
 	sts("%s G !%lu %s %lu", me.numeric, (unsigned long)CURRTIME, me.name, (unsigned long)CURRTIME);
 }
 
 /* protocol-specific stuff to do on login */
 static void p10_on_login(user_t *u, myuser_t *mu, const char *wantedhost)
 {
-	if (!me.connected || u == NULL)
-		return;
+	return_if_fail(u != NULL);
 
 	sts("%s AC %s %s %lu", me.numeric, u->uid, entity(mu)->name,
 			(unsigned long)mu->registered);
@@ -322,25 +299,16 @@ static void p10_on_login(user_t *u, myuser_t *mu, const char *wantedhost)
  * we can't keep track of which logins are stale and which aren't -- jilles */
 static bool p10_on_logout(user_t *u, const char *account)
 {
-	if (!me.connected || u == NULL)
-		return false;
+	return_val_if_fail(u != NULL, false);
 
-	if (u != NULL)
-	{
-		kill_user(NULL, u, "Forcing logout %s -> %s", u->nick, account);
-		return true;
-	}
-	else
-		return false;
+	kill_user(NULL, u, "Forcing logout %s -> %s", u->nick, account);
+	return true;
 }
 
 static void p10_jupe(const char *server, const char *reason)
 {
 	server_t *s;
 
-	if (!me.connected)
-		return;
-
 	/* hold it for a day (arbitrary) -- jilles */
 	/* get rid of local deactivation too */
 	s = server_find(server);
@@ -349,10 +317,20 @@ static void p10_jupe(const char *server, const char *reason)
 	sts("%s JU * +%s %d %lu :%s", me.numeric, server, 86400, (unsigned long)CURRTIME, reason);
 }
 
+static void p10_sasl_sts(char *target, char mode, char *data)
+{
+	sts("%s XR %c%c %s :SASL:%c:%s", me.numeric, target[0], target[1], target, mode, data);
+}
+
+static void p10_svslogin_sts(char *target, char *nick, char *user, char *host, myuser_t *account)
+{
+	sts("%s XR %c%c %s :SASL:L:%s", me.numeric, target[0], target[1], target, entity(account)->name);
+}
+
 static void m_topic(sourceinfo_t *si, int parc, char *parv[])
 {
 	channel_t *c = channel_find(parv[0]);
-	char *source;
+	const char *source;
 	time_t ts = 0;
 
 	if (!c)
@@ -451,7 +429,7 @@ static void m_create(sourceinfo_t *si, int parc, char *parv[])
 		else
 			buf[0] = '\0';
 
-		strlcat(buf, si->su->uid, BUFSIZE);
+		mowgli_strlcat(buf, si->su->uid, BUFSIZE);
 
 		chanuser_add(c, buf);
 	}
@@ -605,8 +583,8 @@ static void m_burst(sourceinfo_t *si, int parc, char *parv[])
 							p++;
 						}
 				}
-				strlcpy(newnick, prefix, sizeof newnick);
-				strlcat(newnick, userv[i], sizeof newnick);
+				mowgli_strlcpy(newnick, prefix, sizeof newnick);
+				mowgli_strlcat(newnick, userv[i], sizeof newnick);
 				chanuser_add(c, newnick);
 			}
 		}
@@ -675,12 +653,12 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 	/* if it's only 2 then it's a nickname change */
 	else if (parc == 2)
 	{
-                if (!si->su)
-                {       
-                        slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
-                        return;
-                }
-                
+		if (!si->su)
+		{
+			slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
+			return;
+		}
+
 		slog(LG_DEBUG, "m_nick(): nickname change from `%s': %s", si->su->nick, parv[0]);
 
 		if (user_changenick(si->su, parv[0], atoi(parv[1])))
@@ -935,7 +913,7 @@ static void m_away(sourceinfo_t *si, int parc, char *parv[])
 
 static void m_pass(sourceinfo_t *si, int parc, char *parv[])
 {
-	if (strcmp(curr_uplink->pass, parv[0]))
+	if (strcmp(curr_uplink->receive_pass, parv[0]))
 	{
 		slog(LG_INFO, "m_pass(): password mismatch from uplink; aborting");
 		runflags |= RF_SHUTDOWN;
@@ -969,6 +947,7 @@ static void m_account(sourceinfo_t *si, int parc, char *parv[])
 static void check_hidehost(user_t *u)
 {
 	static bool warned = false;
+	char buf[HOSTLEN + 1];
 
 	/* do they qualify? */
 	if (!(u->flags & UF_HIDEHOSTREQ) || u->myuser == NULL || (u->myuser->flags & MU_WAITAUTH))
@@ -988,13 +967,20 @@ static void check_hidehost(user_t *u)
 		}
 		return;
 	}
-	snprintf(u->vhost, sizeof u->vhost, "%s.%s", entity(u->myuser)->name,
-			me.hidehostsuffix);
+
+	snprintf(buf, sizeof buf, "%s.%s", entity(u->myuser)->name, me.hidehostsuffix);
+
+	strshare_unref(u->vhost);
+	u->vhost = strshare_get(buf);
+
 	slog(LG_DEBUG, "check_hidehost(): %s -> %s", u->nick, u->vhost);
 }
 
 void _modinit(module_t * m)
 {
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "transport/p10");
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "protocol/base36uid");
+
 	/* Symbol relocation voodoo. */
 	server_login = &p10_server_login;
 	introduce_nick = &p10_introduce_nick;
@@ -1025,8 +1011,8 @@ void _modinit(module_t * m)
 	ircd_on_logout = &p10_on_logout;
 	jupe = &p10_jupe;
 	invite_sts = &p10_invite_sts;
-
-	parse = &p10_parse;
+	sasl_sts = &p10_sasl_sts;
+	svslogin_sts = &p10_svslogin_sts;
 
 	pcommand_add("G", m_ping, 1, MSRC_USER | MSRC_SERVER);
 	pcommand_add("Z", m_pong, 1, MSRC_SERVER);
