@@ -9,7 +9,7 @@
 /*
  * protocol/ts6-generic
  *
- * This module is not very useful on it's own.  It implements
+ * This module is not very useful on it's own.	It implements
  * the basis of the TS6 extended linking profile, which is used
  * by hybrid, charybdis, ratbox (--enable-services), weircd
  * and others.
@@ -71,12 +71,12 @@ static unsigned int ts6_server_login(void)
 	if (!me.numeric)
 	{
 		ircd->uses_uid = false;
-		ret = sts("PASS %s :TS", curr_uplink->pass);
+		ret = sts("PASS %s :TS", curr_uplink->send_pass);
 	}
-	else if (strlen(me.numeric) == 3 && isdigit(*me.numeric))
+	else if (strlen(me.numeric) == 3 && isdigit((unsigned char)*me.numeric))
 	{
 		ircd->uses_uid = true;
-		ret = sts("PASS %s TS 6 :%s", curr_uplink->pass, me.numeric);
+		ret = sts("PASS %s TS 6 :%s", curr_uplink->send_pass, me.numeric);
 	}
 	else
 	{
@@ -88,7 +88,7 @@ static unsigned int ts6_server_login(void)
 	me.bursting = true;
 
 	sts("CAPAB :QS EX IE KLN UNKLN ENCAP TB SERVICES EUID EOPMOD MLOCK");
-	sts("SERVER %s 1 :%s", me.name, me.desc);
+	sts("SERVER %s 1 :%s%s", me.name, me.hidden ? "(H) " : "", me.desc);
 	sts("SVINFO %d 3 0 :%lu", ircd->uses_uid ? 6 : 5,
 			(unsigned long)CURRTIME);
 
@@ -101,11 +101,11 @@ static void ts6_introduce_nick(user_t *u)
 	const char *umode = user_get_umodestr(u);
 
 	if (ircd->uses_uid && use_euid)
-		sts(":%s EUID %s 1 %lu %sS %s %s 0 %s * * :%s", me.numeric, u->nick, (unsigned long)u->ts, umode, u->user, u->host, u->uid, u->gecos);
+		sts(":%s EUID %s 1 %lu %s %s %s 0 %s * * :%s", me.numeric, u->nick, (unsigned long)u->ts, umode, u->user, u->host, u->uid, u->gecos);
 	else if (ircd->uses_uid)
-		sts(":%s UID %s 1 %lu %sS %s %s 0 %s :%s", me.numeric, u->nick, (unsigned long)u->ts, umode, u->user, u->host, u->uid, u->gecos);
+		sts(":%s UID %s 1 %lu %s %s %s 0 %s :%s", me.numeric, u->nick, (unsigned long)u->ts, umode, u->user, u->host, u->uid, u->gecos);
 	else
-		sts("NICK %s 1 %lu %sS %s %s %s :%s", u->nick, (unsigned long)u->ts, umode, u->user, u->host, me.name, u->gecos);
+		sts("NICK %s 1 %lu %s %s %s %s :%s", u->nick, (unsigned long)u->ts, umode, u->user, u->host, me.name, u->gecos);
 }
 
 /* invite a user to a channel */
@@ -116,9 +116,6 @@ static void ts6_invite_sts(user_t *sender, user_t *target, channel_t *channel)
 
 static void ts6_quit_sts(user_t *u, const char *reason)
 {
-	if (!me.connected)
-		return;
-
 	sts(":%s QUIT :%s", CLIENT_NAME(u), reason);
 }
 
@@ -179,7 +176,7 @@ static void ts6_msg(const char *from, const char *target, const char *fmt, ...)
 	 * is on it -- jilles
 	 *
 	 * Well, now it's operserv, but yes it's safe to assume that
-	 * the source would be able to send to whatever target it is 
+	 * the source would be able to send to whatever target it is
 	 * sending to. --nenolod
 	 */
 	sts(":%s PRIVMSG %s :%s", CLIENT_NAME(u), t ? CLIENT_NAME(t) : target, buf);
@@ -275,9 +272,6 @@ static void ts6_kline_sts(const char *server, const char *user, const char *host
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s ENCAP %s KLINE %ld %s %s :%s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, duration, user, host, reason);
 }
@@ -287,9 +281,6 @@ static void ts6_unkline_sts(const char *server, const char *user, const char *ho
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s ENCAP %s UNKLINE %s %s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, user, host);
 }
@@ -299,9 +290,6 @@ static void ts6_xline_sts(const char *server, const char *realname, long duratio
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s ENCAP %s XLINE %ld %s 2 :%s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, duration, realname, reason);
 }
@@ -311,9 +299,6 @@ static void ts6_unxline_sts(const char *server, const char *realname)
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s ENCAP %s UNXLINE %s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, realname);
 }
@@ -323,9 +308,6 @@ static void ts6_qline_sts(const char *server, const char *name, long duration, c
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s ENCAP %s RESV %ld %s 0 :%s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, duration, name, reason);
 }
@@ -335,20 +317,35 @@ static void ts6_unqline_sts(const char *server, const char *name)
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
 	sts(":%s ENCAP %s UNRESV %s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, name);
 }
 
+/* server-to-server DLINE wrapper */
+static void ts6_dline_sts(const char *server, const char *host, long duration, const char *reason)
+{
+	service_t *svs;
+
+	svs = service_find("operserv");
+	sts(":%s ENCAP %s DLINE %ld %s :%s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, duration, host, reason);
+}
+
+/* server-to-server UNDLINE wrapper */
+static void ts6_undline_sts(const char *server, const char *host)
+{
+	service_t *svs;
+
+	svs = service_find("operserv");
+	sts(":%s ENCAP %s UNDLINE %s", svs != NULL ? CLIENT_NAME(svs->me) : ME, server, host);
+}
+
 /* topic wrapper */
 static void ts6_topic_sts(channel_t *c, user_t *source, const char *setter, time_t ts, time_t prevts, const char *topic)
 {
 	int joined = 0;
 
-	if (!me.connected || !c)
-		return;
+	return_if_fail(c != NULL);
+	return_if_fail(source != NULL);
 
 	/* If possible, try to use ETB */
 	if (use_eopmod && (c->ts > 0 || ts > prevts))
@@ -404,10 +401,15 @@ static void ts6_topic_sts(channel_t *c, user_t *source, const char *setter, time
 /* mode wrapper */
 static void ts6_mode_sts(char *sender, channel_t *target, char *modes)
 {
-	user_t *u = user_find(sender);
+	user_t *u;
 
-	if (!me.connected || !u)
-		return;
+	return_if_fail(sender != NULL);
+	return_if_fail(target != NULL);
+	return_if_fail(modes != NULL);
+
+	u = user_find(sender);
+
+	return_if_fail(u != NULL);
 
 	if (ircd->uses_uid)
 		sts(":%s TMODE %lu %s %s", CLIENT_NAME(u), (unsigned long)target->ts, target->name, modes);
@@ -418,42 +420,40 @@ static void ts6_mode_sts(char *sender, channel_t *target, char *modes)
 /* ping wrapper */
 static void ts6_ping_sts(void)
 {
-	if (!me.connected)
-		return;
-
 	sts("PING :%s", me.name);
 }
 
 /* protocol-specific stuff to do on login */
 static void ts6_on_login(user_t *u, myuser_t *mu, const char *wantedhost)
 {
-	if (!me.connected || !use_rserv_support || u == NULL)
+	if (!use_rserv_support)
 		return;
 
+	return_if_fail(u != NULL);
+
 	sts(":%s ENCAP * SU %s %s", ME, CLIENT_NAME(u), entity(mu)->name);
 }
 
 /* protocol-specific stuff to do on login */
 static bool ts6_on_logout(user_t *u, const char *account)
 {
-	if (!me.connected || !use_rserv_support || u == NULL)
+	if (!use_rserv_support)
 		return false;
 
+	return_val_if_fail(u != NULL, false);
+
 	sts(":%s ENCAP * SU %s", ME, CLIENT_NAME(u));
 	return false;
 }
 
 /* XXX we don't have an appropriate API for this, what about making JUPE
  * serverside like in P10?
- *       --nenolod
+ *	 --nenolod
  */
 static void ts6_jupe(const char *server, const char *reason)
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	server_delete(server);
 
 	svs = service_find("operserv");
@@ -469,7 +469,7 @@ static void ts6_sethost_sts(user_t *source, user_t *target, const char *host)
 		sts(":%s ENCAP * CHGHOST %s :%s", ME, target->nick, host);
 }
 
-static void ts6_fnc_sts(user_t *source, user_t *u, char *newnick, int type)
+static void ts6_fnc_sts(user_t *source, user_t *u, const char *newnick, int type)
 {
 	/* XXX assumes the server will accept this -- jilles */
 	sts(":%s ENCAP %s RSFNC %s %s %lu %lu", ME,
@@ -479,7 +479,7 @@ static void ts6_fnc_sts(user_t *source, user_t *u, char *newnick, int type)
 			(unsigned long)u->ts);
 }
 
-static void ts6_svslogin_sts(char *target, char *nick, char *user, char *host, char *login)
+static void ts6_svslogin_sts(char *target, char *nick, char *user, char *host, myuser_t *account)
 {
 	user_t *tu = user_find(target);
 	server_t *s;
@@ -492,7 +492,7 @@ static void ts6_svslogin_sts(char *target, char *nick, char *user, char *host, c
 		return;
 
 	sts(":%s ENCAP %s SVSLOGIN %s %s %s %s %s", ME, s->name,
-			target, nick, user, host, login);
+			target, nick, user, host, entity(account)->name);
 }
 
 static void ts6_sasl_sts(char *target, char mode, char *data)
@@ -534,7 +534,7 @@ static void ts6_holdnick_sts(user_t *source, int duration, const char *nick, myu
 
 static void ts6_mlock_sts(channel_t *c)
 {
-	mychan_t *mc = mychan_find(c->name);
+	mychan_t *mc = mychan_from(c);
 
 	if (use_mlock == false)
 		return;
@@ -559,7 +559,7 @@ static void m_mlock(sourceinfo_t *si, int parc, char *parv[])
 	if (!(c = channel_find(parv[1])))
 		return;
 
-	if (!(mc = mychan_find(c->name)))
+	if (!(mc = mychan_from(c)))
 	{
 		/* Unregistered channel. Clear the MLOCK. */
 		sts(":%s MLOCK %lu %s :", ME, (unsigned long)c->ts, c->name);
@@ -647,7 +647,7 @@ static void m_pong(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	handle_eob(s);
 
-	if (irccasecmp(me.actual, parv[0]))
+	if (s != si->s)
 		return;
 
 	me.uplinkpong = CURRTIME;
@@ -868,7 +868,7 @@ static void m_bmask(sourceinfo_t *si, int parc, char *parv[])
 
 	if (atol(parv[0]) > c->ts)
 		return;
-	
+
 	type = *parv[2];
 	if (!strchr(ircd->ban_like_modes, type))
 	{
@@ -931,11 +931,11 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 	/* if it's only 2 then it's a nickname change */
 	else if (parc == 2)
 	{
-                if (!si->su)
-                {       
-                        slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
-                        return;
-                }
+		if (!si->su)
+		{
+			slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
+			return;
+		}
 
 		slog(LG_DEBUG, "m_nick(): nickname change from `%s': %s", si->su->nick, parv[0]);
 
@@ -1195,14 +1195,14 @@ static void m_pass(sourceinfo_t *si, int parc, char *parv[])
 {
 	/* TS5: PASS mypassword :TS
 	 * TS6: PASS mypassword TS 6 :sid */
-	if (strcmp(curr_uplink->pass, parv[0]))
+	if (strcmp(curr_uplink->receive_pass, parv[0]))
 	{
 		slog(LG_INFO, "m_pass(): password mismatch from uplink; aborting");
 		runflags |= RF_SHUTDOWN;
 	}
 
 	if (ircd->uses_uid && parc > 3 && atoi(parv[2]) >= 6)
-		strlcpy(ts6sid, parv[3], sizeof(ts6sid));
+		mowgli_strlcpy(ts6sid, parv[3], sizeof(ts6sid));
 	else
 	{
 		if (ircd->uses_uid)
@@ -1266,7 +1266,9 @@ static void m_encap(sourceinfo_t *si, int parc, char *parv[])
 		u = si->su;
 		if (u == NULL)
 			return;
-		strlcpy(u->host, parv[2], HOSTLEN);
+
+		strshare_unref(u->host);
+		u->host = strshare_get(parv[2]);
 	}
 	else if (!irccasecmp(parv[1], "CHGHOST"))
 	{
@@ -1275,7 +1277,10 @@ static void m_encap(sourceinfo_t *si, int parc, char *parv[])
 		u = user_find(parv[2]);
 		if (u == NULL)
 			return;
-		strlcpy(u->vhost, parv[3], HOSTLEN);
+
+		strshare_unref(u->vhost);
+		u->vhost = strshare_get(parv[3]);
+
 		slog(LG_DEBUG, "m_encap(): chghost %s -> %s", u->nick,
 				u->vhost);
 	}
@@ -1291,6 +1296,7 @@ static void m_encap(sourceinfo_t *si, int parc, char *parv[])
 		smsg.mode = *parv[4];
 		smsg.buf = parv[5];
 		smsg.ext = parc >= 6 ? parv[6] : NULL;
+		smsg.server = si->s ? si->s : NULL;
 		hook_call_sasl_input(&smsg);
 	}
 	else if (!irccasecmp(parv[1], "RSMSG"))
@@ -1305,8 +1311,8 @@ static void m_encap(sourceinfo_t *si, int parc, char *parv[])
 		for (i = 3; i < parc; i++)
 		{
 			if (i > 3)
-				strlcat(buf, " ", sizeof buf);
-			strlcat(buf, parv[i], sizeof buf);
+				mowgli_strlcat(buf, " ", sizeof buf);
+			mowgli_strlcat(buf, parv[i], sizeof buf);
 		}
 		snprintf(dest, sizeof dest, "%s@%s", parv[2], me.name);
 		handle_message(si, dest, false, buf);
@@ -1338,10 +1344,12 @@ static void m_signon(sourceinfo_t *si, int parc, char *parv[])
 	handle_nickchange(u); /* If they're logging out, this will bug them about identifying. Or something. */
 
 	/* USER */
-	strlcpy(u->user, parv[1], USERLEN);
+	strshare_unref(u->user);
+	u->user = strshare_get(parv[1]);
 
 	/* HOST */
-	strlcpy(u->vhost, parv[2], HOSTLEN);
+	strshare_unref(u->vhost);
+	u->vhost = strshare_get(parv[2]);
 
 	/* LOGIN */
 	if(*parv[4] == '*') /* explicitly unchanged */
@@ -1392,7 +1400,7 @@ static void m_capab(sourceinfo_t *si, int parc, char *parv[])
 
 	/* Now we know whether or not we should enable services support,
 	 * so burst the clients.
-	 *       --nenolod
+	 *	 --nenolod
 	 */
 	services_init();
 }
@@ -1404,7 +1412,8 @@ static void m_chghost(sourceinfo_t *si, int parc, char *parv[])
 	if (!u)
 		return;
 
-	strlcpy(u->vhost, parv[1], HOSTLEN);
+	strshare_unref(u->vhost);
+	u->vhost = strshare_get(parv[1]);
 }
 
 static void m_motd(sourceinfo_t *si, int parc, char *parv[])
@@ -1439,12 +1448,15 @@ static void channel_drop(mychan_t *mc)
 static server_t *sid_find(char *name)
 {
 	char sid[4];
-	strlcpy(sid, name, 4);
+	mowgli_strlcpy(sid, name, 4);
 	return server_find(sid);
 }
 
 void _modinit(module_t * m)
 {
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "transport/rfc1459");
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "protocol/base36uid");
+
 	/* Symbol relocation voodoo. */
 	server_login = &ts6_server_login;
 	introduce_nick = &ts6_introduce_nick;
@@ -1482,6 +1494,8 @@ void _modinit(module_t * m)
 	sasl_sts = &ts6_sasl_sts;
 	is_valid_host = &ts6_is_valid_host;
 	mlock_sts = &ts6_mlock_sts;
+	dline_sts = &ts6_dline_sts;
+	undline_sts = &ts6_undline_sts;
 
 	pcommand_add("PING", m_ping, 1, MSRC_USER | MSRC_SERVER);
 	pcommand_add("PONG", m_pong, 1, MSRC_SERVER);
