@@ -14,57 +14,62 @@
 
 DECLARE_MODULE_V1("protocol/unreal", true, _modinit, NULL, PACKAGE_STRING, "Atheme Development Group <http://www.atheme.org>");
 
+static bool has_protoctl = false;
 static bool use_esvid = false;
+static bool use_mlock = false;
+static char ts6sid[3 + 1] = "";
 
 /* *INDENT-OFF* */
 
 ircd_t Unreal = {
-        "UnrealIRCd 3.1 or later",      /* IRCd name */
-        "$",                            /* TLD Prefix, used by Global. */
-        false,                          /* Whether or not we use IRCNet/TS6 UID */
-        false,                          /* Whether or not we use RCOMMAND */
-        true,                           /* Whether or not we support channel owners. */
-        true,                           /* Whether or not we support channel protection. */
-        true,                           /* Whether or not we support halfops. */
-	false,				/* Whether or not we use P10 */
-	true,				/* Whether or not we use vHosts. */
-	CMODE_OPERONLY | CMODE_ADMONLY, /* Oper-only cmodes */
-        CSTATUS_OWNER,                    /* Integer flag for owner channel flag. */
-        CSTATUS_PROTECT,                  /* Integer flag for protect channel flag. */
-        CSTATUS_HALFOP,                   /* Integer flag for halfops. */
-        "+q",                           /* Mode we set for owner. */
-        "+a",                           /* Mode we set for protect. */
-        "+h",                           /* Mode we set for halfops. */
-	PROTOCOL_UNREAL,		/* Protocol type */
-	0,                              /* Permanent cmodes */
-	0,                              /* Oper-immune cmode */
-	"beI",                          /* Ban-like cmodes */
-	'e',                            /* Except mchar */
-	'I',                            /* Invex mchar */
-	IRCD_HOLDNICK                   /* Flags */
+	.ircdname = "UnrealIRCd 3.1 or later",
+	.tldprefix = "$",
+	.uses_uid = true,
+	.uses_rcommand = false,
+	.uses_owner = true,
+	.uses_protect = true,
+	.uses_halfops = true,
+	.uses_p10 = false,
+	.uses_vhost = true,
+	.oper_only_modes = CMODE_OPERONLY | CMODE_ADMONLY,
+	.owner_mode = CSTATUS_OWNER,
+	.protect_mode = CSTATUS_PROTECT,
+	.halfops_mode = CSTATUS_HALFOP,
+	.owner_mchar = "+q",
+	.protect_mchar = "+a",
+	.halfops_mchar = "+h",
+	.type = PROTOCOL_UNREAL,
+	.perm_mode = CMODE_PERM,
+	.oimmune_mode = 0,
+	.ban_like_modes = "beI",
+	.except_mchar = 'e',
+	.invex_mchar = 'I',
+	.flags = IRCD_HOLDNICK | IRCD_SASL_USE_PUID,
 };
 
 struct cmode_ unreal_mode_list[] = {
-  { 'i', CMODE_INVITE   },
-  { 'm', CMODE_MOD      },
-  { 'n', CMODE_NOEXT    },
-  { 'p', CMODE_PRIV     },
-  { 's', CMODE_SEC      },
-  { 't', CMODE_TOPIC    },
-  { 'c', CMODE_NOCOLOR  },
-  { 'M', CMODE_MODREG   },
-  { 'R', CMODE_REGONLY  },
+  { 'i', CMODE_INVITE	},
+  { 'm', CMODE_MOD	},
+  { 'n', CMODE_NOEXT	},
+  { 'p', CMODE_PRIV	},
+  { 's', CMODE_SEC	},
+  { 't', CMODE_TOPIC	},
+  { 'c', CMODE_NOCOLOR	},
+  { 'M', CMODE_MODREG	},
+  { 'R', CMODE_REGONLY	},
   { 'O', CMODE_OPERONLY },
-  { 'A', CMODE_ADMONLY  },
-  { 'Q', CMODE_PEACE    },
-  { 'S', CMODE_STRIP    },
-  { 'K', CMODE_NOKNOCK  },
+  { 'A', CMODE_ADMONLY	},
+  { 'Q', CMODE_PEACE	},
+  { 'S', CMODE_STRIP	},
+  { 'K', CMODE_NOKNOCK	},
   { 'V', CMODE_NOINVITE },
-  { 'C', CMODE_NOCTCP   },
-  { 'u', CMODE_HIDING   },
-  { 'z', CMODE_SSLONLY  },
-  { 'N', CMODE_STICKY   },
-  { 'G', CMODE_CENSOR   },
+  { 'C', CMODE_NOCTCP	},
+  { 'u', CMODE_HIDING	},
+  { 'z', CMODE_SSLONLY	},
+  { 'N', CMODE_STICKY	},
+  { 'G', CMODE_CENSOR	},
+  { 'r', CMODE_CHANREG	},
+  { 'P', CMODE_PERM	},
   { '\0', 0 }
 };
 
@@ -80,20 +85,20 @@ struct extmode unreal_ignore_mode_list[] = {
 };
 
 struct cmode_ unreal_status_mode_list[] = {
-  { 'q', CSTATUS_OWNER   },
+  { 'q', CSTATUS_OWNER	 },
   { 'a', CSTATUS_PROTECT },
-  { 'o', CSTATUS_OP      },
+  { 'o', CSTATUS_OP	 },
   { 'h', CSTATUS_HALFOP  },
-  { 'v', CSTATUS_VOICE   },
+  { 'v', CSTATUS_VOICE	 },
   { '\0', 0 }
 };
 
 struct cmode_ unreal_prefix_mode_list[] = {
-  { '*', CSTATUS_OWNER   },
+  { '*', CSTATUS_OWNER	 },
   { '~', CSTATUS_PROTECT },
-  { '@', CSTATUS_OP      },
+  { '@', CSTATUS_OP	 },
   { '%', CSTATUS_HALFOP  },
-  { '+', CSTATUS_VOICE   },
+  { '+', CSTATUS_VOICE	 },
   { '\0', 0 }
 };
 
@@ -120,7 +125,7 @@ static bool check_jointhrottle(const char *value, channel_t *c, mychan_t *mc, us
 				return false;
 			arg2 = p + 1;
 		}
-		else if (!isdigit(*p))
+		else if (!isdigit((unsigned char)*p))
 			return false;
 		p++;
 	}
@@ -131,10 +136,98 @@ static bool check_jointhrottle(const char *value, channel_t *c, mychan_t *mc, us
 	return true;
 }
 
+/* +f 3:1 or +f *3:1 (which is like +f [3t]:1 or +f [3t#b]:1) */
+static inline bool check_flood_old(const char *value, channel_t *c, mychan_t *mc, user_t *u, myuser_t *mu)
+{
+	bool found_colon = false;
+
+	return_val_if_fail(value != NULL, false);
+
+	/* x:y is 3 bytes long, so that is the minimum length of a +f parameter. */
+	if (strlen(value) < 3)
+		return false;
+
+	/* skip past the * if present */
+	if (*value == '*')
+		value++;
+
+	/* check to make sure all bytes are numbers, allowing for one colon */
+	while (*value != '\0')
+	{
+		if (*value == ':' && !found_colon)
+			found_colon = true;
+		else if (!isdigit((unsigned char)*value))
+			return false;
+
+		value++;
+	}
+
+	/* we have to have a colon in order for it to be valid */
+	if (!found_colon)
+		return false;
+
+	return true;
+}
+
+#define VALID_FLOOD_CHAR(c)	((c == 'c') || (c == 'j') || (c == 'k') || (c == 'm') || (c == 'n') || (c == 't'))
+#define VALID_ACTION_CHAR(c)	((c == 'm') || (c == 'M') || (c == 'C') || (c == 'R') || (c == 'i') || (c == 'K') \
+				 || (c == 'N') || (c == 'b'))
+
+/*
+ * +f *X:Y	 (handled by check_flood_old)
+ * +f X:Y	 (handled by check_flood_old)
+ *
+ * +f [<number><letter>(#<letter>)(,...)]
+ */
 static bool check_flood(const char *value, channel_t *c, mychan_t *mc, user_t *u, myuser_t *mu)
 {
-	/* way too complicated */
-	return false;
+	char evalbuf[BUFSIZE], *ep, *p;
+
+	if (*value != '[')
+		return check_flood_old(value, c, mc, u, mu);
+
+	/* copy this to a local buffer for evaluation */
+	mowgli_strlcpy(evalbuf, value, sizeof evalbuf);
+	ep = evalbuf + 1;
+
+	/* check that the parameter ends with a ] */
+	if ((p = strchr(ep, ']')) == NULL)
+		return false;
+
+	/* we have a ], blast it away and check for a colon. */
+	*p = '\0';
+	if (*(p + 1) != ':')
+		return false;
+
+	for (p = strtok(ep, ","); p != NULL; p = strtok(NULL, ","))
+	{
+		while (isdigit((unsigned char)*p))
+			p++;
+
+		if (!VALID_FLOOD_CHAR(*p))
+			return false;
+
+		*p = '\0';
+		p++;
+
+		if (*p != '\0')
+		{
+			if (*p == '#')
+			{
+				p++;
+
+				if (!VALID_ACTION_CHAR(*p))
+					return false;
+
+				continue;
+			}
+
+			/* not valid, needs to be # or nothing */
+			return false;
+		}
+	}
+
+	return true;
 }
 
 static bool check_forward(const char *value, channel_t *c, mychan_t *mc, user_t *u, myuser_t *mu)
@@ -142,32 +235,103 @@ static bool check_forward(const char *value, channel_t *c, mychan_t *mc, user_t
 	channel_t *target_c;
 	mychan_t *target_mc;
 
-	if (*value != '#' || strlen(value) > 50)
+	if (!VALID_GLOBAL_CHANNEL_PFX(value) || strlen(value) > 50)
 		return false;
 	if (u == NULL && mu == NULL)
 		return true;
 	target_c = channel_find(value);
-	target_mc = mychan_find(value);
+	target_mc = mychan_from(target_c);
 	if (target_c == NULL && target_mc == NULL)
 		return false;
 	return true;
 }
 
+static mowgli_node_t *unreal_next_matching_ban(channel_t *c, user_t *u, int type, mowgli_node_t *first)
+{
+	chanban_t *cb;
+	mowgli_node_t *n;
+	char hostbuf[NICKLEN+USERLEN+HOSTLEN];
+	char realbuf[NICKLEN+USERLEN+HOSTLEN];
+	char ipbuf[NICKLEN+USERLEN+HOSTLEN];
+	char *p;
+	bool matched;
+	int exttype;
+	channel_t *target_c;
+
+	snprintf(hostbuf, sizeof hostbuf, "%s!%s@%s", u->nick, u->user, u->vhost);
+	snprintf(realbuf, sizeof realbuf, "%s!%s@%s", u->nick, u->user, u->host);
+	/* will be nick!user@ if ip unknown, doesn't matter */
+	snprintf(ipbuf, sizeof ipbuf, "%s!%s@%s", u->nick, u->user, u->ip);
+
+	MOWGLI_ITER_FOREACH(n, first)
+	{
+		cb = n->data;
+
+		if (cb->type != type)
+			continue;
+
+		if ((!match(cb->mask, hostbuf) || !match(cb->mask, realbuf) || !match(cb->mask, ipbuf)))
+			return n;
+		if (cb->mask[0] == '~')
+		{
+			p = cb->mask + 1;
+			exttype = *p++;
+			if (exttype == '\0')
+				continue;
+			/* check parameter */
+			if (*p++ != ':')
+				p = NULL;
+			switch (exttype)
+			{
+				case 'a':
+					matched = u->myuser != NULL && !(u->myuser->flags & MU_WAITAUTH) && (p == NULL || !match(p, entity(u->myuser)->name));
+					break;
+				case 'c':
+					if (p == NULL)
+						continue;
+					target_c = channel_find(p);
+					if (target_c == NULL || (target_c->modes & (CMODE_PRIV | CMODE_SEC)))
+						continue;
+					matched = chanuser_find(target_c, u) != NULL;
+					break;
+				case 'r':
+					if (p == NULL)
+						continue;
+					matched = !match(p, u->gecos);
+					break;
+				case 'R':
+					matched = should_reg_umode(u);
+					break;
+				case 'q':
+					matched = !match(p, hostbuf) || !match(p, ipbuf);
+					break;
+				default:
+					continue;
+			}
+			if (matched)
+				return n;
+		}
+	}
+	return NULL;
+}
+
 /* login to our uplink */
 static unsigned int unreal_server_login(void)
 {
 	int ret;
 
-	ret = sts("PASS %s", curr_uplink->pass);
+	ret = sts("PASS %s", curr_uplink->send_pass);
 	if (ret == 1)
 		return 1;
 
 	me.bursting = true;
+	has_protoctl = false;
 
-	sts("PROTOCTL TOKEN NICKv2 VHP NICKIP UMODE2 SJOIN SJOIN2 SJ3 NOQUIT TKLEXT ESVID");
+	sts("PROTOCTL NICKv2 VHP NICKIP UMODE2 SJOIN SJOIN2 SJ3 NOQUIT TKLEXT ESVID MLOCK");
+	sts("PROTOCTL SID=%s", me.numeric);
 	sts("SERVER %s 1 :%s", me.name, me.desc);
 
-	services_init();
+	sts(":%s EOS", ME);
 
 	return 0;
 }
@@ -177,38 +341,38 @@ static void unreal_introduce_nick(user_t *u)
 {
 	const char *umode = user_get_umodestr(u);
 
-	sts("NICK %s 1 %lu %s %s %s * %sS * :%s", u->nick, (unsigned long)u->ts, u->user, u->host, me.name, umode, u->gecos);
+	if (!ircd->uses_uid)
+		sts("NICK %s 1 %lu %s %s %s 0 %sS * :%s", u->nick, (unsigned long)u->ts, u->user, u->host, me.name, umode, u->gecos);
+	else
+		sts(":%s UID %s 1 %lu %s %s %s * %sS * * * :%s", ME, u->nick, (unsigned long)u->ts, u->user, u->host, u->uid, umode, u->gecos);
 }
 
 /* invite a user to a channel */
 static void unreal_invite_sts(user_t *sender, user_t *target, channel_t *channel)
 {
-	sts(":%s INVITE %s %s", sender->nick, target->nick, channel->name);
+	sts(":%s INVITE %s %s", CLIENT_NAME(sender), CLIENT_NAME(target), channel->name);
 }
 
 static void unreal_quit_sts(user_t *u, const char *reason)
 {
-	if (!me.connected)
-		return;
-
-	sts(":%s QUIT :%s", u->nick, reason);
+	sts(":%s QUIT :%s", CLIENT_NAME(u), reason);
 }
 
 /* WALLOPS wrapper */
 static void unreal_wallops_sts(const char *text)
 {
-	sts(":%s GLOBOPS :%s", me.name, text);
+	sts(":%s GLOBOPS :%s", ME, text);
 }
 
 /* join a channel */
 static void unreal_join_sts(channel_t *c, user_t *u, bool isnew, char *modes)
 {
 	if (isnew)
-		sts(":%s SJOIN %lu %s %s :@%s", me.name, (unsigned long)c->ts,
-				c->name, modes, u->nick);
+		sts(":%s SJOIN %lu %s %s :@%s", ME, (unsigned long)c->ts,
+				c->name, modes, CLIENT_NAME(u));
 	else
-		sts(":%s SJOIN %lu %s + :@%s", me.name, (unsigned long)c->ts,
-				c->name, u->nick);
+		sts(":%s SJOIN %lu %s + :@%s", ME, (unsigned long)c->ts,
+				c->name, CLIENT_NAME(u));
 }
 
 /* lower TS */
@@ -251,17 +415,17 @@ static void unreal_msg_global_sts(user_t *from, const char *mask, const char *te
 		MOWGLI_ITER_FOREACH(n, tldlist.head)
 		{
 			tld = n->data;
-			sts(":%s PRIVMSG %s*%s :%s", from ? from->nick : me.name, ircd->tldprefix, tld->name, text);
+			sts(":%s PRIVMSG %s*%s :%s", from ? CLIENT_NAME(from) : ME, ircd->tldprefix, tld->name, text);
 		}
 	}
 	else
-		sts(":%s PRIVMSG %s%s :%s", from ? from->nick : me.name, ircd->tldprefix, mask, text);
+		sts(":%s PRIVMSG %s%s :%s", from ? CLIENT_NAME(from) : ME, ircd->tldprefix, mask, text);
 }
 
 /* NOTICE wrapper */
 static void unreal_notice_user_sts(user_t *from, user_t *target, const char *text)
 {
-	sts(":%s NOTICE %s :%s", from ? from->nick : me.name, target->nick, text);
+	sts(":%s NOTICE %s :%s", from ? CLIENT_NAME(from) : ME, CLIENT_NAME(target), text);
 }
 
 static void unreal_notice_global_sts(user_t *from, const char *mask, const char *text)
@@ -274,16 +438,16 @@ static void unreal_notice_global_sts(user_t *from, const char *mask, const char
 		MOWGLI_ITER_FOREACH(n, tldlist.head)
 		{
 			tld = n->data;
-			sts(":%s NOTICE %s*%s :%s", from ? from->nick : me.name, ircd->tldprefix, tld->name, text);
+			sts(":%s NOTICE %s*%s :%s", from ? CLIENT_NAME(from) : ME, ircd->tldprefix, tld->name, text);
 		}
 	}
 	else
-		sts(":%s NOTICE %s%s :%s", from ? from->nick : me.name, ircd->tldprefix, mask, text);
+		sts(":%s NOTICE %s%s :%s", from ? CLIENT_NAME(from) : ME, ircd->tldprefix, mask, text);
 }
 
 static void unreal_notice_channel_sts(user_t *from, channel_t *target, const char *text)
 {
-	sts(":%s NOTICE %s :%s", from ? from->nick : me.name, target->name, text);
+	sts(":%s NOTICE %s :%s", from ? CLIENT_NAME(from) : ME, target->name, text);
 }
 
 static void unreal_numeric_sts(server_t *from, int numeric, user_t *target, const char *fmt, ...)
@@ -295,7 +459,7 @@ static void unreal_numeric_sts(server_t *from, int numeric, user_t *target, cons
 	vsnprintf(buf, BUFSIZE, fmt, ap);
 	va_end(ap);
 
-	sts(":%s %d %s %s", from->name, numeric, target->nick, buf);
+	sts(":%s %d %s %s", SERVER_NAME(from), numeric, CLIENT_NAME(target), buf);
 }
 
 /* KILL wrapper */
@@ -322,16 +486,16 @@ static void unreal_kill_id_sts(user_t *killer, const char *id, const char *reaso
 			 * dropped.
 			 */
 		}
-		sts(":%s KILL %s :%s!%s (%s)", killer->nick, id, killer->host, killer->nick, reason);
+		sts(":%s KILL %s :%s!%s (%s)", CLIENT_NAME(killer), id, killer->host, killer->nick, reason);
 	}
 	else
-		sts(":%s KILL %s :%s (%s)", me.name, id, me.name, reason);
+		sts(":%s KILL %s :%s (%s)", ME, id, me.name, reason);
 }
 
 /* PART wrapper */
 static void unreal_part_sts(channel_t *c, user_t *u)
 {
-	sts(":%s PART %s", u->nick, c->name);
+	sts(":%s PART %s", CLIENT_NAME(u), c->name);
 }
 
 /* server-to-server KLINE wrapper */
@@ -339,11 +503,8 @@ static void unreal_kline_sts(const char *server, const char *user, const char *h
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
-	sts(":%s TKL + G %s %s %s %lu %lu :%s", me.name, user, host, svs != NULL ? svs->nick : me.name, (unsigned long)(duration > 0 ? CURRTIME + duration : 0), (unsigned long)CURRTIME, reason);
+	sts(":%s TKL + G %s %s %s %lu %lu :%s", ME, user, host, svs != NULL ? svs->nick : me.name, (unsigned long)(duration > 0 ? CURRTIME + duration : 0), (unsigned long)CURRTIME, reason);
 }
 
 /* server-to-server UNKLINE wrapper */
@@ -351,51 +512,39 @@ static void unreal_unkline_sts(const char *server, const char *user, const char
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
-	sts(":%s TKL - G %s %s %s", me.name, user, host, svs != NULL ? svs->nick : me.name);
+	sts(":%s TKL - G %s %s %s", ME, user, host, svs != NULL ? svs->nick : me.name);
 }
 
 static void unreal_xline_sts(const char *server, const char *realname, long duration, const char *reason)
 {
 	char escapedreason[512], *p;
 
-	if (!me.connected)
-		return;
-
 	if (duration > 0)
 	{
 		slog(LG_INFO, "SGLINE: Could not set temporary SGLINE on \2%s\2, not supported by unrealircd.", realname);
 		return;
 	}
 
-	strlcpy(escapedreason, reason, sizeof escapedreason);
+	mowgli_strlcpy(escapedreason, reason, sizeof escapedreason);
 	for (p = escapedreason; *p != '\0'; p++)
 		if (*p == ' ')
 			*p = '_';
 	if (*escapedreason == ':')
 		*escapedreason = ';';
 
-	sts(":%s BR + %s :%s", me.name, escapedreason, realname);
+	sts(":%s SVSNLINE + %s :%s", ME, escapedreason, realname);
 }
 
 static void unreal_unxline_sts(const char *server, const char *realname)
 {
-	if (!me.connected)
-		return;
-
-	sts(":%s BR - :%s", me.name, realname);
+	sts(":%s SVSNLINE - :%s", ME, realname);
 }
 
 static void unreal_qline_sts(const char *server, const char *name, long duration, const char *reason)
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	if (*name == '#' || *name == '&')
 	{
 		slog(LG_INFO, "SQLINE: Could not set SQLINE on \2%s\2, not supported by unrealircd.", name);
@@ -403,25 +552,22 @@ static void unreal_qline_sts(const char *server, const char *name, long duration
 	}
 
 	svs = service_find("operserv");
-	sts(":%s TKL + Q * %s %s %lu %lu :%s", me.name, name, svs != NULL ? svs->nick : me.name, (unsigned long)(duration > 0 ? CURRTIME + duration : 0), (unsigned long)CURRTIME, reason);
+	sts(":%s TKL + Q * %s %s %lu %lu :%s", ME, name, svs != NULL ? svs->nick : me.name, (unsigned long)(duration > 0 ? CURRTIME + duration : 0), (unsigned long)CURRTIME, reason);
 }
 
 static void unreal_unqline_sts(const char *server, const char *name)
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	svs = service_find("operserv");
-	sts(":%s TKL - Q * %s %s", me.name, name, svs != NULL ? svs->nick : me.name);
+	sts(":%s TKL - Q * %s %s", ME, name, svs != NULL ? svs->nick : me.name);
 }
 
 /* topic wrapper */
 static void unreal_topic_sts(channel_t *c, user_t *source, const char *setter, time_t ts, time_t prevts, const char *topic)
 {
-	if (!me.connected || !c)
-		return;
+	return_if_fail(c != NULL);
+	return_if_fail(source != NULL);
 
 	sts(":%s TOPIC %s %s %lu :%s", source->nick, c->name, setter, (unsigned long)ts, topic);
 }
@@ -429,8 +575,9 @@ static void unreal_topic_sts(channel_t *c, user_t *source, const char *setter, t
 /* mode wrapper */
 static void unreal_mode_sts(char *sender, channel_t *target, char *modes)
 {
-	if (!me.connected)
-		return;
+	return_if_fail(sender != NULL);
+	return_if_fail(target != NULL);
+	return_if_fail(modes != NULL);
 
 	sts(":%s MODE %s %s", sender, target->name, modes);
 }
@@ -438,17 +585,14 @@ static void unreal_mode_sts(char *sender, channel_t *target, char *modes)
 /* ping wrapper */
 static void unreal_ping_sts(void)
 {
-	if (!me.connected)
-		return;
-
-	sts("PING :%s", me.name);
+	sts("PING :%s", ME);
 }
 
 /* protocol-specific stuff to do on login */
 static void unreal_on_login(user_t *u, myuser_t *account, const char *wantedhost)
 {
-	if (!me.connected || u == NULL)
-		return;
+	return_if_fail(u != NULL);
+	return_if_fail(account != NULL);
 
 	if (!use_esvid)
 	{
@@ -462,19 +606,19 @@ static void unreal_on_login(user_t *u, myuser_t *account, const char *wantedhost
 		return;
 	}
 
-	sts(":%s SVS2MODE %s +rd %s", nicksvs.nick, u->nick, entity(account)->name);
+	if (should_reg_umode(u))
+		sts(":%s SVS2MODE %s +rd %s", nicksvs.nick, u->nick, entity(account)->name);
+	else
+		sts(":%s SVS2MODE %s +d %s", nicksvs.nick, u->nick, entity(account)->name);
 }
 
 /* protocol-specific stuff to do on logout */
 static bool unreal_on_logout(user_t *u, const char *account)
 {
-	if (!me.connected || u == NULL)
-		return false;
+	return_val_if_fail(u != NULL, false);
 
-	if (!use_esvid && !nicksvs.no_nick_ownership)
+	if (use_esvid || !nicksvs.no_nick_ownership)
 		sts(":%s SVS2MODE %s -r+d 0", nicksvs.nick, u->nick);
-	else
-		sts(":%s SVS2MODE %s -r+d *", nicksvs.nick, u->nick);
 
 	return false;
 }
@@ -483,21 +627,15 @@ static void unreal_jupe(const char *server, const char *reason)
 {
 	service_t *svs;
 
-	if (!me.connected)
-		return;
-
 	server_delete(server);
 
 	svs = service_find("operserv");
-	sts(":%s SQUIT %s :%s", svs != NULL ? svs->nick : me.name, server, reason);
+	sts(":%s SQUIT %s :%s", svs != NULL ? svs->nick : ME, server, reason);
 	sts(":%s SERVER %s 2 :%s", me.name, server, reason);
 }
 
 static void unreal_sethost_sts(user_t *source, user_t *target, const char *host)
 {
-	if (!me.connected)
-		return;
-
 	sts(":%s CHGHOST %s :%s", source->nick, target->nick, host);
 
 	if (irccasecmp(target->host, host))
@@ -509,9 +647,9 @@ static void unreal_sethost_sts(user_t *source, user_t *target, const char *host)
 	}
 }
 
-static void unreal_fnc_sts(user_t *source, user_t *u, char *newnick, int type)
+static void unreal_fnc_sts(user_t *source, user_t *u, const char *newnick, int type)
 {
-	sts(":%s SVSNICK %s %s %lu", source->nick, u->nick, newnick,
+	sts(":%s SVSNICK %s %s %lu", CLIENT_NAME(source), CLIENT_NAME(u), newnick,
 			(unsigned long)(CURRTIME - 60));
 }
 
@@ -519,13 +657,117 @@ static void unreal_holdnick_sts(user_t *source, int duration, const char *nick,
 {
 	if (duration > 0)
 		sts(":%s TKL + Q H %s %s %lu %lu :Reserved by %s for nickname owner (%s)",
-				me.name, nick, source->nick,
+				ME, nick, source->nick,
 				(unsigned long)(CURRTIME + duration),
 				(unsigned long)CURRTIME,
 				source->nick,
 				mu ? entity(mu)->name : nick);
 	else
-		sts(":%s TKL - Q H %s %s", me.name, nick, source->nick);
+		sts(":%s TKL - Q H %s %s", ME, nick, source->nick);
+}
+
+static void unreal_quarantine_sts(user_t *source, user_t *victim, long duration, const char *reason)
+{
+	sts(":%s SHUN +*@%s %ld :%s", source->nick, victim->host, duration, reason);
+}
+
+static void unreal_sasl_sts(char *target, char mode, char *data)
+{
+	char servermask[BUFSIZE], *p;
+	service_t *saslserv;
+
+	saslserv = service_find("saslserv");
+	if (saslserv == NULL)
+		return;
+
+	/* extract the servername from the target. */
+	mowgli_strlcpy(servermask, target, sizeof servermask);
+	p = strchr(servermask, '!');
+	if (p != NULL)
+		*p = '\0';
+
+	sts(":%s SASL %s %s %c %s", saslserv->me->nick, servermask, target, mode, data);
+}
+
+static void unreal_svslogin_sts(char *target, char *nick, char *user, char *host, myuser_t *account)
+{
+	char servermask[BUFSIZE], *p;
+	service_t *saslserv;
+
+	saslserv = service_find("saslserv");
+	if (saslserv == NULL)
+		return;
+
+	/* extract the servername from the target. */
+	mowgli_strlcpy(servermask, target, sizeof servermask);
+	p = strchr(servermask, '!');
+	if (p != NULL)
+		*p = '\0';
+
+	sts(":%s SVSLOGIN %s %s %s", saslserv->me->nick, servermask, target, entity(account)->name);
+}
+
+static void unreal_mlock_sts(channel_t *c)
+{
+	mychan_t *mc = mychan_from(c);
+
+	if (use_mlock == false)
+		return;
+
+	if (mc == NULL)
+		return;
+
+	sts(":%s MLOCK %lu %s :%s", ME, (unsigned long)c->ts, c->name,
+				    mychan_get_sts_mlock(mc));
+}
+
+static void m_mlock(sourceinfo_t *si, int parc, char *parv[])
+{
+	channel_t *c;
+	mychan_t *mc;
+	const char *mlock;
+
+	/* Ignore MLOCK if the server isn't bursting, to avoid 'war' conditions */
+	if (si->s->flags & SF_EOB)
+		return;
+
+	if (!(c = channel_find(parv[1])))
+		return;
+
+	if (!(mc = mychan_from(c)))
+	{
+		/* Unregistered channel. Clear the MLOCK. */
+		sts(":%s MLOCK %lu %s :", ME, (unsigned long)c->ts, c->name);
+		return;
+	}
+
+	time_t ts = atol(parv[0]);
+	if (ts > c->ts)
+		return;
+
+	mlock = mychan_get_sts_mlock(mc);
+	if (0 != strcmp(parv[2], mlock))
+	{
+		/* MLOCK is changing, with the same TS. Bounce back the correct one. */
+		sts(":%s MLOCK %lu %s :%s", ME, (unsigned long)c->ts, c->name, mlock);
+	}
+}
+
+static void m_sasl(sourceinfo_t *si, int parc, char *parv[])
+{
+	sasl_message_t smsg;
+
+	/* :irc.loldongs.eu SASL * irc.goatse.cx!42 S d29vTklOSkFTAGRhdGEgaW4gZmlyc3QgbGluZQ== */
+	if (parc < 4)
+		return;
+
+	smsg.uid = parv[1];
+	smsg.mode = *parv[2];
+	smsg.buf = parv[3];
+	smsg.ext = parc >= 4 ? parv[4] : NULL;
+	smsg.server = si->s ? si->s : NULL;
+
+	hook_call_sasl_input(&smsg);
 }
 
 static void m_topic(sourceinfo_t *si, int parc, char *parv[])
@@ -548,7 +790,7 @@ static void m_topic(sourceinfo_t *si, int parc, char *parv[])
 static void m_ping(sourceinfo_t *si, int parc, char *parv[])
 {
 	/* reply to PING's */
-	sts(":%s PONG %s %s", me.name, me.name, parv[0]);
+	sts(":%s PONG %s %s", ME, me.name, parv[0]);
 }
 
 static void m_pong(sourceinfo_t *si, int parc, char *parv[])
@@ -563,7 +805,7 @@ static void m_pong(sourceinfo_t *si, int parc, char *parv[])
 		return;
 	handle_eob(s);
 
-	if (irccasecmp(me.actual, parv[0]))
+	if (s != si->s)
 		return;
 
 	me.uplinkpong = CURRTIME;
@@ -625,9 +867,9 @@ static void m_sjoin(sourceinfo_t *si, int parc, char *parv[])
 {
 	/*
 	 *  -> :proteus.malkier.net SJOIN 1073516550 #shrike +tn :@sycobuny @+rakaur
-	 *      also:
+	 *	also:
 	 *  -> :nenolod_ SJOIN 1117334567 #chat
-	 *      also:
+	 *	also:
 	 *  -> SJOIN 1117334567 #chat :@nenolod
 	 */
 
@@ -749,6 +991,93 @@ static void m_part(sourceinfo_t *si, int parc, char *parv[])
 	}
 }
 
+static void m_uid(sourceinfo_t *si, int parc, char *parv[])
+{
+	server_t *s;
+	user_t *u;
+	bool realchange;
+	const char *vhost;
+	const char *ipb64;
+	char ipstring[64];
+	int af;
+	size_t iplen;
+	char ipdata[16];
+
+	if (parc == 12)
+	{
+		s = si->s;
+		if (!s)
+		{
+			slog(LG_DEBUG, "m_uid(): new user on nonexistant server: %s", parv[0]);
+			return;
+		}
+
+		slog(LG_DEBUG, "m_uid(): new user on `%s': %s", s->name, si->s->name);
+
+		vhost = strcmp(parv[8], "*") ? parv[8] : NULL;
+		iplen = 0;
+		if (parc == 11 && strcmp(parv[parc - 2], "*"))
+		{
+			ipb64 = parv[parc - 2];
+			af = AF_INET;
+			if (strlen(ipb64) == 8)
+			{
+				iplen = 4;
+				if (base64_decode(ipb64, ipdata, iplen) != iplen)
+					iplen = 0;
+				af = AF_INET;
+			}
+#ifdef AF_INET6
+			else if (strlen(ipb64) == 24)
+			{
+				iplen = 16;
+				if (base64_decode(ipb64, ipdata, iplen) != iplen)
+					iplen = 0;
+				af = AF_INET6;
+			}
+#endif
+			if (iplen != 0)
+				if (!inet_ntop(af, ipdata, ipstring, sizeof ipstring))
+					iplen = 0;
+		}
+		u = user_add(parv[0], parv[3], parv[4], vhost, iplen != 0 ? ipstring : NULL, parv[5], parv[parc - 1], s, atoi(parv[2]));
+		if (u == NULL)
+			return;
+
+		user_mode(u, parv[7]);
+
+		/*
+		 * with ESVID:
+		 * If the user's SVID is equal to their accountname,
+		 * they're properly logged in.	Alternatively, the
+		 * 'without ESVID' criteria is used. --nenolod
+		 *
+		 * without ESVID:
+		 * If the user's SVID is equal to their nick TS,
+		 * they're properly logged in --jilles
+		 */
+		if (use_esvid && !IsDigit(*parv[6]))
+		{
+			handle_burstlogin(u, parv[6], 0);
+
+			if (authservice_loaded && should_reg_umode(u))
+				sts(":%s SVS2MODE %s +r", nicksvs.nick, u->nick);
+		}
+		else if (u->ts > 100 && (time_t)atoi(parv[6]) == u->ts)
+			handle_burstlogin(u, NULL, 0);
+
+		handle_nickchange(u);
+	}
+	else
+	{
+		int i;
+		slog(LG_DEBUG, "m_uid(): got UID with wrong number of params");
+
+		for (i = 0; i < parc; i++)
+			slog(LG_DEBUG, "m_uid():   parv[%d] = %s", i, parv[i]);
+	}
+}
+
 static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 {
 	server_t *s;
@@ -761,7 +1090,7 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 	size_t iplen;
 	char ipdata[16];
 
-	if (parc == 10 || parc == 11)
+	if (parc > 10)
 	{
 		s = server_find(parv[5]);
 		if (!s)
@@ -781,8 +1110,7 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 			if (strlen(ipb64) == 8)
 			{
 				iplen = 4;
-				if (!base64_decode(ipb64, 8, ipdata, &iplen) ||
-						iplen != 4)
+				if (base64_decode(ipb64, ipdata, iplen) != iplen)
 					iplen = 0;
 				af = AF_INET;
 			}
@@ -790,8 +1118,7 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 			else if (strlen(ipb64) == 24)
 			{
 				iplen = 16;
-				if (!base64_decode(ipb64, 24, ipdata, &iplen) ||
-						iplen != 16)
+				if (base64_decode(ipb64, ipdata, iplen) != iplen)
 					iplen = 0;
 				af = AF_INET6;
 			}
@@ -809,15 +1136,20 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 		/*
 		 * with ESVID:
 		 * If the user's SVID is equal to their accountname,
-		 * they're properly logged in.  Alternatively, the
+		 * they're properly logged in.	Alternatively, the
 		 * 'without ESVID' criteria is used. --nenolod
 		 *
 		 * without ESVID:
 		 * If the user's SVID is equal to their nick TS,
 		 * they're properly logged in --jilles
 		 */
-		if (use_esvid && *parv[6] != '*' && !IsDigit(*parv[6]))
+		if (use_esvid && !IsDigit(*parv[6]))
+		{
 			handle_burstlogin(u, parv[6], 0);
+
+			if (authservice_loaded && should_reg_umode(u))
+				sts(":%s SVS2MODE %s +r", nicksvs.nick, u->nick);
+		}
 		else if (u->ts > 100 && (time_t)atoi(parv[6]) == u->ts)
 			handle_burstlogin(u, NULL, 0);
 
@@ -827,12 +1159,12 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 	/* if it's only 2 then it's a nickname change */
 	else if (parc == 2)
 	{
-                if (!si->su)
-                {       
-                        slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
-                        return;
-                }
-                
+		if (!si->su)
+		{
+			slog(LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
+			return;
+		}
+
 		slog(LG_DEBUG, "m_nick(): nickname change from `%s': %s", si->su->nick, parv[0]);
 
 		realchange = irccasecmp(si->su->nick, parv[0]);
@@ -850,6 +1182,13 @@ static void m_nick(sourceinfo_t *si, int parc, char *parv[])
 				/* changed from registered nick, remove +r */
 				sts(":%s SVS2MODE %s -r+d 0", nicksvs.nick, parv[0]);
 		}
+		else if (realchange && !nicksvs.no_nick_ownership && use_esvid)
+		{
+			if (should_reg_umode(si->su))
+				sts(":%s SVS2MODE %s +r", nicksvs.nick, parv[0]);
+			else
+				sts(":%s SVS2MODE %s -r", nicksvs.nick, parv[0]);
+		}
 
 		handle_nickchange(si->su);
 	}
@@ -894,14 +1233,34 @@ static void unreal_user_mode(user_t *u, const char *changes)
 					 * vhost instead of their cloaked host. - Adam
 					 */
 					if (strcmp(u->vhost, u->chost))
-						strlcpy(u->chost, u->vhost, HOSTLEN);
+					{
+						strshare_unref(u->chost);
+						u->chost = strshare_get(u->vhost);
+					}
 				}
 				else if (dir == MTYPE_DEL)
-					strlcpy(u->vhost, u->host, HOSTLEN);
+				{
+					strshare_unref(u->vhost);
+					u->vhost = strshare_get(u->host);
+				}
 				break;
 		}
 }
 
+static bool unreal_is_extban(const char *mask)
+{
+	const char mask_len = strlen(mask);
+	unsigned char offset = 0;
+
+	if (mask_len < 4 || mask[0] != '~' || mask[2] != ':' || strchr(mask, ' '))
+		return false;
+
+	if ((mask[1] < 'a' || mask[1] > 'z') && (mask[1] < 'A' || mask[1] > 'Z'))
+		return false;
+
+	return true;
+}
+
 static void m_mode(sourceinfo_t *si, int parc, char *parv[])
 {
 	if (*parv[0] == '#')
@@ -967,19 +1326,61 @@ static void m_server(sourceinfo_t *si, int parc, char *parv[])
 	server_t *s;
 	const char *inf;
 
+	/* because multiple PROTOCTL messages are allowed without a PROTOCTL END,
+	 * and even if we added a PROTOCTL END in unreal 3.4, we'd still have to do
+	 * this hack in order to support 3.2... -- kaniini
+	 */
+	if (has_protoctl)
+	{
+		if (ts6sid[0] == '\0')
+		{
+			ircd->uses_uid = false;
+			if (me.me->sid)
+			{
+				/* XXX: this is a hack that fixes *most*, though not
+				 * all, of the brokenness when linking to Unreal 3.2
+				 *
+				 * (the EOB is sent before this, therefore still with
+				 * a SID, but apparently still works) --grawity */
+				slog(LG_DEBUG, "m_server(): erasing our SID");
+				free(me.me->sid);
+				me.me->sid = NULL;
+			}
+		}
+
+		services_init();
+		has_protoctl = false;	/* only once after PROTOCTL message. */
+	}
+
 	slog(LG_DEBUG, "m_server(): new server: %s", parv[0]);
 	if (si->s == NULL && (inf = strchr(parv[2], ' ')) != NULL)
 		inf++;
 	else
 		inf = parv[2];
-	s = handle_server(si, parv[0], NULL, atoi(parv[1]), inf);
+	s = handle_server(si, parv[0], si->s || !ircd->uses_uid ? NULL : ts6sid, atoi(parv[1]), inf);
 
 	if (s != NULL && s->uplink != me.me)
 	{
 		/* elicit PONG for EOB detection; pinging uplink is
 		 * already done elsewhere -- jilles
 		 */
-		sts(":%s PING %s %s", me.name, me.name, s->name);
+		sts(":%s PING %s %s", ME, me.name, s->name);
+	}
+}
+
+static void m_sid(sourceinfo_t *si, int parc, char *parv[])
+{
+	server_t *s;
+	const char *inf;
+
+	s = handle_server(si, parv[0], parv[2], atoi(parv[1]), parv[3]);
+
+	if (s != NULL && s->uplink != me.me)
+	{
+		/* elicit PONG for EOB detection; pinging uplink is
+		 * already done elsewhere -- jilles
+		 */
+		sts(":%s PING %s %s", ME, me.name, s->sid);
 	}
 }
 
@@ -1036,7 +1437,7 @@ static void m_join(sourceinfo_t *si, int parc, char *parv[])
 
 static void m_pass(sourceinfo_t *si, int parc, char *parv[])
 {
-	if (strcmp(curr_uplink->pass, parv[0]))
+	if (strcmp(curr_uplink->receive_pass, parv[0]))
 	{
 		slog(LG_INFO, "m_pass(): password mismatch from uplink; aborting");
 		runflags |= RF_SHUTDOWN;
@@ -1050,7 +1451,8 @@ static void m_error(sourceinfo_t *si, int parc, char *parv[])
 
 static void m_sethost(sourceinfo_t *si, int parc, char *parv[])
 {
-	strlcpy(si->su->vhost, parv[0], HOSTLEN);
+	strshare_unref(si->su->vhost);
+	si->su->vhost = strshare_get(parv[0]);
 }
 
 static void m_chghost(sourceinfo_t *si, int parc, char *parv[])
@@ -1060,7 +1462,8 @@ static void m_chghost(sourceinfo_t *si, int parc, char *parv[])
 	if (!u)
 		return;
 
-	strlcpy(u->vhost, parv[1], HOSTLEN);
+	strshare_unref(u->vhost);
+	u->vhost = strshare_get(parv[1]);
 }
 
 static void m_motd(sourceinfo_t *si, int parc, char *parv[])
@@ -1083,23 +1486,35 @@ static void nick_ungroup(hook_user_req_t *hdata)
 	user_t *u;
 
 	u = hdata->si->su != NULL && !irccasecmp(hdata->si->su->nick, hdata->mn->nick) ? hdata->si->su : user_find_named(hdata->mn->nick);
-	if (u != NULL && !nicksvs.no_nick_ownership && !use_esvid)
-		sts(":%s SVS2MODE %s -r+d *", nicksvs.nick, u->nick);
+	if (u != NULL && (!use_esvid || !nicksvs.no_nick_ownership))
+		sts(":%s SVS2MODE %s -r+d 0", nicksvs.nick, u->nick);
 }
 
 static void m_protoctl(sourceinfo_t *si, int parc, char *parv[])
 {
 	int i;
 
+	has_protoctl = true;
+
 	for (i = 0; i < parc; i++)
 	{
 		if (!irccasecmp(parv[i], "ESVID"))
 			use_esvid = true;
+		else if (!irccasecmp(parv[i], "MLOCK"))
+			use_mlock = true;
+		else if (!strncmp(parv[i], "SID=", 4))
+		{
+			ircd->uses_uid = true;
+			mowgli_strlcpy(ts6sid, (parv[i] + 4), sizeof(ts6sid));
+		}
 	}
 }
 
 void _modinit(module_t * m)
 {
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "transport/rfc1459");
+	MODULE_TRY_REQUEST_DEPENDENCY(m, "protocol/base36uid");
+
 	/* Symbol relocation voodoo. */
 	server_login = &unreal_server_login;
 	introduce_nick = &unreal_introduce_nick;
@@ -1132,7 +1547,13 @@ void _modinit(module_t * m)
 	invite_sts = &unreal_invite_sts;
 	holdnick_sts = &unreal_holdnick_sts;
 	chan_lowerts = &unreal_chan_lowerts;
+	sasl_sts = &unreal_sasl_sts;
+	svslogin_sts = &unreal_svslogin_sts;
+	quarantine_sts = &unreal_quarantine_sts;
+	mlock_sts = &unreal_mlock_sts;
+	is_extban = &unreal_is_extban;
 
+	next_matching_ban = &unreal_next_matching_ban;
 	mode_list = unreal_mode_list;
 	ignore_mode_list = unreal_ignore_mode_list;
 	status_mode_list = unreal_status_mode_list;
@@ -1149,6 +1570,7 @@ void _modinit(module_t * m)
 	pcommand_add("SJOIN", m_sjoin, 2, MSRC_USER | MSRC_SERVER);
 	pcommand_add("PART", m_part, 1, MSRC_USER);
 	pcommand_add("NICK", m_nick, 2, MSRC_USER | MSRC_SERVER);
+	pcommand_add("UID", m_uid, 10, MSRC_SERVER);
 	pcommand_add("QUIT", m_quit, 1, MSRC_USER);
 	pcommand_add("UMODE2", m_umode, 1, MSRC_USER);
 	pcommand_add("MODE", m_mode, 2, MSRC_USER | MSRC_SERVER);
@@ -1156,6 +1578,7 @@ void _modinit(module_t * m)
 	pcommand_add("KILL", m_kill, 1, MSRC_USER | MSRC_SERVER);
 	pcommand_add("SQUIT", m_squit, 1, MSRC_USER | MSRC_SERVER);
 	pcommand_add("SERVER", m_server, 3, MSRC_UNREG | MSRC_SERVER);
+	pcommand_add("SID", m_sid, 4, MSRC_UNREG | MSRC_SERVER);
 	pcommand_add("STATS", m_stats, 2, MSRC_USER);
 	pcommand_add("ADMIN", m_admin, 1, MSRC_USER);
 	pcommand_add("VERSION", m_version, 1, MSRC_USER);
@@ -1170,47 +1593,9 @@ void _modinit(module_t * m)
 	pcommand_add("SETHOST", m_sethost, 1, MSRC_USER);
 	pcommand_add("CHGHOST", m_chghost, 2, MSRC_USER | MSRC_SERVER);
 	pcommand_add("MOTD", m_motd, 1, MSRC_USER);
-	pcommand_add("PROTOCTL", m_protoctl, 10, MSRC_UNREG);
-
-	/* 
-	 * for fun, and to give nenolod a heart attack
-	 * whenever he reads unreal.c, let us do tokens... --w00t
-	 * 
-	 * oh, as a side warning: don't remove the above. not only
-	 * are they for compatability, but unreal's server protocol
-	 * can be odd in places (ie not sending JOIN token after synch
-	 * with no SJOIN, not using KILL token... etc. --w00t.
-	 */
-
-	pcommand_add("8", m_ping, 1, MSRC_USER | MSRC_SERVER);
-	pcommand_add("9", m_pong, 1, MSRC_SERVER);
-	pcommand_add("!", m_privmsg, 2, MSRC_USER);
-	pcommand_add("B", m_notice, 2, MSRC_UNREG | MSRC_USER | MSRC_SERVER);
-	pcommand_add("~", m_sjoin, 2, MSRC_USER | MSRC_SERVER);
-	pcommand_add("D", m_part, 1, MSRC_USER);
-	pcommand_add("&", m_nick, 2, MSRC_USER | MSRC_SERVER);
-	pcommand_add(",", m_quit, 1, MSRC_USER);
-	pcommand_add("|", m_umode, 1, MSRC_USER);
-	pcommand_add("G", m_mode, 2, MSRC_USER | MSRC_SERVER);
-	pcommand_add("H", m_kick, 2, MSRC_USER | MSRC_SERVER);
-	pcommand_add(".", m_kill, 1, MSRC_USER | MSRC_SERVER);
-	pcommand_add("-", m_squit, 1, MSRC_USER | MSRC_SERVER);
-	pcommand_add("'", m_server, 3, MSRC_UNREG | MSRC_SERVER);
-	pcommand_add("2", m_stats, 2, MSRC_USER);
-	pcommand_add("@", m_admin, 1, MSRC_USER);
-	pcommand_add("+", m_version, 1, MSRC_USER);
-	pcommand_add("/", m_info, 1, MSRC_USER);
-	pcommand_add("#", m_whois, 2, MSRC_USER);
-	pcommand_add("b", m_trace, 1, MSRC_USER);
-	pcommand_add("6", m_away, 0, MSRC_USER);
-	pcommand_add("C", m_join, 1, MSRC_USER);
-	pcommand_add("<", m_pass, 1, MSRC_UNREG);
-	pcommand_add("5", m_error, 1, MSRC_UNREG | MSRC_SERVER);
-	pcommand_add(")", m_topic, 4, MSRC_USER | MSRC_SERVER);
-	pcommand_add("AA", m_sethost, 1, MSRC_USER);
-	pcommand_add("AL", m_chghost, 2, MSRC_USER | MSRC_SERVER);
-	pcommand_add("F", m_motd, 1, MSRC_USER);
-	pcommand_add("_", m_protoctl, 10, MSRC_UNREG);
+	pcommand_add("PROTOCTL", m_protoctl, 1, MSRC_UNREG);
+	pcommand_add("SASL", m_sasl, 4, MSRC_SERVER);
+	pcommand_add("MLOCK", m_mlock, 3, MSRC_SERVER);
 
 	hook_add_event("nick_group");
 	hook_add_nick_group(nick_group);
